#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Morten Sæthre

Paper: "On the Role of Parallel Trade on Manufacturers and Retailers Profits in the Pharmaceutical Sector"
Authors: Pierre Dubois (TSE) and Morten Saethre (NHH)

Create descriptives.

Use product-market data created in clean_data.py to generate descriptives for
Atorvastatin market.
"""


#%%     SETUP

import pandas as pd
from matplotlib import pyplot as plt
import mkl

mkl.set_num_threads(mkl.get_max_threads())
pd.options.io.parquet.engine = 'fastparquet'

# Directories
rootdir = '/home/morten/'
projectroot = f'{rootdir}projects/parallel/'
datadir = f'{projectroot}data/'
tabfigdir = f'{projectroot}tabfig/'



#%% LOAD MARKET DATA

marketdata = pd.read_parquet(f'{datadir}market.pq')



#%% TABLE 4.1: Market size, parallel share, retail and wholesale prices

marketdata['year'] = marketdata.date.dt.year

desctable = (marketdata
             .groupby(['strength', 'year'])
             .agg({'ddd': lambda x: x.sum() / 1e6, 'maxprice': 'mean'})
             )

temp = (marketdata
        .groupby(['parallel', 'strength', 'year'])
        .agg({'ddd': 'sum', 'gip': 'mean'})
        )

desctable['parshare'] = temp.ddd.loc[True] / 1e6 / desctable.ddd
desctable['digip'] = temp.gip.loc[False]
desctable['pigip'] = temp.gip.loc[True]

vardict = {'ddd': 'Defined Daily Doses (millions DDD)',
           'parshare': 'Share parallel import',
           'maxprice': 'Consumer price',
           'digip': 'Direct import wholesale price',
           'pigip': 'Parallel import wholesale price'}

desctable = desctable.rename(columns=vardict).unstack(0).T.swaplevel().sort_index()
caption = 'Market size (DDD), share of parallel imports, consumers and wholesale prices'
key = 'tbl:aggsizeshare'

print(desctable.to_latex(na_rep='-', float_format='%.2f', 
                         index_names=False, multirow=True,
                         caption=caption, label=key)
      )

with open(f'{tabfigdir}marketdesc.tex', 'w') as f:
    f.write(desctable.to_latex(na_rep='-', float_format='%.2f', 
                               index_names=False, multirow=True,
                               caption=caption, label=key))

#%% FIGURE 4.1: Parallel vs. producer sales

marketdata = marketdata.set_index(['strength', 'chainid', 'date']).sort_index()

strengthlist = [40, 80]
chainlist = [1, 2, 3]

stylemap = {'Pfizer': 'k-', 'Farmagon': 'bs--',
            'Orifarm': 'ro-.', 'Euromedica': 'gv:'}

stylemap = {'Producer': 'k-', 'Parallel': 'bs--'}


fig, ax = plt.subplots(len(chainlist), len(strengthlist),
                       figsize=(12.4, 9.8), sharex='col')

i = 0
for chain in chainlist:
    j = 0
    for strength in strengthlist:
        temp = (marketdata.loc[strength, chain]
                .set_index('parallel', append=True)
                .ddd.unstack('parallel') / 1000
               ).rename(columns={True: 'Parallel', False: 'Producer'})
        #if strength == 40:
         #   temp = temp.loc[:'2006-02']
        for firm in temp.columns:
            temp[firm].plot(ax=ax[i, j], style=stylemap[firm], legend=False)
        ax[i, j].set_title('Chain {:d}, {:d} mg'.format(chain, strength),
                           fontdict={'fontweight': 500})
        ax[i, j].set_xlabel('')
        j += 1
    i += 1
ax[1, 0].set_ylabel('DDD (1000)')
lines, labels = ax[1, 0].get_legend_handles_labels()
plt.figlegend(lines, labels, ncol=4, loc='lower center')
plt.tight_layout(rect=(0, 0.05, 1, 1))

plt.savefig(f'{tabfigdir}lipitorchainsegsales.pdf')



#%% FIGURE 4.2: Margin relative difference

strengthlist = [40, 80]
chainlist = [1, 2, 3]

stylemap = {'Pfizer': 'k--', 'Farmagon': 'bs',
            'Orifarm': 'ro', 'Euromedica': 'gv'}

temp = (marketdata.query('parallel == True').margin
        / marketdata.query('parallel == False').margin * 100 - 100)

fig, ax = plt.subplots(len(chainlist), len(strengthlist),
                       figsize=(12.4, 9.8), sharex='col', sharey=True)

i = 0
for chain in chainlist:
    j = 0
    for strength in strengthlist:
        plottemp = temp.loc[strength, chain]
        plottemp.plot(ax=ax[i, j], style='k-', legend=False)
        ax[i, j].set_title('Chain {:d}, {:d} mg'.format(chain, strength),
                           fontdict={'fontweight': 500})
        ax[i, j].set_xlabel('')
        j += 1
    i += 1

vals = ax[0, 0].get_yticks()
ax[0, 0].set_yticklabels(['{:3.0f}%'.format(x) for x in vals])
#ax[0, 0].set_ylim(0, 30)
ax[1, 0].set_ylabel(r'$\%\Delta m$', rotation=0, labelpad=30, size=18)
plt.tight_layout()

plt.savefig(f'{tabfigdir}lipitorchainsegmargindiff.pdf')