#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Morten Sæthre

Paper: "On the Role of Parallel Trade on Manufacturers and Retailers Profits in the Pharmaceutical Sector"
Authors: Pierre Dubois (TSE) and Morten Saethre (NHH)

Create dataset for Atorvastatin/Lipitor (ATC C10AA05)

Inputs:
    Transaction data from the Norwegian Directorate of Health 
    Wholesale price data for 2004--2007 received from Kurt Brekke (originating
from the Norwegian Institute of Public Health)
    Producer/marketing firm of each drug based on information from the Norwegian
Medicines Agency
    Chain affiliation of each pharmacy based on information from the Norwegian
Medicines Agency
    Data on regulation (price ceilings) from the Norwegian Medicines Agency
    Wholesale prices in (potential) source countries based on information from IMS
    Parallel import license files from the Norwegian Medicines Agency to
determine source area of products
    Exchange rates (monthly) downloaded from the central bank of Norway

Outputs:
    Individual level choice data
    Market level aggregate sales data
    Drug ID and chain ID sales data
    Doctor statin prescription data

Years 2004 to 2007
"""

#%% SETUP

import numpy as np
import pandas as pd
import mkl
import argparse
mkl.set_num_threads(mkl.get_max_threads())

# Directories
rootdir = '/home/morten/'
datadir = f'{rootdir}data/'
projectroot = f'{rootdir}projects/parallel/'

# Get command line arguments (or use default)
parser = argparse.ArgumentParser()
parser.add_argument('-ks', '--dropsmall', action='store_true',
                    help='Keep small packages')

args = parser.parse_args()


# Check if set to keep small packages (30 tablets)
if args.dropsmall:
    print("Removing small packages")
else:
    print("Keeping small packages")



#%% LOAD DATA

### Load information on each drug ID
# Utility file created from the full transaction data from the Directorate of Health
# Contains ID, name and ATC code for all prescription drug products in the period
druginfo = pd.read_pickle(f'{datadir}heldir/druginfo.pkl')
# Get drug IDs for Atorvastatin
lipitoratc = 'C10AA05'
druglist = druginfo.query(f'atccode == "{lipitoratc}"').drugid.astype(int).unique()


### Load transaction data
# List of years and files to use
yearlist = range(2004, 2008)
filelist = [f'{datadir}heldir/heldir_data_{y}.pkl' for y  in yearlist]

### Load transaction data for Atorvastatin
data = pd.DataFrame()
for f in filelist:
    temp = pd.read_pickle(f)
    # Drop drugs not in list of drugs with relevant ATC code
    dropmask = temp.drugid.isin(druglist) == False
    temp.drop(temp[dropmask].index, inplace=True)
    data = data.append(temp, ignore_index=True)

# Make date monthly, keeping daily date
data.rename(columns={'date': 'daydate'}, inplace=True)
data['date'] = pd.to_datetime(data.daydate.dt.year * 100 + data.daydate.dt.month,
                              format='%Y%m')
data['drugid'] = data.drugid.astype(int)

### Load regulatory data
regdata = pd.read_csv(f'{datadir}drugs/SLVregulation.csv')
#Only keep relevant products
dropmask = regdata.drugid.isin(druglist) == False
regdata.drop(regdata[dropmask].index, inplace=True)


### Load drug company name
# Utility file with information from the Norwegian Medicines Agency on
# companies holding the marketing license for each drug ID
drugcompany = pd.read_csv(f'{datadir}drugs/drugcompany.csv')
drugcompany = drugcompany.set_index('varenummer').company
drugcompany.name = 'firmname'

### Load pharmacy chain affiliation
phidchain = pd.read_csv(f'{datadir}pharmacy/pharmacy_affiliation.csv',
                        parse_dates=['date'], infer_datetime_format=True,
                        encoding='latin1')


### Load GIP (wholesale price) data
gipdata = pd.read_csv(f'{datadir}drugs/gip_atorvastatin.csv')


### Load parallel import license data
parlic = pd.read_csv(f'{datadir}drugs/parallelmktlicense.csv')
parlic.rename(columns={'varenummer': 'drugid'}, inplace=True)

### Load IMS wholesale data
imsprices = pd.read_pickle(f'{projectroot}data/ims_quarterly.pkl')

### Load exchange rates
exchrate = pd.read_table(f'{projectroot}data/exchrates.csv',
                         sep=';')



#%% Prepare regulatory data for Atorvastatin

for c in ['containeramount', 'strength']:
    regdata[c] = regdata[c].astype(int)
# Set containeramount in cases where registered as amount per unit (e.g., tablet)
regdata['containeramount'] = (regdata.containeramount
                              .where(regdata.containers.isnull(),
                                     regdata.containers * regdata.containeramount)
                             ).astype(int)

# Assert unique drug characteristics (containeramount and strength)
assert(regdata.groupby('drugid').containeramount.nunique().max() == 1)
assert(regdata.groupby('drugid').strength.nunique().max() == 1)



#%% Get drug characteristics

# From regulatory data
drugchar = regdata.groupby('drugid')[['containeramount', 'strength']].max()

### Add drug upstream firm name to characteristics
drugchar = drugchar.join(drugcompany)

# Parallel import indicator (any other firm than Pfizer)
drugchar['parallel'] = drugchar.firmname != 'Pfizer'

# Add characteristics to data
data = data.join(drugchar, on='drugid')

# Create DDD measure
data['ddd'] = data.containeramount * data.packages * data.strength / 10



#%% Correct individual pharmacy sales and merge on pharmacy chain affiliation

# Correct sales attributed to wrong pharmacy before 2010 (203 transactions)
temp = ~((data.phid == 1711) & (data.date < '2010-01'))
data['phid'] = data.phid.where(temp, 1601)

# Fix observations to match up with date of pharmacy operation time
# Affects 8 (1 + 6 + 1) transactions registered in the month before registered opening of 3 pharmacies
# Move to registered month of opening
phidlist = [1580, 1625, 1653]
datelist = pd.to_datetime(['2005-10-01', '2007-04-01', '2007-10-01'])
for p, d in zip(phidlist, datelist):
    temp = (data.phid == p) & (data.date < d)
    print(f'Number of transactions for {p}: {temp.sum()}')
    data['date'] = data.date.where(~temp, d)

# Add anonymized chain affiliation to sales data
data = data.join(phidchain.set_index(['phid', 'date']).chainid,
                 on=['phid', 'date'])



#%% Cleaning of sales data

# Create bool masks for dropping
# Drop one transaction at one pharmacy, registered after pharmacy closed
drop_afterclose = (data.phid == 1604) & (data.date > '2007-11')

# Drop two months of very high sales registered before pharmacy opening
# Sales in Jan-Feb 2004, pharmacy opens in Feb 2006
# (wholesale transactions erroneously registered on pharmacy)
drop_wholesale = (data.phid == 1586) & (data.date < '2004-06-01')

### Drugs which have very few sales within chain
temp = data.groupby(['chainid', 'drugid']).price.count()
# Drop all drugs which have less than 20 transactions within chain
# (16 is the highest #transactions satisfying this, while next lowest is 48)
salelim = 20
# Remove combinations of chain and drug ID (83 transactions)
drop_verylow = np.zeros(len(data), dtype=bool)
for c, d in temp[temp < salelim].index:
    drop_verylow = drop_verylow | ((data.chainid == c) & (data.drugid == d))

# Also remove a rare product featuring in 2004, which lacks regulatory information
# (name could indicate samples or similar, 2,057 transactions)
drop_noregulate = data.drugid == 245183

# Remove observations without pharmacy ID (2,069 transactions)
drop_nophid = data.phid.isnull()

# Drop negative quantities (81 transactions)
# (Ignore returns due to small number)
drop_negative = data.packages < 0

# Drop transactions with more than 4 packages (15 transactions)
data.packages.value_counts().sort_index().tail(6)
drop_manypacks = data.packages > 4

# Combination of drop criteria so far
drop_combine = (drop_afterclose | drop_wholesale | drop_verylow | drop_noregulate 
                | drop_nophid | drop_negative | drop_manypacks)

# Drop hospital pharmacies (221,102 transactions)
# List of main chain ID's
mainchainlist = [x for x in range(1, 4)]
drop_hospital = ~data.chainid.isin(mainchainlist)
# Overlap of 2,687 transactions with previous drop criteria (60%)
(drop_combine & drop_hospital).sum()

data.drop(data[drop_combine | drop_hospital].index, inplace=True)

# Drop markets (month-strength combinations) lacking parallel trade
temp = (data.groupby(['chainid', 'date', 'strength']).parallel.max()
        .groupby(['date', 'strength']).sum()
       )
# 40% of markets feature parallel trade in all three chains
temp.unstack()
drop_nopar = data[temp.index.names].join(temp, on=temp.index.names).parallel < 3

data.drop(data[drop_nopar].index, inplace=True)

# Transaction sample size (238,268 transactions)
len(data)

if args.dropsmall:
    ### Drop small package sales (1,681 transactions, 0.3% of revenue)
    drop_smallpack = data.containeramount == 30
    data.drop(data[drop_smallpack].index, inplace=True)


# Relevant drug IDs
datadrugs = data.drugid.unique()

# Make chain ID integer
data['chainid'] = data.chainid.astype(int)



#%% Create price ceiling series from information on regulation

# Create date information in regulatory data
regdata['year'] = regdata.period.str[:4].astype(int)

# Find maximal price series based on their changes
# Choice of drug representative of maximum prices
maxdrug = 3048
drugmask = regdata.drugid == maxdrug
regdatemask = regdata.year <= 2007
umaxaup = regdata[drugmask & regdatemask].maxaup.unique()
# Get the (minimum) period indices for these maximum prices as well
umaxaupperiod = [regdata[drugmask & (regdata.maxaup == x)].period.min() for x in umaxaup]
# Get the minimu m date in data where the price of the drug is equal to each of the maximum prices
umaxaupdate = [data[(data.drugid == maxdrug) & (data.price == x)].daydate.min()
               for x in umaxaup]
# Turn into month-start dates
umaxaupdate = pd.to_datetime([x.year * 100 + x.month for x in umaxaupdate],
                             format='%Y%m').tolist()

# Check that all period indices are unique
# (that one will actually get separate maximum prices for each predicted change)
assert(len(np.unique(umaxaupperiod)) == len(umaxaupperiod))

maxaupchange = pd.DataFrame({'period': umaxaupperiod, 'date': umaxaupdate})

# Generate maximum price series by strength and package size
maxaupgroupcols = ['strength', 'containeramount', 'period']
# Assert that there is no variation in maximum prices within this class in each period
assert(regdata[regdatemask].groupby(maxaupgroupcols).maxaup.std().max() == 0)
# Average within grouping (no variation)
temp = regdata[regdatemask].groupby(maxaupgroupcols).maxaup.mean()
maxaupdata = maxaupchange.merge(temp.reset_index(), on='period', how='inner')
maxaupdata = maxaupdata.drop('period', axis=1)

# Make series cover until the end of the sample
maxsaledate = data.date.max()
# Repeat last date entry, changing the date values to the last date in sample
temp = maxaupdata[maxaupdata.date == max(umaxaupdate)].copy()
temp['date'] = maxsaledate
maxaupdata = maxaupdata.append(temp)
maxaupdata.sort_values(['date', 'strength', 'containeramount'], inplace=True)

maxaupseries = (maxaupdata.set_index('date')
                .groupby(['strength', 'containeramount'])
                .resample('MS')
                .ffill()
                .maxaup
               )
maxaupseries.name = 'maxprice'

# Add price ceiling information to sales data
data = data.join(maxaupseries, on=['strength', 'containeramount', 'date'])

# Assert that all transactions have price ceiling assigned
assert(data.maxprice.isnull().sum() == 0)



#%% Add wholesale prices

gipdata['date'] = pd.to_datetime(gipdata.date)
gipdata = gipdata.set_index(['date', 'chainid', 'drugid']).gip.sort_index()

data = data.join(gipdata, on=['date', 'chainid', 'drugid'])


# Assert that all transactions have wholesale price assigned
assert(data.gip.isnull().sum() == 0)



#%% Scale price variable to NOK per DDD

renamedict = {'price': 'pricepack', 'maxprice': 'maxpricepack', 'gip': 'gippack'}
data.rename(columns=renamedict, inplace=True)

for k, v in renamedict.items():
    data[k] = data[v] * data.packages / data.ddd



#%% Create drug-pharmacy sales data

# Create volume data for each monetary measure (market, wholesale and max price)
pricevol = list(zip(['price', 'maxprice', 'gip'], ['revenue', 'maxrevenue', 'giprevenue']))
for col, v in pricevol:
    data[v] = data[col] * data.ddd

marketsumcols = ['revenue', 'maxrevenue', 'giprevenue', 'ddd']
marketixcols = ['date', 'chainid', 'drugid']
product = data.groupby(marketixcols, as_index=False)[marketsumcols].sum()

for col, v in pricevol:
    product[col] = product[v] / product.ddd

product = product.join(drugchar, on='drugid')

product['margin'] = product.price - product.gip



#%% Select relevant parallel importer per chain

# Highest share parallel trade (average share is 94%)
# (smaller sales of previous stocks will occur)
tempcols = ['parallel', 'date', 'strength', 'chainid']
temp = product.groupby(tempcols).ddd.transform('max')
# Average share of highest share parallel trade brand is 94%
(product.groupby(tempcols).ddd.max() / product.groupby(tempcols).ddd.sum()).loc[True].mean()

propcompany = product[product.ddd == temp].groupby(['date', 'strength', 'chainid', 'parallel']).firmname.first()



#%% Get source country from license data

# Get drug ID's from product-market aggregate data
temp = parlic[parlic.drugid.isin(datadrugs)].copy()
temp['drugid'] = temp.drugid.astype(int)

# Two products have two registries
# Use the updated license in 2004
drop_oldlicense = temp.licyear != temp.groupby('drugid').licyear.transform('max')
temp.drop(temp[drop_oldlicense].index, inplace=True)
temp.set_index('drugid', inplace=True)
temp.rename(columns={'exportcountry': 'sourcecountry'}, inplace=True)

temp['sourcecountry'] = temp.sourcecountry.replace('Czech Republic', 'Czech')


# Get list of source countries
sourcelist = temp.sourcecountry.unique()

# Add source country to product sales data (setting to "None" for Pfizer)
product = product.join(temp.sourcecountry, on='drugid').fillna('None')

# Create indicators for each country multiplied by ddd
product = product.join(pd.get_dummies(product.sourcecountry, prefix='share', prefix_sep='').mul(product.ddd, axis=0))



#%% Create choice-market data

salesumcols = ['revenue', 'maxrevenue', 'giprevenue', 'ddd']
saleixcols = ['date', 'strength', 'chainid', 'parallel']

market = product.groupby(saleixcols)[salesumcols].sum()

# Add parallel import company used
market = market.join(propcompany)

# Add source sales for each choice-market observation
sourcecols = [f'share{x}' for x in sourcelist]
market = market.join(product.groupby(saleixcols)[sourcecols].sum())

# Scale to share of DDD
for c in sourcecols:
    market[c] = market[c] / market.ddd


market.reset_index(inplace=True)

# Fix somy upstream firm assignments
temp = market.parallel & (market.chainid == 1) & (market.strength == 40) & (market.date > '2005')
market['firmname'] = market.firmname.where(~temp, 'Orifarm')

temp = market.parallel & (market.chainid == 2) & (market.strength == 80)
market['firmname'] = market.firmname.where(~temp, 'Farmagon')

temp = market.parallel & (market.chainid == 3) & (market.strength == 80) & (market.date >= '2005')
market['firmname'] = market.firmname.where(~temp, 'Farmagon')

# Prices per DDD
for col, v in pricevol:
    market[col] = market[v] / market.ddd


# Market share and share within chain (conditional share)
market['share'] = market.ddd / market.groupby(['date', 'strength']).ddd.transform('sum')
market['condshare'] = market.ddd / market.groupby(['date', 'strength', 'chainid']).ddd.transform('sum')


market['margin'] = market.maxprice - market.gip

market.sort_values(['date', 'strength', 'chainid', 'parallel'], inplace=True)



#%% Assign source area

# Highest share country
A = market[sourcecols].values
B = np.column_stack((1 - A.sum(1), A))
# Highest share in each row
source_maxshare = np.hstack(('None', sourcelist))[B.argmax(1)]
market['sourcemax'] = source_maxshare

# The distribution of max share sources for each upstream company suggests a 
# correspondence between firm and source area
# (4 cases total, at most 2.5% of cases for any upstream firm is for a different country)
market.groupby('firmname').sourcemax.value_counts()
sourcecountrymap = {'Euromedica': 'UK', 'Farmagon': 'France',
                    'Orifarm': 'UK', 'Pfizer': 'None'}

market['sourcecountry'] = market.firmname.map(sourcecountrymap)



#%% Create market, choice, market-choice and upstream firm numbering/ID

market.sort_values(['strength', 'date', 'chainid', 'parallel'], inplace=True)
market.reset_index(drop=True, inplace=True)
# Market ID (combination of concentration/strength and date)
market['market'] = market.groupby(['strength', 'date']).ngroup()
market['marketddd'] = market.groupby(['strength', 'date']).ddd.transform('sum')

# Add market ID to other aggregations
marketinfo = market.groupby(['strength', 'date'])[['market', 'marketddd']].first()
product = product.join(marketinfo, on=['strength', 'date'])
data = data.join(marketinfo.market, on=['strength', 'date'])


# Choice ID (combination of chain and parallel/direct)
choiceix = pd.MultiIndex.from_product([mainchainlist, [False, True]],
                                       names=['chainid', 'parallel'])
choiceid = pd.Series(range(len(choiceix)), choiceix, name='choiceid')

market = market.join(choiceid, on=['chainid', 'parallel'])
product = product.join(choiceid, on=['chainid', 'parallel'])
data = data.join(choiceid, on=['chainid', 'parallel'])


# Market-choice ID (combination of market and choice ID's)
market['mcid'] = market.groupby(['market', 'choiceid']).ngroup()
mcid = market.groupby(['market', 'choiceid']).mcid.first()

product = product.join(mcid, on=['market', 'choiceid'])
data = data.join(mcid, on=['market', 'choiceid'])


# Upstream firm ID
firmmap = {'Pfizer': 0, 'Farmagon': 1, 'Orifarm': 2, 'Euromedica': 3}

market['firm'] = market.firmname.map(firmmap)
product['firm'] = product.firmname.map(firmmap)
data['firm'] = data.firmname.map(firmmap)


# Sort data, ensure parallel is boolean and reset indices
market.sort_values(['market', 'choiceid'], inplace=True)
product.sort_values(['market', 'choiceid', 'drugid'], inplace=True)
data.sort_values(['patid', 'market', 'choiceid'], inplace=True)

for d in [market, product, data]:
    d['parallel'] = d.parallel.astype(bool)
    d.reset_index(drop=True, inplace=True)



#%% Add exchange rates to market data

exchrate['date'] = pd.to_datetime(exchrate.date, format='%Y-%m-%d')
market['date'] = pd.to_datetime(market.date)

exchrate = exchrate.set_index('date')

# Select rates to keep and add to market data
keeprates = ['nokusd', 'nokeur', 'nokgbp', 'nokchf', 'nokpln', 'nokczk']
market = market.join(exchrate[keeprates], on='date')



#%% Add IMS wholesale data to market data

imsprices.sort_values(['date', 'country', 'strength'], inplace=True)
# Resample data to monthly level
imsprices = (imsprices.set_index(['date', 'country', 'strength'])
             .unstack(['country', 'strength'])
             .resample('MS').bfill()
             .stack(['country', 'strength'])
            )

imsprices = imsprices.price.unstack('country')

market = market.join(imsprices, on=['date', 'strength'])

# Scale price by exchange rate and DDD per tablet (NOK per DDD)
for c in imsprices.columns:
    market[c + 'n'] = market[c] * market.nokusd / market.strength * 10



#%% Store individual level data

indcols = ['date', 'strength', 'chainid', 'mcid',
           'parallel', 'firm',
           'ddd', 'packages', 'copayment',
           'female', 'age']

# Drop missing patient ID 
# (included in market shares, but less useful for estimating preference heterogeneity)
data.drop(data[data.patid.isnull()].index, inplace=True)

# Some cases of multiple transactions of the same choice in the same month (3% of cases)
data.groupby(['patid', 'market', 'choiceid']).ddd.count().value_counts()
# Collapse multiple transactions of choice within market
icdict = {'date': 'first', 'strength': 'first', 'chainid': 'first', 'mcid': 'first',
          'parallel': 'first', 'firm': 'first',
          'ddd': 'sum', 'packages': 'sum', 'copayment': 'sum',
          'female': 'first', 'age': 'first'}

data = data.groupby(['patid', 'market', 'choiceid'], as_index=False).agg(icdict)

# Take random sample of 50,000 individuals
ninddraw = 50000
patidlist = data.patid.unique()
usepatid = np.random.choice(patidlist, ninddraw, replace=False)

dropmask = data.patid.isin(usepatid) == False
data.drop(data[dropmask].index, inplace=True)

data.to_parquet(f'{projectroot}data/individual.pq')



#%% Prepare sale data for storage

dropcols = ['revenue', 'maxrevenue', 'giprevenue']

market.drop(dropcols, axis=1, inplace=True)
# Indicator for maximum-margin option in chain and market
temp = market.groupby(['market', 'chainid']).margin.transform('max')
market['ismaxoption'] = market.margin == temp

# Separate indicators for each chain and each firm
market = market.join(pd.get_dummies(market.chainid, prefix='chain', prefix_sep=''))

market.to_parquet(f'{projectroot}data/market.pq')



#%% Store drug-chain data as well

product.to_parquet(f'{projectroot}data/product.pq')



#%% Notes: Remaining
# TODO: Generate doctor prescription data
# TODO: Incorporate IMS wholesale prices
# TODO: Generate parallel trade sales reduced form data for other molecules