#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Morten Sæthre

Paper: "On the Role of Parallel Trade on Manufacturers and Retailers Profits in the Pharmaceutical Sector"
Authors: Pierre Dubois (TSE) and Morten Saethre (NHH)

Estimation of upstream model for parallel trade.

Inputs:
    Product-market level data containing market shares and margins stored in "data/market.pq" created in "clean_data.py"
    Estimates from demand model:
        Estimated parameters governing preference heterogeneity
        Mean utility parameters for each product-market combination
        Foreclosure parameters for each chain-market combination

Outputs:
    Table of bargaining parameter estimates (Table 4.5)
    Tables for counterfactual simulations (Tables 5.1, 5.2, 5.3 and 5.4)
"""

#%% SETUP

from numpy import array, diag, arange, eye, zeros, ones, log, exp, hstack, \
    row_stack, column_stack, repeat
from numpy.linalg import solve
from numpy.random import rand
import numpy as np
from scipy.optimize import minimize, root
from scipy.stats import chi2
from scipy.linalg import block_diag
from timeit import default_timer as dtim
import pandas as pd
import pickle
from sklearn.utils.extmath import cartesian
from numpy.polynomial.hermite import hermgauss
from patsy import dmatrix
import theano
import theano.tensor as T
from theano.tensor.slinalg import solve as tsolve
import argparse
import re
import mkl

mkl.set_num_threads(mkl.get_max_threads())
pd.options.io.parquet.engine = 'fastparquet'

# Get command line arguments for running estimation (or use default)
parser = argparse.ArgumentParser()
parser.add_argument('-nu', '--nugroup', default=4, type=int,
                    help='Number of latent classes')
parser.add_argument('-qn', '--quadnodes', default=5, type=int,
                    help='Number of quadrature nodes per dimension of heterogeneity')
parser.add_argument('-fp', '--filepostfix', default=f'',
                    help="Postfix for result files")
args = parser.parse_args()

print(f'Number of latent classes: \t {args.nugroup}')
print(f'Number of quadrature nodes: \t {args.quadnodes}')
print(f'File postfix: \t {args.filepostfix}')

# Directories
rootdir = '/home/morten/'
projectroot = f'{rootdir}projects/parallel/'
datadir = f'{projectroot}data/'
resultdir = f'{projectroot}results/'
tabfigdir = f'{projectroot}tabfig/'



#%%    LOAD DATA

print("Loading data")

marketdata = pd.read_parquet(f'{datadir}market.pq')

# Load demand estimates
demand_res = pickle.load(open(f'{resultdir}demand_res{args.filepostfix}.pkl',
                              mode='rb'))
theta = pd.read_pickle(f'{resultdir}theta{args.filepostfix}.pkl')
delta = pd.read_pickle(f'{resultdir}delta{args.filepostfix}.pkl')

# Load demand estimates and parameters for France
demand_france = pd.read_stata(f'{datadir}blp_coef.dta')

# Extract estimated parameters
fr_b_cols = [c for c in demand_france.columns if re.match('b\d+', c)]
# Parameters governing gross utility (mean utility without price)
fr_grossutil_param = demand_france[fr_b_cols].iloc[0, :-2].values
fr_price_param = demand_france[fr_b_cols].iloc[0, -2]
fr_sd_param = demand_france[fr_b_cols].iloc[0, -1]
fr_bargain_param = demand_france.bFRoptimal.iloc[0]
# Drop columns containing parameters
demand_france.drop(fr_b_cols + ['bFRoptimal'], axis=1, inplace=True)

demand_france['strength'] = demand_france.strength.str.replace('"|MG', '').astype('int')
demand_france['date'] = pd.to_datetime(demand_france.quarter, format='%Y%m')

rightsholder = {'ATORVASTATIN': 'PFIZER',
                'FLUVASTATIN': 'NOVARTIS',
                'PRAVASTATIN': 'BRISTOL-MYERS SQB.',
                'SIMVASTATIN': 'MERCK & CO',
                'ROSUVASTATIN': 'ASTRAZENECA'}

demand_france['bargainer'] = 0
for k, v in rightsholder.items():
    mask = (demand_france.molecule == k) & (demand_france.corp == v)
    demand_france.loc[mask, 'bargainer'] = 1

demand_france['bargainer'] = demand_france.bargainer.astype(bool)

# Ensure sorting by market
demand_france.sort_values(['date', 'strength', 'molecule', 'corp'],
                          inplace=True)

# Drop unnecessary columns
demand_france.drop(['num_mkt', 'quarter'], axis=1, inplace=True)

# Columns corresponding to parameters of gross utility
fr_molstr_var = [c for c in demand_france.columns if c.startswith('molstr_')]
fr_q_var = [c for c in demand_france.columns if c.startswith('q_')]
fr_grossutil_var = fr_molstr_var +  fr_q_var + ['molecule_loe']

# Calculate gross utility
# Observed part
demand_france['obsgrossutil'] =\
    fr_grossutil_param[0] + demand_france[fr_grossutil_var] @ fr_grossutil_param[1:]
# Total
demand_france['grossutil'] = demand_france.obsgrossutil + demand_france.delta

demand_france.drop(fr_grossutil_var, axis=1, inplace=True)

# Keep only relevant dates
dropmask = (demand_france.date < '2003-12-01') | (demand_france.date > '2007-12-01')
demand_france.drop(demand_france[dropmask].index, inplace=True)

# Ratio of generic price to originator price (regulated)
fr_genpriceratio = 0.6
# Relative reduction in generic price from originator price
fr_rgenprice = 1 - fr_genpriceratio

# Nodes and preference heterogeneity
fr_nodes = pd.read_stata(f'{datadir}blp_draws.dta').values
fr_msim = len(fr_nodes)
fr_pricepref = fr_price_param + fr_sd_param * fr_nodes.T

# Make monthly, repeating quarterly values
demand_france = (demand_france.set_index('date')
                 .groupby(['molecule', 'strength', 'corp'])
                 .resample('MS').bfill()
                 .drop(['molecule', 'strength', 'corp'], axis=1)
                 .reset_index()
                 )

marketdef = marketdata.groupby(['strength', 'date']).market.first()
demand_france = demand_france.join(marketdef, on=['strength', 'date'])
mask = demand_france.market.isnull()
demand_france.drop(demand_france[mask].index, inplace=True)
demand_france['market'] = demand_france.market.astype(int)
#demand_france.reset_index(drop=True, inplace=True)

# Add molecule ID
moleculedict = {'ATORVASTATIN': 0,
                'FLUVASTATIN': 1,
                'PRAVASTATIN': 2,
                'SIMVASTATIN': 3}
demand_france['moleculeid'] = demand_france.molecule.map(moleculedict)
# Molecule dummies
demand_france = demand_france.join(pd.get_dummies(demand_france.moleculeid, 'mol', ''))
dmolcols = [f'mol{i}' for i in range(demand_france.moleculeid.nunique())]

# Sort data
demand_france.sort_values(['market', 'moleculeid', 'generic'],
                          inplace=True)
# Add product ID
demand_france['prodid'] = demand_france.groupby(['moleculeid', 'corp']).ngroup()

demand_france.set_index(['market', 'prodid'], inplace=True)

# Market info - sizes, weights and conversions
marketinfo = demand_france.groupby('market')[['date', 'strength']].first()
# Exchange rate
exchrates = pd.read_csv(f'{datadir}exchrates.csv', parse_dates=['date'])
exchrates.set_index('date', inplace=True)
marketinfo = marketinfo.join(exchrates.nokusd, on='date')
# Adjust from quarterly to monthly size in France
frsize = demand_france.eval('su_ / share / 3').groupby('market').mean().rename('frsize')
norsize = marketdata.groupby('market').marketddd.first().rename('norsize') / 1000
marketinfo = marketinfo.join([frsize, norsize])
# Weight on France, monetary terms with conversion to NOK
marketinfo['frweight'] = marketinfo.eval('frsize / norsize * nokusd')
# Lipitor price
mask = demand_france.molecule == 'ATORVASTATIN'
marketinfo['frlipitorprice'] = demand_france[mask].reset_index('prodid').price
# Scaling from French lipitor price in USD/tablet to NOK/DDD (old DDD definition)
marketinfo['lipitorpriceconvert'] = marketinfo.eval('nokusd / strength * 10')
# price in NOK/DDD 
marketinfo['frlipitorpricen'] = marketinfo.eval('frlipitorprice * nokusd / strength * 10')

# Indexing objects
# Index of market x molecule (for prices and MC of bargaining companies)
frmbargainix = (demand_france.query('bargainer').reset_index()
                .set_index(['market', 'moleculeid'])
                .index
                )



#%%    SET VARIABLES

# Choice ID to normalize utility against
basechoice = 0

# Number of chains
nchain = marketdata.chainid.nunique()
# Chain dummy columns
chaincols = ['chain' + str(i) for i in range(1, 4)]

# Get list of markets
marketlist = marketdata.market.unique()
nmarket = len(marketlist)

# Make indicators for maximum margin option and parallel trade 0/1
for c in ['ismaxoption', 'parallel']:
    marketdata[c] = marketdata[c].astype(int)

# Dummies for each upstream firm
marketdata = marketdata.join(pd.get_dummies(marketdata.firm, prefix='firm', prefix_sep=''))
firmcols = [f'firm{i}' for i in range(4)]
# Dummy for French sourced parallel trade
marketdata['francesource'] = marketdata.sourcecountry == 'France'

# Series mapping choice ID to parallel status
choiceidtopar = marketdata.groupby('choiceid').parallel.first()


#%% PREPARE ID'S AND OBJECTS FOR SIMULATION

# Features of unobserved heterogeneity
varcols = ['parallel']
varcols.extend(chaincols[1:])
prefdistdim = len(varcols)
nugroup = args.nugroup
# Number of active parameters for each latent class
nugroupparams = prefdistdim

simdraws = args.quadnodes
simlength = args.quadnodes ** prefdistdim

#Number of parameters to be estimated by maximum likelihood
"""
One parameter per dimension of heterogeneity for each latent class
Separate mean utility for each dimension of heterogeneity for each latent class
One nesting paremeter per chain
Parmeters for probability simplex of latent classes (#classes - 1)
"""
nmlparam = nugroup * prefdistdim + (nugroup - 1) * prefdistdim + nchain + nugroup - 1

# Verify setup has same number of parameters as in loaded estimates
assert nmlparam == len(demand_res.x)

### Index objects
#Index for latent class parameters
ugroupdistix = pd.MultiIndex.from_product((varcols, arange(nugroup)),
                                          names=['variable', 'latclass'])
# Make simulation draw index
simix = pd.RangeIndex(simlength, name='sim')

#Bargaining parameter index
bargainix = pd.MultiIndex.from_arrays([[0, 1, 2, 3] * 3, 
                                       [2, 3, 3, 3, 4, 5, 5, 5, 6, 7, 7, 7]],
                                      names=['firm', 'choiceid']
                                      )
#Sequence of bargaining parameters (For matching with observed choices)
bargainparseq = pd.Series(range(len(bargainix)), bargainix)

#Market index
marketix = pd.Index(marketlist, name='market')

#Unique choice ID's
choiceidlist = marketdata.choiceid.unique()

#Get quadrature for market share integration
singlenodes, singleweights = hermgauss(args.quadnodes)
# Cartesian product of nodes over dimensions of heterogeneity
basenodes = cartesian([singlenodes] * prefdistdim).T
# Make weights reflect cartesian product of nodes over dimensions and rescale (sum to one)
weightscale = np.pi ** (-prefdistdim / 2.)
baseweights = weightscale * cartesian([singleweights] * prefdistdim).prod(1)
# Repeated over latent classes (rows indexed as in ugroupdistix)
repnodes = repeat(basenodes, nugroup, 0)

# Set index for market-choice data
marketindexcols = ['market', 'chainid', 'choiceid']
marketdata.set_index(marketindexcols[0], inplace=True)
marketdata.sort_index(inplace=True)

# Matrices for selecting bargaining parameters for the relevant choices
bwsel = dict(zip(marketlist, [array([[1 * (bargainparseq[x[0], x[1]] == j)
                                         for j in range(12)]
                                        for x in marketdata.loc[m][['firm', 'choiceid']].values])
                                 for m in marketlist]))

marketdata.set_index(marketindexcols[1:], append=True, inplace=True)
marketdata.sort_index(inplace=True)

# Estimated parameter vector
param = demand_res.x



#%%     CALCULATE UTILITY FROM DEMAND ESTIMATES

def createindparams(param):
    """
    Create parameters governing distribution of individual preferences based
    on array with stacked parameters.

    Parameters
    ----------
    param : ndarray (nmlparam,)
        Array of stacked parameters governing distrbution of individual preferences.

    Returns
    -------
    ugroupmean : pandas.Series (nugroup x nugroupparams,)
        Parameters shifting average utility for each characteristic for latent
        classes (first latent class is baseline with zeros).
    ugroupdist : pandas.Series (nugroup x nugroupparams,)
        Parameters governing dispersion of utility for each characteristic for
        latent classes.
    lda : ndarray (nchain,)
        Nesting parameter for each chain.
    eta : ndarray (nugroup - 1,)
        Parameters governing logit probability for latent classes (implicitly
        zero for first class).
    """
    j = 0
    # Utility dispersion for latent classes
    ndist = nugroupparams * nugroup
    ugroupdist = pd.Series(param[j:j + ndist],
                          index=ugroupdistix,
                          name='latclassvar')
    j = j + ndist
    # Mean utility shifts for latent classes
    nmean = nugroupparams * (nugroup - 1)
    meanpar = column_stack((zeros(nugroupparams),
                            param[j:j + nmean].reshape(nugroupparams, nugroup - 1)
                            ))
    ugroupmean = pd.Series(meanpar.flatten(),
                           index=ugroupdistix,
                           name='latclassmean')
    j = j + nmean
    # Nesting parameters
    lda = param[j:j + nchain]
    j = j + nchain
    # Probability of latent classes
    nprob = nugroup - 1
    eta = param[j:j + nprob]
    return ugroupmean, ugroupdist, lda, eta

def getutil(eta, ugroupmean, ugroupdist, delta):
    """
    Create data on utility at each integration node from parameters governing
    distribution of individual preferences and mean utility (delta).

    Parameters
    ----------
    eta : ndarray (nugroup - 1,)
        Parameters governing logit probability for latent classes.
    ugroupmean : ndarray (nugroup x nugroupparams,)
        Parameters shifting average utility for each characteristic for latent
        classes (first latent class is baseline with zeros).
    ugroupdist : ndarray (nugroup x nugroupparams,)
        Parameters governing dispersion of utility for each characteristic for
        latent classes.
    delta : ndarray (nmarket x nprod,)
        Vector of mean utility parameters.

    Returns
    -------
    devutil : pandas.DataFrame (nmarket x nprods, nsim x nugroup)
        Utility deviations/over integration nodes x latent classes.
    weights : ndarray (nsim x nugroup)
        Weights for integrating over nodes and latent classes.
    """
    # Preference distribution
    prefdist = pd.DataFrame(repnodes * ugroupdist[:, None] + ugroupmean[:, None],
                            index=ugroupdistix, columns=simix)
    prefdist = (prefdist.unstack(1).swaplevel(axis=1).
                sort_index(axis=1).loc[varcols]
                )

    devutil = marketdata[varcols].dot(prefdist)
    # Normalize utility deviations per draw per market
    normcols = ['market']
    devutil = devutil - devutil.groupby(level=normcols).max()
    util = devutil.add(delta, axis=0)
    
    expgroup = hstack((1., exp(eta)))
    ugroupprob = expgroup / expgroup.sum()
    weights = hstack([baseweights * z for z in ugroupprob])
    return util, weights

ugroupmean, ugroupdist, lda, eta = createindparams(param)
mu, mw = getutil(eta, ugroupmean, ugroupdist, delta.values)



#%%     THEANO EXPRESSIONS

print("Setting up Theano functions")

# Variables
tutil = T.matrix('util') #Matrix of individual utility
tm = T.vector('m') #Vector of profit margins
town = T.matrix('owner') #Ownership dummies
ttheta = T.vector('theta') #Vector of supply shading
tulo = T.vector('u_l') #Vector of shadow prices for lower limit of theta (0)
tuhi = T.vector('u_h') #Vector of shadow prices for upper limit of theta (1)
thi = T.vector('high') #Vector of highest margin good
trowsel = T.matrix('rowsel') #Matrix for selecting rows where theta can be set

#Vector of weights for utility deviations
tw = T.constant(mw, name='w') 

# theta, lda and hi as column vectors
tthetacol = ttheta.dimshuffle([0, 'x'])
tldacol = T.constant(lda[:, None], name='lambda')#tlda.dimshuffle([0,'x'])
thicol = thi.dimshuffle([0, 'x'])

# theta and lambda expanded to product (as column)
tthetaprod = town.T.dot(tthetacol)
tldaprod = town.T.dot(tldacol)

#Utility per good, per simulation
#Exponentiated utility
tmu = tutil / tldaprod
texpu = T.exp(tmu)
#Within chain sum of exp utility
tsumexpu = town.dot(texpu)
#Individual within-chain choice probability (raw, without distortion)
tcondcpnodist = texpu / town.T.dot(tsumexpu)
#Conditional choice probability, given chain choice
tcondcp = tthetaprod * tcondcpnodist + thicol * (1 - tthetaprod)

# Inclusive values
tIVb = T.log(tsumexpu)
tIVhi = (town * thi).dot(tutil)
# Exponentiated inclusive value
texpI = T.exp(tldacol * tthetacol * tIVb + (1 - tthetacol) * tIVhi)
# Chain choice probability
tchaincp = texpI / texpI.sum(0)
# Choice probability for chain-product combination, per node
tsimchoiceprob = tcondcp * town.T.dot(tchaincp)
# Choice probability for chain-product combination, aggregate
tchoiceprob = tsimchoiceprob.dot(tw)

#Jacobian of choice probability wrt theta
tJtchoiceprob = T.jacobian(tchoiceprob, ttheta)

# Lagrangian of Profit
tlagprofit = town.dot(tm * tchoiceprob) + tulo * ttheta + tuhi * (1 - ttheta)
# Derivative of profit lagrangian with respect to theta, only own-theta derivatives
tfoc, _ = theano.scan(fn=lambda i: T.grad(tlagprofit[i], ttheta)[i],
                      sequences=[T.arange(ttheta.shape[0])])

# Jacobian of FOC wrt input theta and margin
tJtmfoc, _ = theano.scan(
        fn=lambda i: T.grad(T.grad(tlagprofit[i], ttheta)[i], [ttheta, tm]),
        sequences=[T.arange(ttheta.shape[0])]
        )

tJwtheta = tsolve(trowsel.dot(tJtmfoc[0]).dot(trowsel.T), trowsel.dot(tJtmfoc[1]))

tJwcp = tJtchoiceprob.dot(trowsel.T).dot(tJwtheta)



###     BLP-type system of equations
#BLP choice probabilities
#Individual within-chain choice probability
tblpcondcp = texpu / town.T.dot(tsumexpu)
#Exponentiated inclusive values (per nest/chain)
tblpexpI = T.exp(tldacol * T.log(tsumexpu))
#Chain choice probability
tblpchaincp = tblpexpI / tblpexpI.sum(0)
#Product choice probability
tblpsimchoiceprob = tblpcondcp * town.T.dot(tblpchaincp)
#Aggregate choice probability/market share
tblpchoiceprob = tblpsimchoiceprob.dot(tw)


### Compile functions
mchoiceprob = theano.function([ttheta,tutil,thi,town], tchoiceprob,
                              name='choiceprob')
mblpchoiceprob = theano.function([tutil, town], tblpchoiceprob,
                                 name = 'blpchoiceprob')
mfoc = theano.function([ttheta,tulo,tuhi,tutil,thi,town,tm], tfoc, name = 'foc')
mJtfoc = theano.function([ttheta,tulo,tuhi,tutil,thi,town,tm], tJtmfoc[0], name = 'jacfoc')
mJwcp = theano.function([ttheta,tulo,tuhi,tutil,thi,town,tm,trowsel], tJwcp, name = 'jacwcp')



#%%    SETUP FUNCTIONS

print("Setting up Python functions")

def partargs(arg):
    theta = arg[:nchain]
    ulo = arg[nchain: 2 * nchain]
    uhi = arg[2 * nchain: 3 * nchain]
    return theta, ulo, uhi
    
def createargs(arg):
    theta = arg[:nchain]
    ulo = arg[nchain: 2 * nchain]
    uhi = arg[2 * nchain: 3 * nchain]
    return theta, ulo, uhi

def fixtheta(theta, ulo, uhi, thetafix):
    for i, t in thetafix:
        theta[i] = t
        ulo[i] = 0
        uhi[i] = 0
    return theta, ulo, uhi

def cobj(arg):
    theta, ulo, uhi = partargs(arg)
    return array([(ulo * theta + uhi * (1 - theta)).sum()])

def gcobj(arg):
    theta, ulo, uhi = partargs(arg)
    return hstack((ulo - uhi, theta, 1 - theta))

def csyseq(arg, mmu, maxmargin, own, margin):
    theta, ulo, uhi = createargs(arg)
    sys = mfoc(theta, ulo, uhi, mmu, maxmargin, own, margin)
    return sys

def Jcsyseq(arg, mmu, maxmargin, own, margin):
    theta, ulo, uhi = createargs(arg)
    tdjac = mJtfoc(theta, ulo, uhi, mmu, maxmargin, own, margin)
    tdjac = tdjac
    ujac = column_stack([eye(nchain), - eye(nchain)])
    return column_stack([tdjac, ujac])

def createguess(targs=None):
    if targs is None:
        return None
    else:
        ntargs = targs.sum()
        return hstack((ones(ntargs), zeros(ntargs), zeros(ntargs)))

def checksoc(theta, ulo, uhi, mmu, maxmargin, own, margin):
    """
    Verify that the second order condition for optimal theta holds for interior theta.
    """
    thetainterior = ~np.isclose(theta, 1) & ~np.isclose(theta, 0)
    soc = mJtfoc(theta, ulo, uhi, mmu, maxmargin, own, margin)
    return all(soc.diagonal()[thetainterior] < 0), soc

def solvemarketeq(mutil, maxmargin, own, margin,
                  guess=None, jac=True, fullret=False, retry=True,
                  singleprod=None):
    if guess is None:
        guess = hstack((ones(nchain), zeros(nchain), zeros(nchain)))
    # Bounds: 0-1 for theta, non-negative for multipliers
    bounds = [(0, 1)] * nchain
    bounds.extend([(0, None)] * 2 * nchain)
    # Constraints
    consdict = {'type': 'eq',
                'fun': lambda z: csyseq(z, mutil, maxmargin, own, margin)}
    if jac:
        consdict['jac'] = lambda z: Jcsyseq(z, mutil, maxmargin, own, margin)
    constraints = [consdict]
    # Add constraint for theta = 1 if single-product chain
    if singleprod is not None:
        cons = {'type': 'eq', 'fun': lambda z: z[singleprod] - 1}
        if jac:
            nsingleprod = len(singleprod)
            def cjac(z):
                jac = np.zeros((nsingleprod, len(z)))
                jac[np.arange(nsingleprod), singleprod] = 1
                return jac
            cons['jac'] = cjac
        constraints.append(cons)
    #Define objective and jacobian
    objective = lambda z: cobj(z)
    gobjective = lambda z: gcobj(z)
    optdict = {'maxiter': 200, 'ftol': 1e-16}
    start = dtim()
    res = minimize(objective, guess, jac=gobjective,
                   bounds=bounds, constraints=constraints, options=optdict)
    
    retried = 0
    if retry and (not res.success or (res.fun > 1e-10)):
        for i in range(150):
            guess = hstack((rand(nchain), rand(nchain), rand(nchain)))
            res = minimize(objective, guess, jac=gobjective,
                           bounds=bounds, constraints=constraints, options=optdict)
            retried += 1
            if res.success & (res.fun < 1e-10):
                break
    
    duration = dtim() - start
    ret = dict(zip(['theta', 'ulo', 'uhi'], createargs(res.x)))
    if fullret:
        ret['optres'] = res
        ret['duration'] = duration
        ret['retry'] = retried
    return ret

def rmarketsolve(market, rid, mu, fullret=False, retry=True,
                 jac=True, guess=None):
    """Solve market equilibrium with product ID rid removed.
    Returns optimization results and possibly diagnostics (with fullret = True)
    Also returns market shares with the new equilibrium."""
    mview = marketdata.loc[market].drop(rid, level='choiceid', errors='ignore')
    nchainprod = mview.groupby('chainid').share.count().values
    margin = mview.margin.fillna(0)
    maxmargin = margin.groupby(level='chainid').transform(lambda x: x == x.max())
    margin = margin.values
    # Fix potential removal of highest margin good
    maxmargin[maxmargin.groupby(level='chainid').transform('sum') == 0] = 1
    maxmargin = maxmargin.values
    own = mview[chaincols].values.T
    
    # Bool for arguments to be optimized over
    targs = nchainprod == 2
    
    # Indices for theta of single-product chains
    if all(targs):
        singleprod = None
    else:
        singleprod = (nchainprod == 1).nonzero()[0]
    
    mutil = mu.loc[market].drop(rid, level='choiceid', errors='ignore').values
    
    ret = solvemarketeq(mutil, maxmargin, own, margin,
                        guess=guess, jac=jac, fullret=fullret,
                        retry=retry, singleprod=singleprod)
    
    share = mchoiceprob(ret['theta'], mutil, maxmargin, own)
    
    soc = checksoc(ret['theta'], ret['ulo'], ret['uhi'], mutil, maxmargin,
                   own, margin)
    return ret, share, soc

def removeproductequil(rid, mu):
    """Remove one product (choice ID) and recalculate equilibrium in all markets"""
    ntheta = []
    ulo = []
    uhi = []
    solvestatus= []
    funval = []
    duration = []
    retries = []
    reoptimize = []
    soc = []
    altshare = pd.Series(0, index=marketdata.index)
    retdictkeys = ['theta', 'ulo', 'uhi', 'duration', 'retry']
    for m in marketlist:
        mix = marketdata.loc[[m]].index
        if rid in mix.get_level_values('choiceid'):
            rmix = marketdata.loc[[m]].drop(rid, level='choiceid').index
            nchainprod = marketdata.loc[m].drop(rid, level='choiceid')[chaincols].sum().values
            #rchains = nchainprod.index.values
            if m in marketlist and any(nchainprod > 1):
                ret, maltshare, msoc = rmarketsolve(m, rid, mu, fullret=True)
            else:
                nrchain = (nchainprod > 0).sum()
                ret = dict(zip(retdictkeys,
                           [ones(nrchain), zeros(nrchain), zeros(nrchain), 0, 0]))
                #Products per chain, except chain 0
                nchainprod = (marketdata.loc[m].drop(rid, level = 'choiceid').
                              groupby('chainid').share.count()
                              )
                rchaincols = ['chain%s' % i for i in nchainprod.index]
                own = marketdata.loc[m].drop(rid, level = 'choiceid')[rchaincols].values.T
                maltshare = mblpchoiceprob(mu.loc[m].drop(rid, level='choiceid'), own)
            maltshare = pd.Series(maltshare, index = rmix)
            altshare.update(maltshare)
            reoptimize.append(True)
        else:
            ret = dict(zip(retdictkeys,
                           [theta.theta.loc[m],
                            theta.ulo.loc[m],
                            theta.uhi.loc[m],
                            0, 0]))
            altshare.update(marketdata.loc[[m]].share)
            reoptimize.append(False)
        ntheta.append(ret['theta'])
        ulo.append(ret['ulo'])
        uhi.append(ret['uhi'])
        soc.append(msoc)
        duration.append(ret['duration'])
        retries.append(ret['retry'])
        if 'optres' in ret:
            solvestatus.append(ret['optres'].success)
            funval.append(np.abs(ret['optres'].fun).max())
        else:
            solvestatus.append(True)
            funval.append(0)
        
    altshare.name = 'share'
    #Index for theta with product removed (market-firm)
    altthetaix = marketdata.drop(rid, level='choiceid').groupby(['market', 'chainid']).share.sum().index
    alttheta = pd.DataFrame({'theta': hstack(ntheta), 'ulo': hstack(ulo), 'uhi': hstack(uhi)}, index=altthetaix)
    diagnostic = pd.DataFrame({'status': solvestatus, 'duration': duration,
                               'funval': funval, 'soc': soc,
                               'retry': retries, 'reoptimize': reoptimize}, index=marketix)
    return altshare, alttheta, diagnostic

def getoutsideshares(mu, rids=None):
    """
    Calculate market shares when removing a single product.
    
    Parameters
    ----------
    mu : pandas.DataFrame, (Nprodnor, Nlatent x Nsim)
        Deviations from mean market utility of each product in market over
        latent groups and simulations.
    rids : list_like, optional
        Choice ID's to remove for calculation of market share differences.
        The default is None. If None, use all products active in market.
    
    Returns
    -------
    outsideshare : pandas.Series, (Nrid x Nmarket x Nprodnor,)
        Series of shares under disagreement for each product given by "rid" for
        each market and product.
    outsidestheta : pandas.Series, (Nrid x Nmarket x Nchain,)
        Series of theta under removal of each product given by "rid" for
        each market and chain.
    outsidediagnostic : pandas.DataFrame, (Nrid x Nmarket, Ndiagnostics)
        DataFrame of diagnostics on the numerical solution for each product
        removal and market combination.
    """
    if rids is None:
        rids = choiceidlist
    outsideshare = pd.DataFrame()
    outsidetheta = pd.DataFrame()
    outsidediagnostic = pd.DataFrame()
    for rid in rids:
        altshare, alttheta, diagnostic = removeproductequil(rid, mu)
        altshare = altshare.reset_index()
        altshare['rid'] = rid
        alttheta = alttheta.reset_index()
        alttheta['rid'] = rid
        diagnostic = diagnostic.reset_index()
        diagnostic['rid'] = rid
        outsideshare = outsideshare.append(altshare, ignore_index=True)
        outsidetheta = outsidetheta.append(alttheta, ignore_index=True)
        outsidediagnostic = outsidediagnostic.append(diagnostic, ignore_index=True)
    outsideshare.set_index(['rid', 'market', 'chainid', 'choiceid'], inplace=True)
    outsidetheta.set_index(['rid', 'market', 'chainid'], inplace=True)
    outsidediagnostic.set_index(['rid', 'market'], inplace=True)
    return outsideshare, outsidetheta, outsidediagnostic

def getnondistshare(mu, dropdict=None):
    """
    Calculate market shares in the absence of market distortions downstream
    Allows exit through specification of market-specific choice ID's to drop in dropdict
    """
    if dropdict is None:
        dropdict = {}
    share = pd.Series(0, index=marketdata.index, name='share')
    for m in marketlist:
        drop = dropdict.get(m)
        if drop is not None:
            mix = marketdata.loc[[m]].drop(drop, level='choiceid').index
            own = marketdata.loc[m].drop(drop, level='choiceid')[chaincols].values.T
            mshare = mblpchoiceprob(mu.loc[m].drop(drop, level='choiceid').values, own)
        else:
            mix = marketdata.loc[[m]].index
            own = marketdata.loc[m][chaincols].values.T
            mshare = mblpchoiceprob(mu.loc[m].values, own)
        mshare = pd.Series(mshare, index=mix)
        share.update(mshare)
    return share

def getnoparshare(mu):
    """
    Calculate market shares in the absence of parallel imports
    """
    share = pd.Series(0, index=marketdata.index, name='share')
    # Choice ID for direct imports
    dichoiceid = marketdata.query('parallel == 0').reset_index().choiceid.unique()
    for m in marketlist:
        mix = marketdata.loc[[m]].query(f'choiceid in {*dichoiceid,}').index
        own = marketdata.loc[m].query(f'choiceid in {*dichoiceid,}')[chaincols].values.T
        mmu = mu.loc[m].query(f'choiceid in {*dichoiceid,}').values
        mshare = mblpchoiceprob(mmu, own)
        mshare = pd.Series(mshare, index=mix)
        share.update(mshare)
    return share

def getnondistoutsideshares(mu, rids=None):
    """
    Calculate full set of market shares in the absence of market distortions downstream with product removal
    rids - list of choice ID's to remove (optional). If none provided, will iterate through all choices
    """
    if rids is None:
        rids = choiceidlist
    outsideshare = pd.DataFrame()
    for rid in rids:
        rshare = pd.Series(0, index=marketdata.index, name='share')
        rmu = mu.drop(rid, level='choiceid')
        for m in marketlist:
            rmix = marketdata.loc[[m]].drop(rid, level='choiceid').index
            own = marketdata.loc[m].drop(rid, level='choiceid')[chaincols].values.T
            mshare = mblpchoiceprob(rmu.loc[m].values, own)
            mshare = pd.Series(mshare, index = rmix)
            rshare.update(mshare)
        rshare = rshare.reset_index()
        rshare['rid'] = rid
        outsideshare = outsideshare.append(rshare, ignore_index=True)
    outsideshare.set_index(['rid', 'market', 'chainid', 'choiceid'], inplace = True)
    return outsideshare

def marketproftdiff(market, sharediff):
    """
    Generate objects pertaining to additional profits from agreement for a given market
    
    Parameters
    ----------
    market : int
        Market ID.
    sharediff : pandas.DataFrame
        Changes in shares from agreement across bargaining situations.
    
    Returns
    -------
    chaindiff : pandas.DataFrame (Nprodnor, Nprodnor)
        Changes in shares from agreement relevant for chain profit.
    firmdiff : pandas.DataFrame (Nprodnor, Nprodnor)
        Changes in shares from agreement relevant for upstream firms revenue.
    parcpdiff : pandas.DataFrame (Nprodnor, Nprodnor)
        Changes in shares from agreement relevant for opportunity cost of parallel import.
    """
    #Relevant choice ID's and parallel trade indicators for market
    mchoice = choiceidlist
    parallel = marketdata.loc[market].parallel.values
    notboth = (marketdata.loc[market].parallel.groupby(level = 'chainid').transform('max') == 0).values
    #Ownership matrices for chains and upstream
    cown = marketdata.loc[market][chaincols].values.T
    cownmat = cown.T.dot(cown)
    uown = marketdata.loc[market][firmcols].values.T
    uownmat = uown.T.dot(uown)
    
    #Make a matrix of market share differences - choice ID in rows and disputed choice ID's in columns
    msdiff = sharediff.loc[market][mchoice].copy()
    #Chain share difference
    chaindiff = msdiff.values.T * cownmat
    #Firm share difference
    firmdiff = msdiff.values.T * uownmat
    #Get active upstream firms in market
    mfirm = marketdata.loc[market].firm.values
    #Create difference in shares for parallel imported product from agreement by upstream firm
    parchoice = mchoice[parallel == 1]
    msdiff = msdiff.reset_index('chainid', drop = True)
    parcpdiff = msdiff.loc[parchoice][[2, 4, 6]]
    parcpdiff.index = parcpdiff.index - 1
    parcpdiff = parcpdiff.append(msdiff.loc[mchoice[notboth]][[2,4,6]])
    for f in range(1, 4):
        fchoice = mchoice[mfirm == f]
        parcpdiff = parcpdiff.append(-msdiff.loc[fchoice][fchoice])
    return chaindiff, firmdiff, parcpdiff.sort_index().fillna(0).values.T

def getshareinputpricejac(market, mu, theta):
    """
    Calculate and return jacobian of market shares wrt input prices/upstream price
    """
    mutil = mu.loc[market].values
    hi = marketdata.loc[market].ismaxoption.values
    margin = marketdata.loc[market].margin.fillna(0).values
    own = marketdata.loc[market][chaincols].values.T
    
    mtheta = theta.theta.loc[market].values
    ulo = theta.ulo.loc[market].values
    uhi = theta.uhi.loc[market].values
    
    targs = ~np.isclose(mtheta, 1, 1e-10) & np.isclose(uhi, 0, 1e-10)
    if sum(targs) == 0:
        nchoice = len(margin)
        return zeros((nchoice, nchoice))
    else:
        ntargs = len(targs)
        rowsel = array([[1 * (i == j) for j in range(ntargs)] for i in range(ntargs) if targs[i]])
        return mJwcp(mtheta, ulo, uhi, mutil, hi, own, margin, rowsel)

def marketjac(market, mu, theta):
    """
    Generate objects pertaining to jacobian of shares wrt input prices for a given market
    
    Parameters
    ----------
    market : int
        Market ID.
    mu : pandas.DataFrame (Nprodnor x Nmarket, Nlatent x Nsim)
        Deviations from mean market utility of each product and market over
        latent groups and simulations.
    theta : pandas.Series (Nprodnor x Nmarket,)
        Series of theta (from demand estimation) for each chain and market.
    
    Returns
    -------
    chainjac : pandas.DataFrame (Nprodnor, Nprodnor)
        Derivatives of shares wrt to wholesale price relevant for each chain.
    firmjac : pandas.DataFrame (Nprodnor, Nprodnor)
        Derivatives of shares wrt wholesale price relevant for each upstream firm.
    parcpjac : pandas.DataFrame (Nprodnor, Nprodnor)
        Derivatives of shares wrt wholesale price relevant for opportunity cost
        of parallel trade in each bargaining situation.
    """
    #Get jacobian of market shares wrt input prices
    mjac = getshareinputpricejac(market, mu, theta)
    
    #Ownership matrices for chains and upstream
    cown = marketdata.loc[market][chaincols].values.T
    cownmat = cown.T.dot(cown)
    uown = marketdata.loc[market][firmcols].values.T
    uownmat = uown.T.dot(uown)
    #Matrix of share derivatives relevant for chain
    chainjac = mjac.T * cownmat
    #Matrix of share derivatives relevant for upstream firm
    firmjac = mjac.T * uownmat
    #Create derivative for parallel imported product by upstream firm
    mchoice = choiceidlist
    parallel = marketdata.loc[market].parallel.values
    parchoice = mchoice[parallel == 1]
    notboth = (marketdata.loc[market].parallel.groupby(level = 'chainid').transform('max') == 0).values
    mfirm = marketdata.loc[market].firm
    mjac = pd.DataFrame(mjac, index=pd.Index(mchoice, name='choiceid'),
                        columns=pd.Index(mchoice, name='rid'))
    parcpjac = mjac.loc[parchoice][[2, 4, 6]]
    parcpjac.index = parcpjac.index - 1
    parcpjac = parcpjac.append(mjac.loc[mchoice[notboth]][[2,4,6]])
    for f in range(1, 4):
        fchoice = mchoice[mfirm.values == f]
        parcpjac = parcpjac.append(-mjac.loc[fchoice][fchoice])
    return chainjac, firmjac, parcpjac.sort_index().fillna(0).values.T

def gathermarketdiffs(mu, theta, sharediff):
    """
    Gather calculated differences in shares from agreement and jacobians wrt
    wholesale prices for each market in dictionaries.
    """
    diffobj = {}
    jacobj = {}    
    for m in marketlist:
        diffobj[m] = marketproftdiff(m, sharediff)
        jacobj[m] = marketjac(m, mu, theta)
    return diffobj, jacobj

def solvemarketcost(bargainparam, market, diffobj, jacobj, jac=True):
    """
    Solve for upstream opportunity cost of parallel trade implied by
    Nash-in-Nash bargaining for a given market. Function of bargaining
    parameters, taking observed shares, retail prices, wholesale prices 
    and (estimated) demand model as given.
    """
    
    b = bwsel[market] @ bargainparam
    # Observed retail price, wholesale price and market shares
    retail = marketdata.loc[market].maxprice.values
    wholesale = marketdata.loc[market].gip.values
    share = marketdata.loc[market].share.values
    
    # Get change in shares from agreement 
    dchain, dfirm, duscost = diffobj[market]
    # Get jacobian of shares wrt wholesale price
    jchain, jfirm, juscost = jacobj[market]
    
    #Differences in profit
    dchainprofit = dchain.dot(retail - wholesale)
    dfirmrev = dfirm.dot(wholesale)
    
    #Jacobians of profit
    jchainprofit = -share + jchain.dot(retail - wholesale)
    jfirmrev = share + jfirm.dot(wholesale)
    
    jcobj = (juscost * (b * dchainprofit)[:, None] 
             + duscost * ((1 - b) * jchainprofit)[:, None]
             )
    # Vector not depending on upstream opportunity cost
    v = b * jfirmrev * dchainprofit + (1 - b) * dfirmrev * jchainprofit
    # pinv more stable across optimizer guesse for bargaining weights
    costpar = -np.linalg.pinv(jcobj) @ v
    
    #Difference and jacobians of upstream cost and profit
    dfirmcost = duscost.dot(costpar)
    dfirmprofit = dfirmrev + dfirmcost
    jfirmcost = juscost.dot(costpar)
    jfirmprofit = jfirmrev + jfirmcost
    
    if jac:
        jbobj = diag(jfirmprofit * dchainprofit - dfirmprofit * jchainprofit)
        jbuscost = (-np.linalg.pinv(jcobj) @ jbobj @ bwsel[market]).T
        ret = costpar, jbuscost
    else:
        ret = costpar
    return ret

def solvecost(bargainparam, diffobj, jacobj, asseries=False):
    """
    Solve for upstream opportunity cost of parallel trade across all markets.
    """
    temp = [solvemarketcost(bargainparam, m, diffobj, jacobj) 
            for m in marketlist]
    if asseries:
        ret = pd.Series(hstack([t[0] for t in temp]),
                        index=marketdata.loc[marketlist].index,
                        name='parprice')
    else:
        ret = hstack([t[0] for t in temp]), hstack([t[1] for t in temp])
    return ret

def nondistmarketdiff(market, sharediff, drop=None):
    """
    Generate objects pertaining to additional profits from agreement for a
    given market in counterfactual without share distortions.
    Allows dropping choice IDs when used for situations inducing exit.
    
    Parameters
    ----------
    market : int
        ID for market
    sharediff : pandas.DataFrame, (Nprodnor, Nprodnor)
        Differences in share for each product in row when product in column
        is available (bargaining agreement).
    drop : array_like, optional
        Product IDs to drop from Norwegian market. The default is None.
    
    Returns
    -------
    chaindiff : ndarray, (Nprodnor, Nprodnor)
        Matrix of share changes relevant for chain costs for each relevant
        bargaining scenario.
    upstreamdiff : ndarray, (Nprodnor, Nprodnor)
        Matrix of share changes relevant for upstream firm revenue for each
        relevant bargaining scenario.
    parcpdiff : ndarray, (Nprodnor, Nprodnor)
        Matrix of share changes relevant for upstream firm opportunity cost 
        for each relevant bargaining scenario.
    """
    
    if drop is None:
        drop = []
    #Relevant choice ID's and parallel trade indicators for market
    mchoice = array([x for x in choiceidlist if x not in drop])
    
    #Simplify by fetching relevant market subsample
    marketview = marketdata.loc[market].drop(drop, level='choiceid', errors='ignore')
    
    #Ownership matrices for chains and upstream
    cown = marketview[chaincols].values.T
    cownmat = cown.T.dot(cown)
    uown = marketview[firmcols].values.T
    uownmat = uown.T.dot(uown)
    
    #Make a matrix of market share differences - choice ID in rows and disputed choice ID's in columns
    msdiff = sharediff.copy()
    #Matrix of share changes relevant for chain
    chaindiff = msdiff.values.T * cownmat
    #Matrix of share changes relevant for upstream firm
    upstreamdiff = msdiff.values.T * uownmat
    
    #Create difference in shares for parallel imported product from agreement by upstream firm
    #Get active upstream firms in market
    mfirm = marketview.firm.values
    parallel = marketview.parallel.values
    parchoice = mchoice[parallel == 1]
    notboth = (marketview.parallel.groupby(level = 'chainid').transform('max') == 0).values
    
    msdiff = msdiff.reset_index('chainid', drop=True)    
    parcpdiff = msdiff.loc[parchoice][[2, 4, 6]]
    parcpdiff.index = parcpdiff.index - 1
    parcpdiff = parcpdiff.append(msdiff.loc[mchoice[notboth]][[2,4,6]])
    
    for f in range(1, 4):
        fchoice = mchoice[mfirm == f]
        parcpdiff = parcpdiff.append(- msdiff.loc[fchoice][fchoice])
    return chaindiff, upstreamdiff, parcpdiff.sort_index().fillna(0).values.T

def solvenondistwholesale(market, sharediff, parprice, bargainparam,
                          drop=None, marketprice=None):
    """
    Solve for wholesale prices (gip) implied by zero-equations/first order 
    conditions for bargaining under no distortion in Norway.

    Parameters
    ----------
    market : int
        Market ID.
    sharediff : pandas.DataFrame, (Nprodnor, Nprodnor)
        Differences in share for each product in row when product in column
        is available (bargaining agreement).
    parprice : ndarray, (Nprodnor,)
        Upstream opportunity cost of parallel trade (NOK/DDD) in Norway.
    bargainparam : ndarray, (Nbargainpair,)
        Bargaining power of upstream firm for each potential bargaining pair
        in Norway.
    drop : list_like, optional
        Product IDs to drop from Norwegian market. The default is None.
    marketprice : ndarray, (Nprodnor), optional
        Pharmacy earnings (NOK/DDD) for each product. The default is None.
        If None, set to observed maxprice for market in data.

    Returns
    -------
    gip : ndarray, (Nprodnor,)
        Wholesale prices (NOK/DDD) that satisfy the bargaining equilibrium conditions.
    """
    # Relative weights on pharmacy chain profit
    b = bargainparam / (1 - bargainparam)
    if drop is None:
        drop = []
    #Matrix for selecting relevant bargaining parameters
    sel = [i for i in range(6) if i not in [x - 2 for x in drop]]
    mb = (bwsel[market] @ b)[sel]
    
    chaindiff, upstreamdiff, parcpdiff = nondistmarketdiff(market, sharediff,
                                                           drop=drop)
    
    if marketprice is None:
        marketprice = (marketdata.loc[market, 'maxprice']
                       .drop(drop, level='choiceid', errors='ignore').values
                       )
    
    A = mb[:, None] * chaindiff + upstreamdiff
    b = mb * (chaindiff @ marketprice) - parcpdiff @ parprice
    return solve(A, b)

def updateparprice(market, parprice, lipitorprice, drop=None):
    """
    Calculate new vector of upstream opportunity cost of parallel trade in
    the Norwegian market given French Lipitor price and old vector of
    opportunity cost. Updates elements where the chain-specific parallel source
    is France.

    Parameters
    ----------
    market : int
        Market ID.
    parprice : ndarray, (Nprodnor,)
        Old upstream opportunity cost of parallel trade (NOK/DDD) in Norway.
    lipitorprice : float
        (New) price (USD/SU) of Lipitor in French market .
    drop : array_like, optional
        Product IDs to drop from Norwegian market. The default is None.

    Returns
    -------
    newparprice : ndarray, (Nprodnor,)
        Updated vector of upstream opportunity cost (NOK/DDD).
    """
    mview = marketdata.loc[market]
    if drop is not None:
        mview = mview.drop(drop, level='choiceid', errors='ignore')
    # Baseline price in France
    lipitorprice0 = marketinfo.loc[market, 'frlipitorprice']
    # Conversion to NOK/DDD
    pricescale = marketinfo.loc[market, 'lipitorpriceconvert']
    francesource = mview.francesource.groupby('chainid').transform('max').values
    newparprice = parprice.copy()
    newparprice[francesource] += (lipitorprice - lipitorprice0) * pricescale
    return newparprice
    
def intlbargaincond(market, frprice, parprice, mc,
                    frsharediff, norsharediff,
                    frbargainparam, norbargainparam, 
                    marketprice=None, drop=None):
    """
    Bargaining zero-equations for France accounting for changes in the Norwegian
    market equilibrium. The zero equations are derivatives of the bargaining
    objective for each of `Nmol` Statin molecules with respect to price, given
    simultaneous changes in the upstream bargaining equilibrium in Norway.

    Parameters
    ----------
    market : int
        Market ID.
    frprice : ndarray, (Nmol,)
        (Wholesale) price (USD/SU) of originator drug within each molecule in
        France.
    parprice : ndarray, (Nprodnor,)
        Old upstream opportunity cost of parallel trade (NOK/DDD) in Norway.
    mc : ndarray, (Nmol,)
        Marginal cost (USD/SU) of originator drug within each molecule in France.
    frsharediff : ndarray, (Nprodnor,)
        Difference in shares in Norwegian market from agreement in France.
    norsharediff : pandas.DataFrame, (Nprodnor, Nprodnor)
        Differences in share in Norwegian market for each product in row when
        product in column is available (bargaining agreement).
    frbargainparam : float
        (Nash) bargaining power of producer vis-a-vis French regulator. 
        The default is 0.5.
    norbargainparam : ndarray, (Nbargainpair,)
        Bargaining power of upstream firm for each potential bargaining pair
        in Norway.
    marketprice : ndarray, (Nprodnor), optional
        Pharmacy earnings (NOK/DDD) for each product. The default is None.
        If None, set to observed maxprice for market in data.
    drop : array_like, optional
        Product IDs to drop from Norwegian market. The default is None.

    Returns
    -------
    cond : ndarray, (Nmol,)
        Values of the derivative of the bargaining objective for each molecule
        in the French market.
    """
    mview = marketdata.loc[market]
    if drop is not None:
        mview = mview.drop(drop, level='choiceid', errors='ignore')
    parallel = mview.parallel.values == 1
    parchain = mview.parallel.groupby('chainid').transform('max').values == 1
    nonparmask = (parallel == False) & parchain
    lipitorprice = frprice[0] # Price of Lipitor first element
    newparprice = updateparprice(market, parprice, lipitorprice, drop=drop)
    gip = solvenondistwholesale(market, norsharediff, newparprice, norbargainparam,
                                drop=drop, marketprice=marketprice)
    producerprice = gip.copy()
    producerprice[parallel] = newparprice[nonparmask]
    return francebargaincond(market, producerprice, frprice, mc,
                             frsharediff, frbargainparam, drop=drop)

def solveintlbargain(market, parprice, mc, frsharediff, norsharediff,
                     frbargainparam, norbargainparam,
                     marketprice=None, drop=None,
                     guess=None, fullret=False, retry=True, usejac=True, eps=1e-5):
    """
    Solve bargaining equilibrium conditions in French market with respect to
    (French) prices.

    Parameters
    ----------
    market : int
        Market ID.
    parprice : ndarray, (Nprodnor,)
        Old upstream opportunity cost of parallel trade (NOK/DDD) in Norway.
    mc : ndarray, (Nmol,)
        Marginal cost (USD/SU) of originator drug within each molecule in France.
    frsharediff : ndarray, (Nprodnor,)
        Difference in shares in Norwegian market from agreement in France.
    norsharediff : pandas.DataFrame, (Nprodnor, Nprodnor)
        Differences in share for each product in row when product in column
        is available (bargaining agreement).
    frbargainparam : float
        (Nash) bargaining power of producer vis-a-vis French regulator. 
        The default is 0.5.
    norbargainparam : ndarray, (Nbargainpair,)
        Bargaining power of upstream firm for each potential bargaining pair
        in Norway.
    marketprice : ndarray, (Nprodnor), optional
        Pharmacy earnings (NOK/DDD) for each product. The default is None.
        If None, set to observed maxprice for market in data.
    drop : array_like, optional
        Product IDs to drop from Norwegian market. The default is None.
    guess : ndarray, (Nmol,), optional
        Initial guess for root finind. The default is None. If None, set to
        a factor of `mc`.
    fullret : bool, optional
        Return full output of root finding and number of retries.
        The default is False.
    retry : bool, optional
        Retry with other guesses if root finding does not converge.
        The default is True.
    jac : bool, optional
        Use numeric jacobian for root finding. The default is True.
    eps : float, optional
        Step-size in calculating two-sided numeric derivatives. The default is 1e-5.

    Returns
    -------
    gip : ndarray, (Nprodnor,)
        Wholesale price (NOK/DDD) in the Norwegain market.
    frprice : ndarray, (Nmol,)
        Wholesale) price (USD/SU) of originator drug within each molecule in
        France.
    success : bool
        Whether root finding converged.
        
    or, if `fullret` is True:
    gip : ndarray, (Nprodnor,)
        Wholesale price (NOK/DDD) in the Norwegain market.
    solution : OptimizeResult
        The result of the root finding routine represented as a
        OptimizeResult object (see `scipy.optimize.root`).
    nretry : int
        Number of retries with changed guesses.
    """
    gfactor = 1.05
    if guess is None:
        #guess = mc * gfactor * np.sign(mc)
        guess = demand_france.query('bargainer').loc[market].price.values
    
    def zerofunc(x):
        return intlbargaincond(market, x, parprice, mc,
                               frsharediff, norsharediff,
                               frbargainparam, norbargainparam,
                               marketprice=marketprice, drop=drop)
    
    if usejac:
        def gzerofunc(x):
            nx = len(x)
            xhi = x[:, None] + eye(nx) * eps
            xlo = x[:, None] - eye(nx) * eps
            fhi = np.apply_along_axis(zerofunc, 0, xhi)
            flo = np.apply_along_axis(zerofunc, 0, xlo)
            return (fhi - flo) / (2 * eps)
        jac = gzerofunc
    else:
        jac = False
    for i in range(30):
        res = root(zerofunc, guess, jac=jac)
        if (not res.success) & retry:
            gfactor += 0.05
            guess = mc * gfactor * np.sign(mc)
        else:
            break
    lipitorprice = res.x[0]
    newparprice = updateparprice(market, parprice, lipitorprice, drop=drop)
    gip = solvenondistwholesale(market, norsharediff, newparprice, norbargainparam,
                                drop=drop, marketprice=marketprice)
    if fullret:
        return gip, res, i
    else:
        return gip, res.x, res.success

def intlequil(parprice, mc, frsharediff, norsharediff,
                 frbargainparam, norbargainparam):
    """
    Solve for the bargaining equilibrium in France and Norway across all markets.
    
    Parameters
    ----------
    parprice : ndarray, (Nmarket x Nprodnor,)
        Old upstream opportunity cost of parallel trade (NOK/DDD) in Norway.
    mc : ndarray, (Nmarket x Nmol,)
        Marginal cost (USD/SU) of originator drug within each molecule in France.
    frsharediff : pandas.Series, (Nmarket x Nprodnor,)
        Difference in shares in Norwegian market from agreement in France.
    norsharediff : pandas.DataFrame, (Nmarket x Nprodnor, Nprodnor)
        Differences in share for each product in row when product in column
        is available (bargaining agreement).
    frbargainparam : float
        (Nash) bargaining power of producer vis-a-vis French regulator. 
        The default is 0.5.
    norbargainparam : ndarray, (Nbargainpair,)
        Bargaining power of upstream firm for each potential bargaining pair
        in Norway.

    Returns
    -------
    gip : pandas.Series, (Nmarket x Nprodnor,)
        Wholesale price (NOK/DDD) in the Norwegain market.
    frprice : pandas.Series, (Nmarket x Nmol,)
        Wholesale) price (USD/SU) of originator drug within each molecule in
        France.
    success : list, (Nmarket,)
        Whether root finding converged in each market.
    """
    gip = []
    frprice = []
    success = []
    for m in marketlist:
        mgip, mfrprice, succ =\
            solveintlbargain(m, parprice.loc[m].values,
                             mc.loc[m].values,
                             frsharediff.loc[m].values,
                             norsharediff.loc[m],
                             frbargainparam, norbargainparam)
        gip.append(mgip)
        frprice.append(mfrprice)
        success.append(succ)
    gip = pd.Series(hstack(gip), index=marketdata.index)
    price = pd.Series(hstack(frprice), index=frmbargainix)
    return gip, price, success

def getridreductionmarketsharediff(market, mu, rids=None, drop=None):
    """
    Calculate market shares difference in the absence of market distortions 
    downstream if products were removed (disagreement outcome).
    Account for potential product exit (see `drop`).

    Parameters
    ----------
    market : int
        Market ID.
    mu : pandas.DataFrame, (Nprodnor, Nlatent x Nsim)
        Deviations from mean market utility of each product in market over
        latent groups and simulations.
    rids : list_like, optional
        Choice ID's to remove for calculation of market share differences.
        The default is None. If None, use all products active in market.
    drop : array_like, optional
        Product IDs to drop from Norwegian market. The default is None.

    Returns
    -------
    norsharediff : pandas.DataFrame, (Nmarket x Nprodnor, Nprodnor)
        Differences in share for each product in row when product in column
        is available (bargaining agreement).
    frsharediff : pandas.Series, (Nprodnor,)
        Difference in shares in Norwegian market from agreement in France.
    norbaseshare : pandas.Series, (Nprodnor,)
        Shares in Norway at baseline, accounting for exit.
    """
    
    if drop is None:
        drop = []
    fullmix = marketdata.loc[[market]].drop(drop, level='choiceid', errors='ignore').index
    mview = marketdata.loc[market].drop(drop, level='choiceid', errors='ignore')
    mix = mview.index
    norsharediff = pd.DataFrame(index=mix)
    if rids is None:
        rids = [x for x in choiceidlist if x not in drop]
    #Base market share without distortion, dropping ID's in drop
    own = mview[chaincols].values.T
    mmu = mu.drop(drop, level='choiceid', errors='ignore').values
    norbaseshare = mblpchoiceprob(mmu, own)
    norbaseshare = pd.Series(norbaseshare, index=mix)
    for rid in rids:
        rdrop = np.unique(hstack((rid, drop)))
        rmview = mview.drop(rid, level='choiceid')
        rmix = rmview.index
        own = rmview[chaincols].values.T
        mshare = mblpchoiceprob(mu.drop(rdrop, level='choiceid').values, own)
        mshare = pd.Series(mshare, index=rmix).reindex(mix).fillna(0)
        norsharediff[rid] = norbaseshare - mshare
    # If disagreement in France
    francesource = (mview.sourcecountry == 'France')
    franceid = mview.loc[francesource].reset_index().choiceid.values
    rdrop = np.unique(hstack((franceid, drop)))
    rmview = mview.drop(franceid, level='choiceid', errors='ignore')
    rmix = rmview.index
    own = rmview[chaincols].values.T
    mmu = mu.drop(rdrop, level='choiceid', errors='ignore').values
    mshare = mblpchoiceprob(mmu, own)
    mshare = pd.Series(mshare, index=rmix).reindex(mix).fillna(0)
    frsharediff = norbaseshare - mshare
    norbaseshare = pd.Series(norbaseshare.values, index=fullmix)
    return norsharediff, frsharediff, norbaseshare


def solvenondistexitmarket(market, mu, parprice, mc,
                           frbargainparam, norbargainparam,
                           price=None, drop=None):
    """
    Solves for wholesale prices in given market in the counterfactual scenario 
    of no supply distortion. Allows for exit of parallel traders.

    Parameters
    ----------
    market : int
        Market ID.
    mu : pandas.DataFrame, (Nprodnor, Nlatent x Nsim)
        Deviations from mean market utility of each product in market over
        latent groups and simulations.
    parprice : pandas.Series, (Nprodnor,)
        Old upstream opportunity cost of parallel trade (NOK/DDD) in Norway.
    mc : ndarray, (Nmol,)
        Marginal cost (USD/SU) of originator drug within each molecule in France.
    frbargainparam : float
        (Nash) bargaining power of producer vis-a-vis French regulator. 
        The default is 0.5.
    norbargainparam : ndarray, (Nbargainpair,)
        Bargaining power of upstream firm for each potential bargaining pair
        in Norway.
    price : pd.Series, (Nprodnor,), optional
        Pharmacy earnings (NOK/DDD) for each product. The default is None.
        If None, set to observed maxprice for market in data.
    drop : array_like, optional
        Product IDs to drop from Norwegian market. The default is None.

    Returns
    -------
    gip : pandas.Series, (Nprodnor,)
        Wholesale price (NOK/DDD) for `market` in Norway.
    norbaseshare : pandas.Series, (Nprodnor,)
        Shares in Norway at baseline, accounting for exit.
    frprice : ndarray, (Nmol,)
        Wholesale) price (USD/SU) of originator drug within each molecule in
        France.
    succ : bool
        Whether root finding converged.
    """
    if drop is None:
        drop = []
    for k in range(3):
        norsharediff, frsharediff, norbaseshare =\
            getridreductionmarketsharediff(market, mu, drop=drop)

        if price is not None:
            argprice = price.drop(drop, level='choiceid', errors='ignore').values
        else:
            argprice = None
        useparprice = parprice.drop(drop, level='choiceid', errors='ignore').values
        
        gip, frprice, succ = solveintlbargain(market, useparprice, mc,
                                              frsharediff.values, norsharediff,
                                              frbargainparam, norbargainparam,
                                              marketprice=argprice,
                                              drop=drop)
        
        mix = marketdata.loc[[market]].drop(drop, level='choiceid', errors='ignore').index
        gip = pd.Series(gip, mix)
        newparprice = updateparprice(market, useparprice, frprice[0], drop=drop)
        mnonpi = [x for x in [2, 4, 6] if x not in drop]
        parmargin = (gip.loc[market] - newparprice).drop(mnonpi, level='choiceid')
        if any(parmargin < 0):
            drop.append(parmargin.idxmin()[1])
        else:
            break
    return gip, norbaseshare, frprice, succ

def solvenondistexit(mu, parprice, mc, frbargainparam, norbargainparam,
                     price=None, drop=None):
    """
    Solves for all wholesale prices in the counterfactual scenario of
    no supply distortion.
    Solves each market by calling `solvenondistexitmarket`.
    
    Parameters
    ----------
    mu : pandas.DataFrame, (Nmarket x Nprodnor, Nlatent x Nsim)
        Deviations from mean market utility for all markets and products over
        latent groups and simulations.
    parprice : pandas.Series, (Nmarket x Nprodnor,)
        Old upstream opportunity cost of parallel trade (NOK/DDD) in Norway
        for all markets.
    mc : pandas.Series, (Nmarket x Nmol,)
        Marginal cost (USD/SU) of originator drug within each molecule in France
        for all markets.
    frbargainparam : float
        (Nash) bargaining power of producer vis-a-vis French regulator. 
        The default is 0.5.
    norbargainparam : ndarray, (Nbargainpair,)
        Bargaining power of upstream firm for each potential bargaining pair
        in Norway.
    price : pandas.Series, (Nmarket x Nprodnor,), optional
        Pharmacy earnings (NOK/DDD) for each product for all markets. The default is None.
        If None, set to observed market maxprices in data.
    drop : array_like, optional
        Product IDs to drop from Norwegian market. The default is None.
    
    Returns
    -------
    gip : pandas.Series, (Nmarket x Nprodnor,)
        Wholesale price (NOK/DDD) in the Norwegain market.
    norshare : pandas.Series, (Nmarket x Nprodnor,)
        Market share of products in Norwegian market,
        potentially zero due to exit.
    frprice : pandas.Series, (Nmarket x Nmol,)
        Wholesale) price (USD/SU) of originator drug within each molecule in
        France.
    success : list, (Nmarket,)
        Whether root finding converged in each market.
    """
    gip = pd.Series(0, index=marketdata.index)
    norshare = pd.Series(0, index=marketdata.index)
    frprice = []
    success = []
    for m in marketlist:
        if price is None:
            argprice = None
        else:
            argprice = price.loc[m]
        mgip, mshare, mfrprice, msucc =\
            solvenondistexitmarket(m, mu.loc[m], parprice.loc[m],
                                   mc.loc[m].values,
                                   frbargainparam, norbargainparam,
                                   price=argprice, drop=drop)
        gip.update(mgip)
        norshare.update(mshare)
        frprice.append(mfrprice)
        success.append(msucc)
    gip.name = 'gip'
    norshare.name = 'share'
    frprice = pd.Series(hstack(frprice), index=frmbargainix)
    return gip, norshare, frprice, success

def pricedictfrance(orgprice):
    """
    Create dictionary mapping molecule ID to originator price.

    Parameters
    ----------
    orgprice : ndarray
        (Wholesale) price (USD/SU) of originator drug within each molecule in France.
        Assumed that price vector follows the ordering of molecule IDs.

    Returns
    -------
    orgpricedict : dict
        Dictorinary mapping molecule ID and originator price (USD/SU).

    """
    norgprice = len(orgprice)
    return dict(zip(range(norgprice), orgprice))
    
def pricefrance(market, orgpricedict):
    """
    Return vector of market prices given prices of originators in each molecule.
    Corresponds to regulatory set prices for originator products marketed by
    other companies (same price as bargaining company) and generics (0.6
    of originator price).

    Parameters
    ----------
    market : int
        Market ID.
    orgpricedict : dict
        Dictorinary mapping molecule ID and originator price (USD/SU).
    
    Returns
    -------
    price : ndarray, (Nprod,)
        Prices (USD/SU) for all products in French market.
    """
    orgprice = demand_france.loc[market, 'moleculeid'].map(orgpricedict)
    generic = demand_france.loc[market, 'generic']
    return orgprice * (1 - fr_rgenprice * generic)

def utilfrance(market, price):
    """
    Calculate matrix of utilities for each product (`Nprod` rows) and each
    simulated individual/utility parameter vector (`Nsim` columns) in French
    market as a function of prices.

    Parameters
    ----------
    market : int
        Market ID.
    price : ndarray, (Nprod,)
        Prices (USD/SU) for all products in French market.

    Returns
    -------
    util : ndarray, (Nprod, Nsim)
        Matrix of simulated utilities for all products in market.

    """
    util = (demand_france.loc[market].grossutil.values[:, None] +
            price[:, None] * fr_pricepref)
    return util

def sharefrance(market, price, jac=True):
    """
    Calculate market shares in the French market.

    Parameters
    ----------
    market : int
        Market ID.
    price : ndarray, (Nprod,)
        Prices (USD/SU) for all products in French market.
    jac : bool, optional
        Also return jacobian of shares with respect to prices. The default is True.
    
    Returns
    -------
    share : ndarray, (Nprod,)
        Market share of each product.
    jacshare : ndarray, (Nprod, Nprod), optional
        Jacobian of market shares with respect to prices. Only if requested.
    """
    expu = exp(utilfrance(market, price))
    prob = expu / (1 + expu.sum(0))
    share = prob.mean(1)
    if jac:
        probpricepref = prob * fr_pricepref
        jacshare = diag((probpricepref).mean(1))  - probpricepref @ prob.T / fr_msim
        return share, jacshare
    else:
        return share

def bargainerprofitfrance(market, price, mc, nok=False):
    """
    Calculate profits of bargaining firms in France.

    Parameters
    ----------
    market : int
        Market ID.
    price : ndarray, (Nmol,)
        (Wholesale) price (USD/SU) of originator drug within each molecule in
        France.
    mc : ndarray, (Nmol,)
        Marginal cost (USD/SU) of originator drug within each molecule in France.
    nok : bool, optional
        Convert profits to NOK (default is USD). The default is False.

    Returns
    -------
    profit : ndarray, (Nmol,)
        Profit of originator drug within each molecule in France.
    """
    fullprice = pricefrance(market, pricedictfrance(price))
    bargainer = demand_france.loc[market, 'bargainer'].values
    share = sharefrance(market, fullprice, jac=False)[bargainer]
    msize = marketinfo.loc[market, 'frsize']
    currencyconvert = marketinfo.loc[market, 'nokusd'] if nok else 1
    profit = (price - mc) * share * msize * currencyconvert
    return profit

def welfarefrance(market, price, jac=True):
    """
    Compute change in consumer surplus in French Statin market from having
    the originator product available in each molecule given prices.

    Parameters
    ----------
    market : int
        Market ID.
    price : numpy.ndarray, (Nprod,)
        Vector of prices in USD/SU (SU=Standard Unit=Tablet).
    jac : bool, optional
        Also return the total derivative of consumer surplus. The default is True.

    Returns
    -------
    deltacs : ndarray, (Nmol,)
        Change in consumer surplus per market capita (market size in SU) when
        originator product is available in each of Nmol molecules at given
        prices.
    graddeltacs : ndarray, (Nmol,), optional
        Total derivative of each molecule-specific change in consumer surplus
        with respect to price of the originator product in that molecule.
    """
    data = demand_france.loc[market]
    expu = exp(utilfrance(market, price))
    dmol = data[dmolcols]
    dmol = dmol.loc[:, dmol.sum() > 0]
    bargaindrop = dmol.mul(1 - data.generic, axis=0).values.T
    sumexpu = 1 + expu.sum(0)
    sumexpu0 = sumexpu[None, :] - bargaindrop @ expu
    diffw = ((log(sumexpu) - log(sumexpu0)) / (- fr_pricepref)).mean(1)
    if jac:
        partdiffw = (expu / sumexpu).mean(1)
        diffrel = dmol.mul(1 - fr_rgenprice * data.generic, axis=0).values.T
        totdiffw = - diffrel @ partdiffw
        return diffw, totdiffw
    else:
        return diffw

def francepriceprep(market, gip, lipitorprice, parprice):
    """
    Calculate upstream opportunity cost and Pfizer's revenue per product in
    Norway (`Nprodnor` products) given the the price of Lipitor in France
    
    Parameters
    ----------
    market : int
        Market ID.
    gip : ndarray, (Nprodnor,)
        Wholesale price (NOK/DDD) of each product in Norway.
    lipitorprice : float
        (New) lipitor price (USD/SU) in France.
    parprice : ndarray, (Nprodnor,)
        Old upstream opportunity cost of parallel trade (NOK/DDD) in Norway.
    
    Returns
    -------
    newparprice : ndarray, (Nprodnor,)
        Updated upstream opportunity cost (NOK/DDD).
    producerprice : ndarray, (Nprodnor,)
        Pfizer's revenue (NOK/DDD) for each product in Norway.
    """
    mview = marketdata.loc[market]
    # Baseline price
    lipitorprice0 = marketinfo.loc[market, 'frlipitorprice']
    # Conversion to NOK/DDD
    pricescale = marketinfo.loc[market, 'lipitorpriceconvert']
    parallel = (mview.parallel == 1).values
    francesource = mview.francesource.groupby('chainid').transform('max').values
    newparprice = parprice.copy()
    newparprice[francesource] += (lipitorprice - lipitorprice0) * pricescale
    producerprice = gip.copy()
    producerprice[parallel] = newparprice[parallel == False]
    return newparprice, producerprice

def francebargaincond(market, producerprice, price, mc, frsharediff,
                      bargainparam=0.5, retparts=False, drop=None):
    """
    Bargaining zero-equations for France, i.e., derivatives of bargaining
    objective for each of `Nmol` Statin molecules with respect to price.
    
    Parameters
    ----------
    market : int
        Market ID.
    producerprice : ndarray, (Nprodnor,)
        Pfizer's revenue (NOK/DDD) for each product in Norway.
    price : ndarray, (Nmol,)
        (Wholesale) price (USD/SU) of originator drug within each molecule in
        France.
    mc : ndarray, (Nmol,)
        Marginal cost (USD/SU) of originator drug within each molecule in France.
    frsharediff : ndarray, (Nprodnor,)
        Difference in shares in Norwegian market from agreement in France.
    bargainparam : float
        (Nash) bargaining power of producer vis-a-vis French regulator. 
        The default is 0.5.
    retparts : bool, optional
        Return the bargaining surplus of each party along with their derivatives
        with respect to price (profit difference, derivative of profit difference,
        change in consumer surplus, and derivative of change in consumer surplus).
        The default is False.
    drop : array_like, optional
        Product IDs to drop from Norwegian market. The default is None.
    
    Returns
    -------
    cond : ndarray, (Nmol,)
        Values of the derivative of the bargaining objective for each molecule
        in the French market.
    """
    mview = marketdata.loc[market]
    if drop is not None:
        mview = mview.drop(drop, level='choiceid', errors='ignore')
    franceparallel = mview.francesource.values
    franceweight = marketinfo.loc[market, 'frweight']
    bargainer = demand_france.loc[market, 'bargainer'].values
    fullprice = pricefrance(market, pricedictfrance(price))
    share, jacshare = sharefrance(market, fullprice)
    bshare, gbshare = share[bargainer], jacshare[bargainer]
    dmol = demand_france.loc[market, dmolcols]
    dmol = dmol.loc[:, dmol.sum() > 0]
    diffrel = dmol.mul(1 - fr_rgenprice * demand_france.loc[market, 'generic'],
                       axis=0).values.T
    # Gain in profit from agreement
    diffPi = franceweight * (price - mc) * bshare
    diffPi[0] += producerprice @ frsharediff
    gdiffPi = franceweight * (bshare + (price - mc) * (gbshare * diffrel).sum(1))
    gdiffPi[0] += frsharediff[franceparallel].sum()
    # Consumer surplus
    diffw, gdiffw = welfarefrance(market, fullprice)
    if retparts:
        return diffPi, gdiffPi, diffw, gdiffw
    else:
        # Relative weight on producer profit
        b = bargainparam / (1 - bargainparam)
        return b * gdiffPi / diffPi + gdiffw / diffw

def solvefrancebargaincond(market, producerprice, mc, frsharediff,
                           bargainparam=0.5, guess=None, retry=True, jac=True, eps=1e-5):
    """
    Find a solution to the French bargaining zero-equations, given by the
    derivative of bargaining objective in each molecule with respect to price.

    Parameters
    ----------
    market : int
        Market ID.
    producerprice : ndarray, (Nprodnor,)
        Pfizer's revenue (NOK/DDD) for each product in Norway.
    mc : ndarray, (Nmol,)
        Marginal cost (USD/SU) of originator drug within each molecule in France.
    frsharediff : ndarray, (Nprodnor,)
        Difference in shares in Norwegian market from agreement in France.
    bargainparam : float
        (Nash) bargaining power of producer vis-a-vis French regulator. 
        The default is 0.5.
    guess : ndarray, (Nmol,), optional
        Guess for root finder. The default is None. If not provided, guess
        is set to :math:`1.1 \\times mc`
    retry : bool, optional
        Retry with other guesses if root finding does not converge.
        The default is True.
    jac : bool, optional
        Use numeric jacobian for root finding. The default is True.
    eps : float, optional
        Step-size in calculating two-sided numeric derivatives. The default is 1e-5.

    Returns
    -------
    solution : OptimizeResult
        The result of the root finding routine represented as a
        OptimizeResult object (see `scipy.optimize.root`).
    """
    def zerofunc(x):
        return francebargaincond(market, producerprice, x, mc, frsharediff,
                                 bargainparam)
    
    gfactor = 1.05
    if guess is None:
        #guess = mc * gfactor * np.sign(mc)
        guess = demand_france.query('bargainer').loc[market].price.values
    
    if jac:
        def gzerofunc(x):
            nx = len(x)
            xhi = x[:, None] + eye(nx) * eps
            xlo = x[:, None] - eye(nx) * eps
            fhi = np.apply_along_axis(zerofunc, 0, xhi)
            flo = np.apply_along_axis(zerofunc, 0, xlo)
            return (fhi - flo) / (2 * eps)
        jac = gzerofunc
    
    for i in range(30):
        solution = root(zerofunc, guess, jac=jac)
        if (not solution.success) & retry:
            gfactor += 0.025
            guess = mc * gfactor * np.sign(mc)
        else:
            break
    return solution

def equilfranceaut(mc, bargainparam=0.5, retry=True, jac=True, eps=1e-5):
    """
    Calculate price equilibria in France when Norwegian market doesn't matter
    (e.g., parallel trade ban).

    Parameters
    ----------
    mc : pandas.Series, (Nmarket x Nmol,)
        Marginal cost (USD/SU) of originator drug within each molecule for 
        each market in France. Indexed by market as first index.
    bargainparam : float
        (Nash) bargaining power of producer vis-a-vis French regulator. 
        The default is 0.5.
    retry : bool, optional
        Retry with other guesses if root finding does not converge.
        The default is True.
    jac : bool, optional
        Use numeric jacobian for root finding. The default is True.
    eps : float, optional
        Step-size in calculating two-sided numeric derivatives. The default is 1e-5.
    
    Returns
    -------
    price : pandas.Series, (Mnarket x Nmol,)
        (Wholesale) price (USD/SU) of originator drug within each molecule for
        each market in France.
    success : list (bool)
        Indicator for each market whether root finding terminated in a
        solution to the system of equations.
    """
    price = []
    success = []
    prodprice = zeros(6)
    frsharediff = zeros(6)
    for m in marketlist:
        res = solvefrancebargaincond(m, prodprice, mc.loc[m].values,
                                     frsharediff, bargainparam,
                                     retry=retry, jac=jac, eps=eps)
        price.append(res.x)
        success.append(res.success)
    return pd.Series(hstack(price), index=frmbargainix), success

def francemcest(parprice, frsharediff, bargainparam=0.5):
    """
    Solve for marginal cost in all French markets. Finds marginal costs that 
    set the system of bargaining equations to zero for each market given 
    observed prices.

    Parameters
    ----------
    parprice : pandas.Series, (Nmarket x Nprodnor,)
        Upstream opportunity cost of parallel trade (NOK/DDD) for each
        market in Norway. Indexed by market as first index.
    frsharediff : pandas.Series, (Nmarket x Nprodnor,)
        Difference in shares in Norwegian market from agreement in France.
        Indexed by market as first index.
    bargainparam : float
        (Nash) bargaining power of producer vis-a-vis French regulator. 
        The default is 0.5.
    
    Returns
    -------
    mc : pandas.Series, (Nmarket x Nmol,)
        Marginal cost (USD/SU) of originator drug within each molecule for 
        each market in France. Indexed by market as first index.
    success : list (bool)
        Indicator for each market whether root finding terminated in a
        solution to the system of equations.
    """
    mclist = []
    success = []
    for m in marketlist:
        price = (demand_france.query('bargainer').
                 loc[m, 'price'].values
                 )
        lipitorprice = price[0]
        mcguess = price * 0.5
        _, producerprice = francepriceprep(m, marketdata.loc[m, 'gip'].values,
                                           lipitorprice, parprice.loc[m].values)
        optmc = lambda x: francebargaincond(m, producerprice, price, x,
                                            frsharediff.loc[m].values,
                                            bargainparam)
        res = root(optmc, mcguess)
        mclist.append(res.x)
        success.append(res.success)
    return pd.Series(hstack(mclist), index=frmbargainix), success

def createregmat(xformula, zformula, resid=None):
    """
    Create matrix for GMM objective, which will be of the form :math:`M Z W Z' M`,
    where :math:`M` is the residual maker with respect to the variables entering
    linearly (representing a concentration of the GMM objective function with
    respect to the parameters on these variables), :math:`Z` is the instrument
    matrix and  :math:`W` is a weighting matrix.
    
    If residuals are not given in `resid`, the weighting matrix will be
    calculated as :math:`W = (Z'Z)^{-1}`
    
    Parameters
    ----------
    xformula : str
        Formula for specifying set of variables entering linearly.
    zformula : str
        Formula for specifying set of instruments.
    resid : ndarray, optional
        Residuals after a first stage solution. The default is None.

    Returns
    -------
    Mx : ndarray
        Residual maker with respect to set of variables entering linearly.
    Omat : ndarray
        Matrix for GMM objective quadratic form with respect to dependent variable.
    W : ndarray
        GMM weighting matrix.
    
    Notes
    -----
    Requires `dmatrix` from `patsy` package.
    """
    
    X = dmatrix(xformula, data=marketdata.reset_index())
    Z = dmatrix(zformula, data=marketdata.reset_index())
    
    if resid is not None:
        Zeps = resid[:, None] * Z
        Zeps = Zeps - Zeps.mean(0)
        W = solve( Zeps.T.dot(Zeps), eye(Z.shape[1]) )
    else:
        W = solve(Z.T.dot(Z), eye(Z.shape[1]))
    
    Mx = eye(len(X)) - X.dot( solve(X.T.dot(Z.dot(W).dot(Z.T)).dot(X), X.T.dot(Z.dot(W).dot(Z.T)) ) )
    Omat = Mx.T.dot(Z).dot(W).dot(Z.T.dot(Mx))
    return Mx, Omat, W

def gmmobj(bargainparam, diffobj, jacobj):
    """
    Returns the GMM objective function and its gradient with respect to the
    bargaining parameters.
    
    Parameters
    ----------
    bargainparam: ndarray
        Inputs for bargaining parameters
    diffobj: dict
        Market profit differences
    jacobj: dict
        Market profit jacobians
    
    Returns
    -------
    objective : float
        Value of objective function
    grad : ndarray
        Gradient of objective function wrt bargaining parameters
    """
    parprice, parjac = solvecost(bargainparam, diffobj, jacobj)
    return parprice.dot(Omat).dot(parprice), 2 * parjac.dot(Omat).dot(parprice)


def gmmcovmat(param, diffobj, jacobj):
    """
    Calculate covariance matrix of GMM estimates.
    """
    parprice, parjac = solvecost(param.dot(pisum.T), diffobj, jacobj)
    parjac = pisum.T.dot(parjac)
    covmat = solve(parjac.dot(Omat).dot(parjac.T), eye(len(param)))
    return covmat

def maketable(param, diffobj, jacobj, optfval,
              caption=None, key=None, notes=None, numdec=2):
    """
    Create table of estimated bargaining parameters
    
    parameter is the vector of estimated parameters
    optfval is the GMM objective function at estimated parameters
    caption specifies table caption (optional)
    key specifies key for table reference (optional)
    numdec specifies the number of decimals to use
    """
    #Index for chain in LaTeX table
    paramindex = ['Chain ' + str(i) for i in range(1, 4)]
    #Column header for upstream firms in LaTeX table
    groupindex = ['Producer', 'Importer']#, 'Parallel import 2']
    
    #Get bargaining parameters
    bargainix = pd.MultiIndex.from_arrays([['di', 'pi'] * 3,
                                           [1, 1, 2, 2, 3, 3]],
                                          names=['firm', 'chainid'])
    
    bargainparam = pd.Series(param, index=bargainix)
    #Get standard errors
    bargainse = np.sqrt(gmmcovmat(param, diffobj, jacobj).diagonal())
    bargainse = pd.Series(bargainse, index=bargainix)
    #Get degrees of freedom for IV-GMM
    df = Z.shape[1] - X.shape[1] - len(param)
    chi2pval = chi2.sf(optfval, df)
    
    #Make array of parameters and std.err. in format appropriate for table generation
    #Chains in rows, upstream firms in columns
    parmat = column_stack([bargainparam['di'].values,
                           bargainparam['pi'].values])
    semat = column_stack([bargainse['di'].values,
                          bargainse['pi'].values])
    
    #Format table to text with number formatting
    parmat = array([[f'${x:.2f}$' for x in z] for z in parmat])
    semat = array([[f'$({x:.2f})$' for x in z] for z in semat])
    
    #Add on parameter labels to parameter matrix and empty column to std.err matrix
    parmat = column_stack([paramindex, parmat])
    semat = column_stack([[''] * len(paramindex), semat])
    
    #Join arrays to one
    tablemat = hstack([row_stack([parmat[i], semat[i]]).T for i in range(3)]).T
    tablerows = ['\t&\t'.join(x) + r' \\ ' + '\n' for x in tablemat]
    
    #Number of columns in table
    ncol = len(groupindex)
    
    #Construct table
    table = r'\begin{table}[h]' + '\n'
    table += r'\centering' + '\n'
    
    #Make caption
    if caption is not None:
        captionstr = r'\caption{' + caption
        if key is not None:
            captionstr += r' \label{' + key + r'}'
        captionstr += r'}' + '\n'
        table += captionstr
    
    table += r'\begin{tabular}{l' + 'c' * ncol + r'}' + '\n'
    table += r'\toprule' + '\n'
    table += '\t&\t' + '\t&\t'.join(groupindex) + r' \\' + ' \n'
    table += r'\midrule' + '\n'
    
    #Construct table body
    for i in range(len(tablerows)):
        table += tablerows[i]
        if i%2 == 1:
            table += r'\addlinespace' + '\n'
    
    #Statistics
    table += r'\midrule' + '\n'
    table += r'$J$ &' + '\t'
    table += rf'\multicolumn{{{ncol}}}{{c}}{{${optfval:,.2f}$}} \\ ' + '\n'
    table += rf'P-value $\chi^2({df})$ &' + '\t'
    table += rf'\multicolumn{{{ncol}}}{{c}}{{${chi2pval:,.2f}$}} \\ ' + '\n'
    table += r'$N$ &' + '\t'
    table += rf'\multicolumn{{{ncol}}}{{c}}{{${len(X)}$}} \\ ' + '\n'
    #End of table
    table += r'\bottomrule' + '\n'
    table += r'\end{tabular}' + '\n'
    # Add notes, if given
    if notes is not None:
        table += rf'\caption*{{\emph{{{notes}}}}}' + '\n'
    table += r'\end{table}'
    
    return table

def calcchangevar(dx, x0, groupvars):
    """
    Calculates relevant changes in level and percent of variable for firms in market
    Takes change and baseline of variable for each product-chain combination as input
    Returns a DataFrame with change in level and percent for both chains and upstream firms
    """
    # Combined index for chains and upstream firms
    usid = {0: 4, 1: 5}
    # Upstream terms
    usdx = dx.sum(level='parallel')
    usdx = usdx.reset_index()
    usdx['chainid'] = usdx.parallel.map(usid)
    usdx = usdx.set_index(groupvars)[0]
    usx0 = x0.sum(level='parallel')
    usx0 = usx0.reset_index()
    usx0['chainid'] = usx0.parallel.map(usid)
    usx0 = usx0.set_index(groupvars)[0]
    
    dx = dx.append(usdx)
    x0 = x0.append(usx0)
    return pd.DataFrame({'level': dx, 'perc': dx / x0})

def modifyindex(ser):
    """Modify index of Series from market, chainid and choiceid to
    market, chainid and parallel to facilitate calculation of counterfactual
    changes."""
    # Check conforming to market, chainid and choiceid
    assert(ser.index.names == marketindexcols)
    name = ser.name
    ser = ser.reset_index()
    ser['parallel'] = ser.choiceid.map(choiceidtopar)
    ser = ser.set_index(['market', 'chainid', 'parallel'])[name]
    return ser

def calccfchanges(newshare, newparprice, newgip, pricefactor=1,
                  baseline=False, scale=1, mc=None):
    """
    Calculate changes from observed outcomes to provided counterfactual outcomes
    """
    # Change indexing of marketdata and provided series
    usedata = marketdata.reset_index().set_index(['market', 'chainid', 'parallel'])
    newshare = modifyindex(newshare)
    newparprice = modifyindex(newparprice)
    newgip = modifyindex(newgip)
    # Get price series
    price = usedata.maxprice
    # Aggregation columns
    groupvars = ['parallel', 'chainid']
    firmgroupvars = 'parallel'
    # Quantitites
    q = [usedata.marketddd * marketdata.share, usedata.marketddd * newshare]
    dq = q[1] - q[0]
    # Change in quantity for parallel traded goods relevant for upstream firms
    # loc[:,:,1] picks out parallel and loc[:,:,0] direct
    pardq = pd.Series(-dq.loc[:,:,1].values, index=dq.loc[:,:,0].index)
    pardq = pardq.append(dq.loc[:,:,1]).sort_index()
    # CF quantity for parallel traded goods
    parq1 = pd.Series(-q[1].loc[:,:,1].values, index=q[1].loc[:,:,0].index)
    parq1 = parq1.append(q[1].loc[:,:,1]).sort_index()
    # Baseline quantity for parallel traded goods (upstream profit relevant)
    parq0 = pd.Series(-q[0].loc[:,:,1].values, index=q[0].loc[:,:,0].index)
    parq0 = parq0.append(q[0].loc[:,:,1]).sort_index()
    
    if mc is not None:
        mc = usedata.join(mc, on='market').mc
        varcost = (q[0] * mc).groupby(groupvars[-1]).sum() * scale
    else:
        varcost = 0
    
    # Change in wholesale price
    dgipchain = (newgip - usedata.gip).groupby(groupvars).mean()
    gip0chain = usedata.gip.groupby(groupvars).mean()
    gipchain = pd.DataFrame({'level': dgipchain, 'perc': dgipchain / gip0chain})
    dgipfirm = (newgip - usedata.gip).groupby(firmgroupvars).mean()
    gip0firm = usedata.gip.groupby(firmgroupvars).mean()
    gipfirm = pd.DataFrame({'level': dgipfirm, 'perc': dgipfirm / gip0firm})
    gipfirm.reset_index(inplace = True)
    gipfirm['chainid'] = gipfirm.parallel.map({0: 4, 1: 5})
    gipfirm.set_index(groupvars, inplace = True)
    gipchange = gipchain.append(gipfirm)
    # Change in quantities per entity, yearly measure
    dqchain = dq.groupby(groupvars).sum() * scale
    q0chain = q[0].groupby(groupvars).sum() * scale
    qchange = calcchangevar(dqchain, q0chain, groupvars)
    # Change in revenue
    drevchain = (q[1] * price * pricefactor - q[0] * price).groupby(groupvars).sum() * scale
    rev0chain = (q[0] * price).groupby(groupvars).sum() * scale
    revchange = calcchangevar(drevchain, rev0chain, groupvars)
    # Change in chain cost/firm revenue
    dcostchain = (q[1] * newgip - q[0] * usedata.gip).groupby(groupvars).sum() * scale
    cost0chain = (q[0] * usedata.gip).groupby(groupvars).sum() * scale
    costchange = calcchangevar(dcostchain, cost0chain, groupvars)
    # Change in firm opportunity cost of parallel trade
    dcostfirm = (parq1 * newparprice - parq0 * usedata.parprice).groupby(groupvars).sum() * scale
    dcostfirm.name = 'cost'
    cost0firm = (parq0 * usedata.parprice).groupby(groupvars).sum() * scale
    cost0firm.name = 'cost'
    # Change in Chain profits, yearly measure
    dprofitchain = drevchain - dcostchain
    profit0chain = rev0chain - cost0chain
    chainprofitchange = calcchangevar(dprofitchain, profit0chain, groupvars)
    # Change in firm profits
    dprofitfirm = dcostchain - dcostfirm
    profit0firm = cost0chain - cost0firm
    # Subtract variable costs for Pfizer (if mc provided)
    profit0firm.loc[0, [1, 2, 3]] -= varcost
    firmprofitchange = calcchangevar(dprofitfirm, profit0firm, groupvars)
    if baseline:
        return q0chain, rev0chain, cost0chain, gip0chain, gip0firm, profit0chain, profit0firm, varcost
    else:
        return qchange, revchange, costchange, gipchange, chainprofitchange, firmprofitchange

def makecftable(newshare, newparprice, newgip, pricefactor=1,
                frprice=None, mc=None,
                caption=None, key=None,
                scale=1, numdec=2):
    """
    Create table of counterfactual numbers
    
    """
    #Index for chain in LaTeX table
    rowindex = ['Pharmacy Chain ' + str(i) for i in range(1, 4)]
    #Firm index for rows
    firmindex = ['Manufacturer', 'Parallel']
    #Column header for counterfactual changes
    groupindex = [r'$\Delta q_0$', r'$\Delta q_1$', r'$\Delta w_0$',
                  r'$\Delta w_1$', r'$\Delta \pi$']
    
    #Get CF numbers
    if mc is not None:
        lipitormc = mc.loc[:, 0].mul(marketinfo.lipitorpriceconvert).rename('mc')
    else:
        lipitormc = None
    qchange, revchange, costchange, gipchange, chainprofitchange, firmprofitchange \
     = calccfchanges(newshare, newparprice, newgip,
                     pricefactor=pricefactor, scale=scale, mc=lipitormc)
    
    # Convert to chain profit changes
    x = chainprofitchange.loc[0].loc[:3]
    y = chainprofitchange.loc[1].loc[:3]
    
    #Total profit change for each entity
    profitchange = (chainprofitchange[['level']].groupby('chainid').sum().loc[:3]
                    .append(firmprofitchange[['level']].groupby('chainid').sum().loc[4:])
                   )
    # Add percentage, converting to percentage change for chains
    profitchange['perc'] = (((x.level + y.level) * x.perc * y.perc / (x.level * y.perc + y.level * x.perc))
                            .append(firmprofitchange.perc.groupby('chainid').sum().loc[4:])
                           )
    
    #Make Frame for chain numbers
    cols = ['dq0', 'dq1', 'dgip0', 'dgip1', 'dprofit']
    #Scaling for columns
    colscale = pd.Series(dict(zip(cols, [1e-6, 1e-6, 1, 1, 1e-6])))
    #Level changes (making sure to get)
    dlevel = [qchange.level.unstack(0)[1], qchange.level.unstack(0)[0],
              gipchange.level.unstack(0)[1], gipchange.level.unstack(0)[0],
              profitchange.level]
    dflevel = pd.DataFrame(dict(zip(cols, dlevel))).mul(colscale)
    #Percent changes
    dperc = [qchange.perc.unstack(0)[1], qchange.perc.unstack(0)[0],
             gipchange.perc.unstack(0)[1], gipchange.perc.unstack(0)[0],
             profitchange.perc]
    dfperc = pd.DataFrame(dict(zip(cols, dperc)))
    
    #Number of columns in table
    ncol = len(cols)
    
    #Construct table
    table = r'\begin{table}[h!]' + '\n'
    table += r'\centering' + '\n'
    
    #Make caption
    if caption is not None:
        captionstr = r'\caption{' + caption
        if key is not None:
            captionstr += r' \label{' + key + r'}'
        captionstr += r'}' + '\n'
        table += captionstr
    
    table += r'\begin{tabular}{l' + 'r' * ncol + r'}' + '\n'
    table += r'\toprule' + '\n'
    table += '\t&\t' + '\t&\t'.join(groupindex) + r' \\' + ' \n'
    table += r'\midrule' + '\n'
    
    
    #Chain values
    i = 1
    for c in rowindex:
        lval = dflevel[cols].loc[i].values
        pval = dfperc[cols].loc[i].values
        table += c + '\t&\t'
        table += '\t&\t'.join(['${:.2f}$'.format(x) for x in lval]) + r' \\' + ' \n'
        table += '\t&\t' + '\t&\t'.join([r'${:.0f}\%$'.format(x * 100) for x in pval]) + r' \\' + ' \n'
        table += r'\addlinespace' + '\n'
        i += 1
    
    #Firm values
    table += r'\midrule' + '\n'
    table += '\t&\t' * ncol + r'$\Delta \Pi$ \\' + ' \n'
    table += r'\midrule' + '\n'
    for c in firmindex:
        lval = dflevel[cols].loc[i].values
        pval = dfperc[cols].loc[i].values
        table += c + '\t&\t'
        table += '\t&\t'.join(['${:.2f}$'.format(x) for x in lval]) + r' \\' + ' \n'
        table += '\t&\t' + '\t&\t'.join([r'${:.0f}\%$'.format(x * 100) for x in pval]) + r' \\' + ' \n'
        table += r'\addlinespace' + '\n'
        i += 1

        
    if frprice is not None:
        frgroup = [0, 1, 1, 1]
        frindex = ['Manufacturer on French market', 'Other on French market']
        # Impact on manufacturers in France
        # Baseline French profits
        bargainerprice0 = demand_france.query('bargainer').price
        pifr0 = [bargainerprofitfrance(m, bargainerprice0.loc[m].values,
                                       mc.loc[m].values, nok=True) / 1000
                 for m in marketlist]
        pifr0 = pd.Series(hstack(pifr0), index=frmbargainix).sum(level='moleculeid')
        # French profits with new prices
        pifrnofc = [bargainerprofitfrance(m, frprice.loc[m].values,
                                          mc.loc[m].values, nok=True) / 1000
                    for m in marketlist]
        pifrnofc = pd.Series(hstack(pifrnofc), index=frmbargainix).sum(level='moleculeid')
        # Change in level and percent
        frdeltalevel = (pifrnofc - pifr0).groupby(frgroup).sum()
        frdeltaperc = frdeltalevel / pifr0.groupby(frgroup).sum()
        
        table += r'\midrule' + '\n'
        for i, c in enumerate(frindex):
            lval = frdeltalevel.loc[i] * scale
            pval = frdeltaperc.loc[i]
            table += c + '\t&\t'
            table += '\t&\t' * (ncol - 1) + rf'${lval:.2f}$' + r' \\' + ' \n'
            table += '\t&\t' * ncol  + rf'${pval * 100:.2f}\%$' + r' \\' + ' \n'
            table += r'\addlinespace' + '\n'
    
    #End of table
    table += r'\bottomrule' + '\n'
    table += r'\end{tabular}' + '\n'
    table += r'\end{table}'
    
    table = table.replace('$nan$', '').replace('$nan\%$', '')
    return table

def makecfseqredtable(sharelist, parprice, giplist, pricereductions,
                      frpricelist, mc,
                      caption=None, key=None,
                      scale=1, numdec=2, mscale=1e-6):
    """
    Create table of profit counterfactuals for sequence of price reductions
    """
    nreductions = len(pricereductions)
    
    # Index for chain in LaTeX table
    rowindex = [f'Pharmacy Chain {i}' for i in range(1, 4)]
    # Firm index for rows
    firmindex = ['Manufacturer', 'Parallel']
    # Value column names
    cols = [f'$-${100 * pr:.0f}\%' for pr in pricereductions]
    ncol = len(cols)
    
    lipitormc = mc.loc[:, 0].mul(marketinfo.lipitorpriceconvert).rename('mc')
    
    # Baseline French profits
    frgroup = [0, 1, 1, 1] # Grouping of molecules in France
    bargainerprice0 = demand_france.query('bargainer').price
    pifr0 = [bargainerprofitfrance(m, bargainerprice0.loc[m].values,
                                   mc.loc[m].values, nok=True) / 1000
             for m in marketlist]
    pifr0 = pd.Series(hstack(pifr0), index=frmbargainix).sum(level='moleculeid')
    pifr0 = pifr0.groupby(frgroup).sum()
    
    # Construct table
    table = r'\begin{table}[h!]' + '\n'
    table += r'\centering' + '\n'
    
    # Make caption
    if caption is not None:
        captionstr = r'\caption{' + caption
        if key is not None:
            captionstr += r' \label{' + key + r'}'
        captionstr += r'}' + '\n'
        table += captionstr
    
    table += r'\begin{tabular}{ll' + 'r' * ncol + r'}' + '\n'
    table += r'\toprule' + '\n'
    table += r'\Delta p_t\t&\t' + '\t&\t' + '\t&\t'.join(cols) + r' \\' + ' \n'
    table += r'\midrule' + '\n'
    table += r'\Delta \pi' + '\t&\t' * (ncol + 1) + r' \\' + ' \n'
    
    dflevel = pd.DataFrame()
    dfperc = pd.DataFrame()
    exits = []
    govt = []
    govtprodnet = []
    frprofit_pfizer = []
    frprofit_other = []
    for i in range(nreductions):
        pr = pricereductions[i]
        pricefactor = 1 - pr
        frprice = frpricelist[i]
        redshare = sharelist[i]
        redgip = giplist[i]
        
        redparprice = (parprice + 
                       (frprice.loc[:, 0] - marketinfo.frlipitorprice) *
                       marketinfo.lipitorpriceconvert
                       )
        redparprice.name = 'parprice'
        exits.append((redshare == 0).sum())
        #Get CF numbers
        _, _, _, _, chainprofitchange, firmprofitchange \
         = calccfchanges(redshare, redparprice, redgip,
                         pricefactor=pricefactor, mc=lipitormc, scale=scale)
         
        # Convert to chain profit changes
        x = chainprofitchange.loc[0].loc[:3]
        y = chainprofitchange.loc[1].loc[:3]
        
        #Total profit change for each entity
        profitchange = (chainprofitchange[['level']].sum(level='chainid').loc[:3]
                        .append(firmprofitchange[['level']].sum(level='chainid').loc[4:])
                        * mscale)
        # Add percentage, converting to percentage change for chains
        profitchange['perc'] = (((x.level + y.level) * x.perc * y.perc / (x.level * y.perc + y.level * x.perc))
                                .append(firmprofitchange.perc.sum(level='chainid').loc[4:])
                               )
        
        dflevel[i] = profitchange.level
        dfperc[i] = profitchange.perc
        
        deltaspend = - marketdata.eval(f'ddd * maxprice * {pr} / 4').sum() * mscale
        govt.append(deltaspend)
        govtprodnet.append(deltaspend - profitchange.loc[4, 'level'])
        
        # French profits with new prices
        pifrcf = [bargainerprofitfrance(m, frprice.loc[m].values,
                                        mc.loc[m].values,
                                        nok=True) / 1000
                  for m in marketlist]
        pifrcf = pd.Series(hstack(pifrcf), index=frmbargainix).sum(level='moleculeid')
        pifrcf = pifrcf.groupby(frgroup).sum()
        
        frprofit_delta = (pifrcf - pifr0) * scale
        
        frprofit_pfizer.append(frprofit_delta.loc[0])
        frprofit_other.append(frprofit_delta.loc[1])
        
    # Chain values
    for i, c in zip(range(1, 4), rowindex):
        lval = dflevel.loc[i].values
        pval = dfperc.loc[i].values
        table += f'\t&\t {c} \t&\t'
        table += '\t&\t'.join([f'${x:.2f}$' for x in lval]) + r' \\' + ' \n'
        table += '\t&\t' * 2 + '\t&\t'.join([fr'${100 * x:.0f}\%$' for x in pval]) + r' \\' + ' \n'
        table += r'\addlinespace' + '\n'
    
    # Firm values
    table += r'\Delta \Pi' + '\t&\t' * (ncol + 1) + r' \\' + ' \n'
    for i, c in zip(range(4, 6), firmindex):
        lval = dflevel.loc[i].values
        pval = dfperc.loc[i].values
        table += f'\t&\t {c} \t&\t'
        table += '\t&\t'.join([f'${x:.2f}$' for x in lval]) + r' \\' + ' \n'
        table += '\t&\t' * 2 + '\t&\t'.join([fr'${100 * x:.0f}\%$' for x in pval]) + r' \\' + ' \n'
        table += r'\addlinespace' + '\n'

    # Exits
    table += r'\multicolumn{2}{l}{Number of chain-market exits} &\t'
    table += '\t&\t'.join([f'${x:d}$' for x in exits]) + r' \\' + ' \n'
    
    table += r'\midrule' + '\n'
    # Government spending
    table += r'\multicolumn{2}{l}{Government spending change ($\Delta \left(p_t q_t\right)$)} &\t'
    table += '\t&\t'.join([f'${x:.2f}$' for x in govt]) + r' \\' + ' \n'
    table += r'\addlinespace' + '\n'
    
    # Government spending net of producer profit
    table += r'\multicolumn{2}{l}{$\Delta \left(p_t q_t\right - \Delta \Pi$)} &\t'
    table += '\t&\t'.join([f'${x:.2f}$' for x in govtprodnet]) + r' \\' + ' \n'
    table += r'\addlinespace' + '\n'
    
    # French market
    table += r'\multicolumn{2}{l}{Manufacturer on French market} &\t'
    table += '\t&\t'.join([f'${x:.2f}$' for x in frprofit_pfizer]) + r' \\' + ' \n'
    table += r'\addlinespace' + '\n'
    
    table += r'\multicolumn{2}{l}{Other on French market} &\t'
    table += '\t&\t'.join([f'${x:.2f}$' for x in frprofit_other]) + r' \\' + ' \n'
    
    # End of table
    table += r'\bottomrule' + '\n'
    table += r'\end{tabular}' + '\n'
    table += r'\end{table}'
    
    table = table.replace('$nan$', '').replace('$nan\%$', '')
    return table

import re

def mergeboundtables(lbtbl, ubtbl):
    """
    Utility function to merge tables of bounds
    Puts unequal items in brackets
    Expects string representation of tables as output by "makecftable"
    """
    lblines = lbtbl.splitlines()
    ublines = ubtbl.splitlines()
    assert len(lblines) == len(ublines)
    mergelines = []
    for lbl, ubl in zip(lblines, ublines):
        if lbl == ubl:
            mergelines.append(lbl)
        else:
            lbcells = lbl.split('\t')
            ubcells = ubl.split('\t')
            assert len(lbcells) == len(ubcells)
            mergecells = []
            for lbc, ubc in zip(lbcells, ubcells):
                if lbc == ubc:
                    mergecells.append(lbc)
                else:
                    # Confident of exactly one match
                    if lbc.find(r'\%') != -1:
                        lb = re.findall('-?\d+\\\%', lbc)[0]
                        ub = re.findall('-?\d+\\\%', ubc)[0]
                    else:
                        lb = re.findall('-?\d+.\d+', lbc)[0]
                        ub = re.findall('-?\d+.\d+', ubc)[0]
                    mcell = r'$[{}, {}]$'.format(lb, ub)
                    # Preserve latex table line shift
                    if lbc.find(r'\\') != -1:
                        mcell += r' \\'
                    mergecells.append(mcell)
            mergelines.append('\t'.join(mergecells))
    return '\n'.join(mergelines)



#%% ADD UPSTREAM FIRM IV

# Create (mean) utilities
util = mu.dot(mw)
util.name = 'util'

# Inclusive value for upstream firm in each market
marketdata = marketdata.join(util)
upstreamutils = marketdata.groupby(['market', 'firm']).util.apply(lambda x: log(exp(x).sum()))
upstreamutils.name = 'upstreamIV'

# Join onto data
marketdata.reset_index(inplace=True)
marketdata = marketdata.join(upstreamutils, on=['market', 'firm'])
marketdata.set_index(marketindexcols, inplace=True)

# Calculate disagreement shares
start = dtim()
outsideshare, outsidetheta, diagnostic = getoutsideshares(mu)
dur = dtim() - start

sharediff = -outsideshare.share.unstack('rid').sub(marketdata.share, 0).sort_index()
diffobj, jacobj = gathermarketdiffs(mu, theta, sharediff)



#%% ESTIMATE BARGAINING MODEL

pisum = block_diag(1, [1, 1, 1], 1, [1, 1, 1], 1, [1, 1, 1]).T

# Wrapper for gmm objective function for minimization
def gmmwrap(x):
    g = x.dot(pisum.T)
    r = gmmobj(g, diffobj, jacobj)
    return r[0], r[1].dot(pisum)


# Set up estimation matrices
xvars = "(UKn + Czechn) + (UKn + Czechn) * C(parallel)"
zvars = xvars + " + C(chainid) + (nokusd + nokeur + nokczk + upstreamIV) * C(parallel)"

X = dmatrix(xvars, data=marketdata.reset_index())
Z = dmatrix(zvars, data=marketdata.reset_index())

# First stage
Mx, Omat, W = createregmat(xvars, zvars)

guess = array([-1.4, 0.25, 1.22, -1.3, 1.02, 0.1])
res = minimize(gmmwrap, guess, jac=True)

print("Convergence in first stage: {} (objective value: {:.3f})".format(res.success, res.fun))

# Second stage
parprice = solvecost(res.x @ pisum.T, diffobj, jacobj)[0]

# Residuals after first stage
resid = Mx.dot(parprice)

# New weighting matrix
Mx, Omat, W = createregmat(xvars, zvars, resid=resid)

res2 = minimize(gmmwrap, res.x, jac=True)

print(f"Convergence in second stage: {res2.success} (objective value: {res2.fun:.3f})")



#%% TABLE 4.5
caption = 'Bargaining parameter estimates (GMM)'
key = 'tbl:bargaining'
tblnote = 'Explanatory variables: {}. Instruments: {}.'.format(xvars, zvars)
print(maketable(res2.x, diffobj, jacobj, gmmwrap(res2.x)[0],
                caption=caption,
                key=key,
                notes=tblnote)
     )

with open('{tabfigdir}bargainparam{args.filepostfix}.tex', 'w') as f:
    f.write(maketable(res2.x, diffobj, jacobj, gmmwrap(res2.x)[0],
                      caption=caption,
                      key=key,
                      notes=tblnote)
           )

param2 = res2.x @ pisum.T
parprice = solvecost(param2, diffobj, jacobj, asseries=True)
# Merge onto market data to facilitiate computation of counterfactual changes
marketdata = marketdata.join(parprice)

# Bargaining power according to choice ID
bargainparam = pd.Series(res2.x,
                         index=pd.Index(arange(2, 8), name='choiceid'),
                         name='power')



#%% PREPARE COUNTERFACTUALS

# Non-distortion shares, outside shares and difference
ndshare = getnondistshare(mu)
ndoutsideshare = getnondistoutsideshares(mu)
ndsharediff = -ndoutsideshare.share.unstack('rid').sub(ndshare, 0).sort_index()

# Bargaining companies' prices in France
bargainerprice0 = demand_france.query('bargainer').price

# Non-distortion shares if dropping French source
francemask = marketdata.sourcecountry == 'France'
francemarketlist = marketdata[francemask].reset_index().market.unique()
francesourcedict = dict(zip(francemarketlist,
                            [marketdata[francemask].reset_index('choiceid').
                             loc[m].choiceid.values
                             for m in francemarketlist]
                            ))

ndnofranceshare = getnondistshare(mu, francesourcedict)
francesharediff = ndshare - ndnofranceshare

# Get MC from bargaining conditions in France
mcest, estsuccess = francemcest(parprice, francesharediff, fr_bargain_param)
assert(sum(estsuccess) == nmarket)
mcest.name = 'mc'



#%% TABLE 5.1
# Equilibrium under no parallel trade in Norway
# Note: Parallel trade opportunity cost not relevant in the counterfactual
# Market shares in Norway with no parallel trade
npshare = getnoparshare(mu)
# Price in France under estimated MC when no parallel trade
autprice_france, autsuccess = equilfranceaut(mcest, fr_bargain_param)
assert(sum(autsuccess) == nmarket)

# Bargained solution (if all sales of disagreeing pharmacy are lost)
mcnokddd = mcest.loc[:, 0].mul(marketinfo.lipitorpriceconvert).rename('mc')
mcexpand = marketdata.join(mcnokddd, on='market').mc
npgip = marketdata.maxprice * bargainparam + mcexpand * (1 - bargainparam)
npgip.update(marketdata[marketdata.parallel == 1].gip)
npgip.name = 'gip'

# Full appropriation of surplus by manufacturer (no sales lost)
npgipf = marketdata.maxprice.copy()
npgipf.update(marketdata[marketdata.parallel == 1].gip)
npgipf.name = 'gip'

# Tables with bounds
caption = r'Impact of removing parallel imports of Atorvastatin'
key = r'tbl:cf_remove'
tbl_cf_remove_lb = makecftable(npshare, parprice, npgip,
                               frprice=autprice_france, mc=mcest,
                               caption=caption, key=key, scale=0.25)
tbl_cf_remove_ub = makecftable(npshare, parprice, npgipf,
                               frprice=autprice_france, mc=mcest,
                               caption=caption, key=key, scale=0.25)

print(mergeboundtables(tbl_cf_remove_lb, tbl_cf_remove_ub))

with open('{tabfigdir}cf_remove{args.filepostfix}.tex', 'w') as f:
    f.write(mergeboundtables(tbl_cf_remove_lb, tbl_cf_remove_ub))

# Relative price change in France
(autprice_france / bargainerprice0.values - 1).describe()



#%% TABLE 5.2
# Equilibrium under no foreclosure
endgip, eprice_france, esuccess =\
    intlequil(parprice, mcest, francesharediff, ndsharediff,
              fr_bargain_param, param2)
endgip.name = 'gip'
assert(sum(esuccess) == nmarket)

eparprice = (parprice + 
             (eprice_france.loc[:, 0] - marketinfo.frlipitorprice) *
             marketinfo.lipitorpriceconvert
             )
eparprice.name = 'parprice'

caption = r'Impact of preventing parallel imports foreclosure ($\theta_{ct} = 1$) of Atorvastatin'
key = r'tbl:cfprofit_france'
print(makecftable(ndshare, eparprice, endgip,
                  frprice=eprice_france, mc=mcest,
                  caption=caption, key=key, scale=0.25))

with open('{tabfigdir}cf_noforeclosure{args.filepostfix}.tex', 'w') as f:
    f.write(makecftable(ndshare, eparprice, endgip,
                        frprice=eprice_france, mc=mcest,
                        caption=caption, key=key, scale=0.25))

# Relative price change in France
(eprice_france / bargainerprice0.values - 1).describe()



#%% TABLE 5.3
# Equilibrium under no foreclosure and reduced prices in Norway
reduction = 0.2

redgip, redshare, redprice_france, redsuccess = \
    solvenondistexit(mu, parprice, mcest, fr_bargain_param, param2,
                     price=marketdata.maxprice * (1 - reduction))
assert(sum(redsuccess) == nmarket)

redparprice = (parprice + 
               (redprice_france.loc[:, 0] -  marketinfo.frlipitorprice) *
               marketinfo.lipitorpriceconvert
               )
redparprice.name = 'parprice'

caption = r'Impact of reducing the price ceiling by {reduction * 100:.0f}\% of Atorvastatin'
key = r'tbl:pricereduction'
print(makecftable(redshare, redparprice, redgip,
                  pricefactor=(1 - reduction),
                  frprice=redprice_france, mc=mcest,
                  caption=caption, key=key, scale=0.25))

with open('{tabfigdir}cf_pricered{args.filepostfix}.tex', 'w') as f:
    f.write(makecftable(redshare, redparprice, redgip,
                        pricefactor=(1 - reduction),
                        frprice=redprice_france, mc=mcest,
                        caption=caption, key=key, scale=0.25))

# Relative price change in France
(redprice_france / bargainerprice0.values - 1).describe()



#%% TABLE 5.4
# Equilibria under no foreclosure and range of price reductions in Norway
redlist = [0.1, 0.15, 0.2, 0.25, 0.3]
lpgiplist = []
lpsharelist = []
lpfrpricelist = []
lpsucclist = []
for red in redlist:
    g1, s1, f1, su1 =\
        solvenondistexit(mu, parprice, mcest, fr_bargain_param, param2,
                         price=marketdata.maxprice * (1 - red))
    g1.name = 'gip'
    
    lpgiplist.append(g1)
    lpsharelist.append(s1)
    lpfrpricelist.append(f1)
    lpsucclist.append(su1)

assert(all([sum(x) == nmarket for x in lpsucclist]))


caption = 'Impact of reducing the price ceiling of Atorvastatin (and preventing parallel imports foreclosure)'
key = r'tbl:reductions'
print(makecfseqredtable(lpsharelist, parprice, lpgiplist, redlist, lpfrpricelist, 
                        mcest, caption=caption, key=key, scale=0.25))

with open('{tabfigdir}cf_pricered_seq{args.filepostfix}.tex', 'w') as f:
    f.write(makecfseqredtable(lpsharelist, parprice, lpgiplist, redlist, lpfrpricelist, 
                              mcest, caption=caption, key=key, scale=0.25))
