#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Morten Sæthre

Paper: "On the Role of Parallel Trade on Manufacturers and Retailers Profits in the Pharmaceutical Sector"
Authors: Pierre Dubois (TSE) and Morten Saethre (NHH)

Estimate demand model

Inputs:
    Product-market level data containing market shares and margins stored in "data/market.pq" created in "clean_data.py"
    Individual choice data stored in "data/individual.pq" created in "clean_data.py"

Outputs:
    Estimated parameters governing preference heterogeneity
    Mean utility parameters for each product-market combination
    Foreclosure parameters for each chain-market combination
    Table of demand estimates (Table 4.3)
"""

#%% SETUP

import pandas as pd
import numpy as np
from numpy import array, hstack, exp, log, zeros, ones, eye, arange, diag, \
    column_stack, row_stack, repeat, tile
from numpy.random import rand, randn
from scipy.optimize import minimize
from scipy.sparse import coo_matrix
from sklearn.utils.extmath import cartesian
from numpy.polynomial.hermite import hermgauss
import theano
import theano.tensor as T
from theano.tensor.slinalg import kron as tkron
from theano.tensor.slinalg import solve as tsolve
from theano import sparse as tsparse
import pickle
from timeit import default_timer as dtim
import argparse
import mkl

mkl.set_num_threads(mkl.get_max_threads())
pd.options.io.parquet.engine = 'fastparquet'

# Get command line arguments for running estimation (or use default)
parser = argparse.ArgumentParser()
parser.add_argument('-nu', '--nugroup', default=4, type=int,
                    help='Number of latent classes')
parser.add_argument('-qn', '--quadnodes', default=5, type=int,
                    help='Number of quadrature nodes per dimension of heterogeneity')
parser.add_argument('-fp', '--filepostfix', default='',
                    help="Postfix for result files")
args = parser.parse_args()

print(f'Number of latent classes: \t {args.nugroup}')
print(f'Number of quadrature nodes: \t {args.quadnodes}')
print(f'File postfix: \t {args.filepostfix}')

# Directories
rootdir = '/home/morten/'
projectroot = f'{rootdir}projects/parallel/'
datadir = f'{projectroot}data/'
resultdir = f'{projectroot}results/'
tabfigdir = f'{projectroot}tabfig/'



#%% LOAD DATA

print("Loading data")
marketdata = pd.read_parquet(f'{datadir}market.pq')
indchoices = pd.read_parquet(f'{datadir}individual.pq')



#%% SET VARIABLES

# Choice ID to normalize utility against
basechoice = 0

# Number of chains
nchain = marketdata.chainid.nunique()
# Chain dummy columns
chaincols = ['chain' + str(i) for i in range(1, 4)]

# Get list of markets
marketlist = marketdata.market.unique()
nmarket = len(marketlist)

# Sort data
indchoices.sort_values(['patid', 'market', 'choiceid'], inplace=True)
marketdata.sort_values(['market', 'choiceid'], inplace=True)

# List of individuals
indlist = indchoices.patid.unique()
# dictionary of individual choices
indchoicedict = indchoices.groupby('patid').mcid.apply(lambda x: x.unique().tolist()).to_dict()
indchoicelist = indchoicedict.values()

indchoices['obsid'] = indchoices.groupby('patid').ngroup()

# Matrix to sum over individual choices
# (Stacked vectors of length nmarkets x nprods with 1 if alternative chosen)
nindobs = len(indchoices)
row = indchoices.obsid.values
col = indchoices.mcid.values
indchoicespmat = coo_matrix((ones(nindobs), (row, col)), dtype=np.int16)

# Make indicators for maximum margin option and parallel trade 0/1
# (Works better than True/False for Theano calculations)
for c in ['ismaxoption', 'parallel']:
    marketdata[c] = marketdata[c].astype(int)



#%% PREPARE ID'S AND OBJECTS FOR SIMULATION

# Get ID's for individuals and number of individuals
indid = indchoices['patid'].unique()
nind = len(indid)

# Features of unobserved heterogeneity
varcols = ['parallel']
varcols.extend(chaincols[1:])
prefdistdim = len(varcols)
nugroup = args.nugroup
# Number of active parameters for each latent class
nugroupparams = prefdistdim

simdraws = args.quadnodes
simlength = args.quadnodes ** prefdistdim

# Number of parameters to be estimated by maximum likelihood
"""
One parameter per dimension of heterogeneity for each latent class
Separate mean utility for each dimension of heterogeneity for each latent class
One nesting paremeter per chain
Parmeters for probability simplex of latent classes (#classes - 1)
"""
nmlparam = nugroup * prefdistdim + (nugroup - 1) * prefdistdim + nchain + nugroup - 1

# Array to hold individual gradients
indgrad = zeros((nind, nmlparam))

### Index objects
#Index for latent class parameters
ugroupdistix = pd.MultiIndex.from_product((varcols, arange(nugroup)),
                                          names=['variable', 'latclass'])
# Make simulation draw index
simix = pd.RangeIndex(simlength, name='sim')
# Index for theta (market-chain)
thetaix = marketdata.groupby(['market', 'chainid']).share.sum().index


#Get quadrature for market share integration
singlenodes, singleweights = hermgauss(args.quadnodes)
# Cartesian product of nodes over dimensions of heterogeneity
basenodes = cartesian([singlenodes] * prefdistdim).T
# Make weights reflect cartesian product of nodes over dimensions and rescale (sum to one)
weightscale = np.pi ** (-prefdistdim / 2.)
baseweights = weightscale * cartesian([singleweights] * prefdistdim).prod(1)
# Repeated over latent classes (rows indexed as in ugroupdistix)
repnodes = repeat(basenodes, nugroup, 0)


#Set market index for fast retrieval
marketdata.set_index('market', inplace=True)

#Make outside good share
temp = ((marketdata.share * (marketdata.choiceid == basechoice))
        .groupby('market').transform('sum')
       )
#Guess for alphas
marketdata['alphalogit'] = marketdata.share.apply(log) - temp.apply(log)



#%%     THEANO EXPRESSIONS

print("Setting up Theano functions")

# Variables
td = T.vector('d')          # Vector of deltas
tmu = T.matrix('mu')        # Matrix of individual utility deviations
tw = T.vector('w')          # Vector of weights for utility deviations
tm = T.vector('m')          # Vector of profit margins
town = T.matrix('owner')    # Ownership dummies
ttheta = T.vector('theta')  # Vector of supply shading
tlda = T.vector('lambda')   # in-chain "inverse" correlation parameter vector
tulo = T.vector('u_l')      # Vector of shadow prices for lower limit of theta (0)
tuhi = T.vector('u_h')      # Vector of shadow prices for upper limit of theta (1)
thi = T.vector('high')      # Vector of dummy for highest margin good
tsobs = T.vector('sobs')    # Vector of observed shares

# d, theta, lda and hi as column vectors
tdcol = td.dimshuffle([0, 'x'])
tthetacol = ttheta.dimshuffle([0, 'x'])
tldacol = tlda.dimshuffle([0,'x'])
thicol = thi.dimshuffle([0, 'x'])

# theta and lambda expanded to product (as column)
tthetaprod = town.T.dot(tthetacol)
tldaprod = town.T.dot(tldacol)

# Utility per good per simulation, weigthed by nesting parameter
tutil = (tdcol + tmu) / tldaprod
# Exponentiated utility
texpu = T.exp(tutil)
# Within chain sum of exp utility
tsumexpu = town.dot(texpu)
# Individual within-chain choice probability (raw, without distortion)
tcondcpnodist = texpu / town.T.dot(tsumexpu)
# Conditional choice probability, given chain choice
tcondcp = tthetaprod * tcondcpnodist + thicol * (1 - tthetaprod)

# Inclusive values
tIVb = T.log(tsumexpu)
tIVhi = (town * thi).dot(tutil)
# Exponentiated inclusive value
texpI = T.exp(tldacol * (tthetacol * tIVb + (1 - tthetacol) * tIVhi))
# Chain choice probability
tchaincp = texpI / texpI.sum(0)
# Choice probability for chain-product combination, per node
tsimchoiceprob = tcondcp * town.T.dot(tchaincp)
# Choice probability for chain-product combination, aggregate
tchoiceprob = tsimchoiceprob.dot(tw)

# Profit function, for testing
tprofit = town.dot(tm * tchoiceprob)

# Lagrangian of Profit
tlagprofit = town.dot(tm * tchoiceprob) + tulo * ttheta + tuhi * (1 - ttheta)
# Derivative of profit lagrangian with respect to theta, only own-theta derivatives
tfoc, _ = theano.scan(fn=lambda i: T.grad(tlagprofit[i], ttheta)[i],
                      sequences=[T.arange(ttheta.shape[0])])

# Share equations, difference in observed and predicted
tshareq = (tsobs - tchoiceprob)[1:]

# System of equations for market contractions (difference in shares and FOC profit)
tsyseq = T.concatenate([tfoc, tshareq], axis=0)

# Jacobian of FOC and share equations wrt theta and delta
tJtdfoc, _ = theano.scan(
        fn=lambda i: T.grad(T.grad(tlagprofit[i], ttheta)[i], [ttheta, td]),
        sequences=[T.arange(ttheta.shape[0])]
        )
tJtdshareq = T.jacobian(tshareq, [ttheta, td])

# Full jacobian, theta (first F entries), delta (next J entries)
tJsyseq = T.concatenate([
        T.concatenate([tJtdfoc[0], tJtdfoc[1][:, 1:]], axis=1),
        T.concatenate([tJtdshareq[0], tJtdshareq[1][:, 1:]], axis=1)],
        axis=0)

####    Add calculation of preference nodes in Theano (for automatic diff)
# For constructing log likelihood wrt. preference parameters
tsigma = T.vector('su') # Vector of latent class dispersion parameters
tmean = T.vector('mu') # Vector of latent class variable specific utility
teta = T.vector('eta') # Parameters governing logit probability of latent class

# Constant objects (summation matrices and data)
# Max margin choice full vector
tahi = T.constant(marketdata.ismaxoption.astype('int8').values[:, None])
# Margin vector
tamargin = T.constant(marketdata.margin.values)
# Observed variables for preference heterogeneity (chain and parallel dummies)
tx = T.constant(marketdata[varcols].values)

# For each market-chain, pick out the highest-margin product
choosehigh = pd.get_dummies(marketdata.ismaxoption.cumsum() * 
                            marketdata.ismaxoption,
                            drop_first=True)
tchoosehigh = tsparse.constant(coo_matrix(choosehigh.T).tocsc())

# Sum over products within chain and market (from nprod x nmarket to nmarket)
mpix = repeat(arange(nmarket * 3), 2)
mchainmat = pd.get_dummies(mpix, sparse=True).sparse.to_coo()
tmchainmat = tsparse.constant(mchainmat.T.tocsc())

# Sum over chains within market (from nchain x nmarket to nmarket)
chainsum = pd.get_dummies(repeat(arange(nmarket), 3), sparse=True).sparse.to_coo()
tchainsum = tsparse.constant(chainsum.T.tocsc())

# Sum over choice sequence for each individual
tchoicemat = tsparse.constant(indchoicespmat.tocsc())

# Matrix to remove base choice from each market by pre-multiplication
basechoicemask = marketdata.choiceid != basechoice
nobase = coo_matrix(pd.get_dummies(marketdata.mcid.where(basechoicemask, 0),
                                   drop_first=True).values)
tnobase = tsparse.constant(nobase.T.tocsc())
tnobaseix = T.constant(basechoicemask).nonzero()[0]

# Reshaping parameters
tsigmacol = tsigma.dimshuffle([0, 'x'])
tmeancol = tmean.dimshuffle([0, 'x'])
# Parameter indexing (overall mean utility parameters set to zero for group 0)
usemeanpar = tile(hstack([0, ones(nugroup - 1)]), nugroupparams).astype(bool)
usepar = hstack((ones(nugroupparams * nugroup, dtype=bool), usemeanpar,
                 ones(nchain + nugroup - 1, dtype=bool)))
tuseparix = T.constant(arange(len(usepar))[usepar])

# theta and lambda expanded to all markets and product-market level (as columns)
taldacol = T.tile(tlda, nmarket).dimshuffle([0, 'x'])
taldaprod = T.repeat(taldacol, 2, axis=0)
tathetacol = ttheta.dimshuffle([0, 'x'])
tathetaprod = T.repeat(tathetacol, 2, axis=0)

# Fixed integration nodes and weights
tnu = T.constant(repnodes)
tbasew = T.constant(baseweights)
# Probability of latent class
texpeta = T.concatenate([[1], T.exp(teta)])
ttau = texpeta / texpeta.sum()
# Weights over integration nodes x latent clasees
tfw = tkron(ttau.dimshuffle([0, 'x']), tbasew)

taprefdev = tsigmacol * tnu + tmeancol
tautildev = tx.dot(taprefdev.reshape((nugroupparams, simlength * nugroup), 2))
tautil = (tdcol + tautildev) / taldaprod
# Exponentiated utility
taexpu = T.exp(tautil)
# Within chain sum of exp utility
tasumexpu = tsparse.dot(tmchainmat, taexpu)
# Individual within-chain choice probability (raw, without distortion)
tacondcpnodist = taexpu / tsparse.dot(tmchainmat.T, tasumexpu)
# Conditional choice probability, given chain choice
tacondcp = tathetaprod * tacondcpnodist + tahi * (1 - tathetaprod)
# Distortion in within-chain choice probability
tachoicedist = tacondcp - tacondcpnodist

# Inclusive values
taIVb = T.log(tasumexpu) #Both available
taIVhi = tsparse.dot(tchoosehigh, tautil) #Only high-margin available
# Exponentiated inclusive value
taexpI = T.exp(taldacol * (tathetacol * taIVb + (1 - tathetacol) * taIVhi))
tasumexpI = tsparse.dot(tchainsum, taexpI)
# Chain choice probability
tachaincp = taexpI / tsparse.dot(tchainsum.T, tasumexpI)
# Choice probability for chain-product combination, per node
tasimchoiceprob = tacondcp * tsparse.dot(tmchainmat.T, tachaincp)
# Choice probabilities
tacp = tasimchoiceprob.dot(tfw)

# Unconstrained theta
tthetauc = ~T.isclose(ttheta, 1)
# Index of unconstrained thetas/active FOC
tfocix = tthetauc.nonzero()[0]

# Profit
taprofit = tsparse.dot(tmchainmat, tamargin * tacp)

# Jacobians of FOC wrt all parameters
taJfoc, _ = theano.scan(
        fn=lambda i: 
                    T.grad(T.grad(taprofit[i], ttheta)[i],
                           [td, ttheta, tsigma, tmean, tlda, teta]),
        sequences=[tfocix]
    )


# Share equations, only derivatives matter for implicit function
# (at solution of system)
tashareq = tsparse.dot(tnobase, tacp)

#Jacobians of shares wrt all parameters
taJshareq = T.jacobian(tashareq, [td, ttheta, tsigma, tmean, tlda, teta])

# Indexing for relevant parts of system of equations
tsysix = T.concatenate([tnobaseix, tfocix + td.shape[0]])
         
# Jacobian of system of equations wrt delta and theta
taJdtsys = T.concatenate(
        [T.concatenate(taJshareq[:2], axis=1),
         T.concatenate(taJfoc[:2], axis=1)], axis=0
        )[:, tsysix]

# Jacobian of system of equations wrt parameters of preference heterogeneity
taJparsys = T.concatenate(
        [T.concatenate(taJshareq[2:], axis=1),
         T.concatenate(taJfoc[2:], axis=1)], axis=0
        )[:, tuseparix]

# Jacobian of delta and theta wrt parameters of preference heterogeneity
# (implicit function th, at solution of system)
taJdtpar = - tsolve(taJdtsys, taJparsys)

# Log likelihood
# Probability of individual choice sequences for each integration node
# (exp of summed log choiceprobs = product of choice probabilities in sequence)
tasimchoiceseqprob = T.exp(tsparse.dot(tchoicemat, T.log(tasimchoiceprob)))
# Integrate over unobserved heterogeneity
taindlik = tasimchoiceseqprob.dot(tfw)
taloglik = T.log(taindlik).sum()

# Gradient with respect to all parameters
taJloglik = T.grad(taloglik, [td, ttheta, tsigma, tmean, tlda, teta])

# Split by gradient wrt delta and theta, and parameters of preference heterogeneity
taJdtloglik = T.concatenate(taJloglik[:2])[tsysix]
taJparloglik = T.concatenate(taJloglik[2:])[tuseparix]
# Total gradient with respect to parameters of preference heterogeneity
# (implicit function th. to account for changes in delta and theta)
taJtotloglik = taJparloglik + taJdtloglik.dot(taJdtpar)

# Scores with respect to all parameters
tascoresall = T.jacobian(T.log(taindlik), [td, ttheta, tsigma, tmean, tlda, teta])
# Split
tadtscores = T.concatenate(tascoresall[:2], axis=1)[:, tsysix]
taparscores = T.concatenate(tascoresall[2:], axis=1)[:, tuseparix]
# Scores with respect to parameters of preference heterogeneity
tascores = taparscores + tadtscores.dot(taJdtpar)


###     BLP-type system of equations
# BLP choice probabilities
# Individual within-chain choice probability
tblpcondcp = texpu / town.T.dot(tsumexpu)
# Exponentiated inclusive values (per nest/chain)
tblpexpI = T.exp(tlda.dimshuffle([0, 'x']) * T.log(tsumexpu))
# Chain choice probability
tblpchaincp = tblpexpI / tblpexpI.sum(0)
# Product choice probability
tblpsimchoiceprob = tblpcondcp * town.T.dot(tblpchaincp)
# Aggrefate choice probability/market share
tblpchoiceprob = tblpsimchoiceprob.dot(tw)

# System of equations for market contractions
tblpsyseq = (tsobs - tblpchoiceprob)[1:]

# Jacobian of equations, analytic
townmat = town.T.dot(town)
tblpsimcpw = tblpsimchoiceprob * tw.dimshuffle(['x', 0])
tblpJcp0 = T.nlinalg.alloc_diag(tblpchoiceprob) / tldaprod - tblpsimcpw.dot(tblpsimchoiceprob.T)
tblpJcp1 = tblpsimcpw.dot((tblpcondcp * (1 - 1 / tldaprod)).T) * townmat
# summed within product type
tblpJcp = - (tblpJcp0 + tblpJcp1)[1:, 1:]


# Compile theno functions
msyseq = theano.function([ttheta,td,tulo,tuhi,tmu,tw,tlda,thi,town,tm,tsobs], tsyseq,
                         name = 'syseq')
Jmsyseq = theano.function([ttheta,td,tulo,tuhi,tmu,tw,tlda,thi,town,tm,tsobs], tJsyseq,
                          name = 'jacsyseq')
# BLP-type
mblpsyseq = theano.function([td,tmu,tw,tlda,town,tsobs], tblpsyseq, name = 'blpsyseq')
mJblpsyseq = theano.function([td,tmu,tw,tlda,town], tblpJcp, name = 'jacblpsyseq')

# Log likelihood function
mloglik = theano.function([td, tsigma, tmean, tlda, ttheta, teta],
                          [taloglik, taJtotloglik],
                          name='loglik')
# Score
mscores = theano.function([td, tsigma, tmean, tlda, ttheta, teta],
                         tascores, name='scores')

#%%    SETUP FUNCTIONS

#Handle error in contractions
class ContractionError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)

def createindparams(param):
    """
    Create parameters governing distribution of individual preferences based
    on array with stacked parameters.

    Parameters
    ----------
    param : ndarray (nmlparam,)
        Array of stacked parameters governing distrbution of individual preferences.

    Returns
    -------
    ugroupmean : pandas.Series (nugroup x nugroupparams,)
        Parameters shifting average utility for each characteristic for latent
        classes (first latent class is baseline with zeros).
    ugroupdist : pandas.Series (nugroup x nugroupparams,)
        Parameters governing dispersion of utility for each characteristic for
        latent classes.
    lda : ndarray (nchain,)
        Nesting parameter for each chain.
    eta : ndarray (nugroup - 1,)
        Parameters governing logit probability for latent classes (implicitly
        zero for first class).
    """
    j = 0
    # Utility dispersion for latent classes
    ndist = nugroupparams * nugroup
    ugroupdist = pd.Series(param[j:j + ndist],
                          index=ugroupdistix,
                          name='latclassvar')
    j = j + ndist
    # Mean utility shifts for latent classes
    nmean = nugroupparams * (nugroup - 1)
    meanpar = column_stack((zeros(nugroupparams),
                            param[j:j + nmean].reshape(nugroupparams, nugroup - 1)
                            ))
    ugroupmean = pd.Series(meanpar.flatten(),
                           index=ugroupdistix,
                           name='latclassmean')
    j = j + nmean
    # Nesting parameters
    lda = param[j:j + nchain]
    j = j + nchain
    # Probability of latent classes
    nprob = nugroup - 1
    eta = param[j:j + nprob]
    return ugroupmean, ugroupdist, lda, eta

def updateutildev(eta, ugroupmean, ugroupdist):
    """
    Create data on utility at each integration node from parameters governing
    distribution of individual preferences.

    Parameters
    ----------
    eta : ndarray (nugroup - 1,)
        Parameters governing logit probability for latent classes.
    ugroupmean : ndarray (nugroup x nugroupparams,)
        Parameters shifting average utility for each characteristic for latent
        classes (first latent class is baseline with zeros).
    ugroupdist : ndarray (nugroup x nugroupparams,)
        Parameters governing dispersion of utility for each characteristic for
        latent classes.

    Returns
    -------
    devutil : pandas.DataFrame (nmarket x nprods, nsim x nugroup)
        Utility deviations/over integration nodes x latent classes.
    weights : ndarray (nsim x nugroup)
        Weights for integrating over nodes and latent classes.
    """
    # Preference distribution
    prefdist = pd.DataFrame(repnodes * ugroupdist[:, None] + ugroupmean[:, None],
                            index=ugroupdistix, columns=simix)
    prefdist = (prefdist.unstack(1).swaplevel(axis=1).
                sort_index(axis=1).loc[varcols]
                )

    devutil = marketdata[varcols].dot(prefdist)
    #Normalize utility deviations per draw per market
    devutil = devutil - devutil.groupby(level='market').max()
    
    expgroup = hstack((1., exp(eta)))
    ugroupprob = expgroup / expgroup.sum()
    weights = hstack([baseweights * z for z in ugroupprob])
    return devutil, weights


# Functions for optimization
def partargs(arg, ntheta, ndelta):
    """
    Split stacked array of theta, delta and multipliers on bounds of theta into
    theta and multipliers

    Parameters
    ----------
    arg : ndarray (ntheta + ndelta + 2 * ntheta,)
        Stacked array of theta, delta and multipliers on bounds of theta.
    ntheta : int
        Number of thetas.
    ndelta : int
        Number of deltas.

    Returns
    -------
    theta : ndarray (ntheta,)
    ulo : ndarray (ntheta,)
        Lagrangian multipliers on lower bound of theta.
    uhi : ndarray (ntheta,)
        Lagrangian multipliers on upper bound of theta.
    """
    theta = arg[:ntheta]
    ulo = arg[ntheta + ndelta: 2 * ntheta + ndelta]
    uhi = arg[2 * ntheta + ndelta: 3 * ntheta + ndelta]
    return theta, ulo, uhi

def makedeltaarg(delta):
    """Add zero to delta (normalization of first product)"""
    return hstack((0, delta))

def createargs(arg, ntheta, ndelta):
    """
    Split stacked array of theta, delta and multipliers on bounds.

    Parameters
    ----------
    arg : ndarray (ntheta + ndelta + 2 * ntheta,)
        Stacked array of theta, delta and multipliers on bounds of theta.
    ntheta : int
        Number of thetas.
    ndelta : int
        Number of deltas.

    Returns
    -------
    theta : ndarray (ntheta,)
    delta : ndarray (ndelta,)
    ulo : ndarray (ntheta,)
        Lagrangian multipliers on lower bound of theta.
    uhi : ndarray (ntheta,)
        Lagrangian multipliers on upper bound of theta.
    """
    theta = arg[:ntheta]
    delta = makedeltaarg(arg[ntheta:ntheta + ndelta])
    ulo = arg[ntheta + ndelta: 2 * ntheta + ndelta]
    uhi = arg[2 * ntheta + ndelta: 3 * ntheta + ndelta]
    return theta, delta, ulo, uhi

def cobj(arg, ntheta, ndelta):
    """Contraction objective (sum of complementary slackness)"""
    theta, ulo, uhi = partargs(arg, ntheta, ndelta)
    return array([(ulo * theta + uhi * (1 - theta)).sum()])

def gcobj(arg, ntheta, ndelta):
    """Gradient of contraction objective"""
    theta, ulo, uhi = partargs(arg, ntheta, ndelta)
    return hstack((ulo - uhi, zeros(ndelta), theta, 1 - theta))

def csyseq(arg, ntheta, ndelta, mmu, mw, lda, maxmargin, own, margin, sobs):
    """Wrapper for system of equations (theano), taking a stacked array of
    theta, delta and multipliers as argument.
    Used as constraint on inner loop solution.
    """
    theta, delta, ulo, uhi = createargs(arg, ntheta, ndelta)
    return msyseq(theta, delta, ulo, uhi, mmu, mw, lda, maxmargin, own, margin, sobs)

def Jcsyseq(arg, ntheta, ndelta, mmu, mw, lda, maxmargin, own, margin, sobs):
    """Wrapper for Jacobian of system of equations (theano), taking a stacked 
    array of theta, delta and multipliers as argument. Adds derivatives wrt
    multipliers.
    Used in constraint on inner loop solution.
    """
    theta, delta, ulo, uhi = createargs(arg, ntheta, ndelta)
    tdjac = Jmsyseq(theta, delta, ulo, uhi, mmu, mw, lda, maxmargin, own, margin, sobs)
    tdjac = tdjac
    # Add derivatives wrt multipliers
    zeropad = len(tdjac) - ntheta
    ujac = row_stack([
        column_stack([eye(ntheta), - eye(ntheta)]),
        zeros((zeropad, 2 * ntheta))
        ])
    return column_stack([tdjac, ujac])

def blpsyseq(delta, mmu, mw, lda, own, sobs):
    """System of equations with theta = 1 (implicitly)."""
    delta = makedeltaarg(delta)
    return mblpsyseq(delta, mmu, mw, lda, own, sobs)

def jacblpsyseq(delta, mmu, mw, lda, own, sobs):
    """Jacobian of system of equations with theta = 1 (implicitly)."""
    delta = makedeltaarg(delta)
    return mJblpsyseq(delta, mmu, mw, lda, own)

def createguess(ndelta, ntheta=None, logitdelta=None):
    """Create guess for inner loop parameters."""
    if logitdelta is None:
        logitdelta = zeros(ndelta)
    else:
        logitdelta = logitdelta[-ndelta:]
    if ntheta is None:
        return logitdelta
    else:
        return hstack((ones(ntheta) * 0.5, logitdelta, zeros(ntheta), zeros(ntheta)))

def solvemarketeq(ntheta, ndelta, mmu, mw, lda, maxmargin, own, margin, sobs,
                  logitdelta=None, guess=None, jac=True, fullret=False,
                  retry=True, maxiter=200, ftol=1e-14):
    """Find parameters solving inner loop system of equations for a given market
    configuration."""
    if guess is None:
        guess = createguess(ndelta, ntheta, logitdelta)
    bounds = [(0, 1)] * ntheta # theta bounds
    bounds.extend([(None, None)] * ndelta) # delta unbounded
    bounds.extend([(0, None)] * 2 * ntheta) # non-negative multipliers on theta bounds
    consdict = {'type': 'eq',
                'fun': lambda z: csyseq(z, ntheta, ndelta, mmu, mw, lda,
                                        maxmargin, own, margin, sobs)}
    if jac:
        consdict['jac'] = lambda z: Jcsyseq(z, ntheta, ndelta, mmu, mw, lda,
                                            maxmargin, own, margin, sobs)
    
    #Define objective and jacobian
    objective = lambda z: cobj(z, ntheta, ndelta)
    gobjective = lambda z: gcobj(z, ntheta, ndelta)
    optdict = {'maxiter': maxiter, 'ftol': ftol}
    start = dtim()
    res = minimize(objective, guess, jac=gobjective, bounds=bounds,
                   constraints=(consdict), options=optdict)

    sysmaxval = np.abs(consdict['fun'](res.x)).max()
    retried = 0
    if retry and (not res.success or (res.fun > 1e-10) or (sysmaxval > 1e-10)):
        guess = createguess(ndelta, ntheta)
        res = minimize(objective, guess, jac=gobjective, bounds=bounds,
                       constraints=(consdict), options=optdict)

        sysmaxval = np.abs(consdict['fun'](res.x)).max()
        retried += 1
        if not res.success or (res.fun > 1e-10) or (sysmaxval > 1e-10):
            tscale = 1
            for i in range(150):
                guess = hstack((rand(ntheta) * tscale, (i + 1) * 0.02 * randn(ndelta), rand(ntheta), rand(ntheta)))
                res = minimize(objective, guess, jac=gobjective, bounds=bounds,
                               constraints=(consdict), options=optdict)

                sysmaxval = np.abs(consdict['fun'](res.x)).max()
                retried += 1
                if res.success & (res.fun < 1e-10) & (sysmaxval < 1e-10):
                    break
    duration = dtim() - start
    # Return tuple with information (add extra diagnostics if fullret)
    if fullret:
        ret = createargs(res.x, ntheta, ndelta), res, duration, retried, sysmaxval
    else:
        ret = createargs(res.x, ntheta, ndelta)
    return ret

def solveblpeq(mmu, mw, lda, own, sobs, logitdelta=None, guess=None,
               jac=False, fullret=False, retry=True):
    """Find deltas equalizing observed and predicted shares (theta implicitly 1)"""
    ndelta = len(sobs) - 1
    if guess is None:
        guess = createguess(ndelta, None, logitdelta)
    consdict = {'type': 'eq',
                'fun': lambda z: blpsyseq(z, mmu, mw, lda, own, sobs)}
    if jac:
        consdict['jac'] = lambda z: jacblpsyseq(z, mmu, mw, lda, own, sobs)
    
    optdict = {'maxiter': 200, 'ftol': 1e-14}
    start = dtim()
    res = minimize(lambda z: 0 * z.sum(), guess, jac=lambda z: zeros(len(z)),
                   constraints=(consdict), options=optdict)

    sysmaxval = np.abs(blpsyseq(res.x, mmu, mw, lda, own, sobs)).max()
    retried = 0
    if retry and (not res.success or (sysmaxval > 1e-10)):
        for i in range(500):
            guess = i * 0.02 * randn(ndelta)
            res = minimize(lambda z: 0 * z.sum(), guess, jac=lambda z: zeros(len(z)),
                           constraints=(consdict), options=optdict)
            sysmaxval = np.abs(blpsyseq(res.x, mmu, mw, lda, own, sobs)).max()
            retried += 1
            if res.success & (sysmaxval < 1e-10):
                break

    duration = dtim() - start

    if fullret:
        ret = hstack((0., res.x)), res, duration, retried, sysmaxval
    else:
        ret = hstack((0., res.x))
    return ret

def marketsolve(market, mu, mw, lda, fullret=False, retry=True, jac=True,
                logitdelta=False, blpdelta=True, guess=None,
                maxiter=200, ftol=1e-14):
    """Find parameters solving inner loop system of equations for a given market.
    Wrapper for solvemarketeq."""
    # Number of active chains
    mnchain = marketdata.loc[market].chainid.nunique()
    # Other data
    margin = marketdata.loc[market].margin.fillna(0).values
    maxmargin = marketdata.loc[market].ismaxoption.values
    own = marketdata.loc[market][chaincols].values.T
    sobs = marketdata.loc[market].share.values

    # Number of theta and delta to optimize over
    ntheta = mnchain
    ndelta = len(sobs) - 1

    # Utility deviations for market
    mmu = mu.loc[market].values
    
    # Initial values
    if logitdelta:
        lgd = marketdata.alphalogit.loc[market].values
    else:
        if blpdelta:
            lgd = solveblpeq(mmu, mw, lda, own, sobs, jac=True,
                             logitdelta=None)
        else:
            lgd = None
    return solvemarketeq(ntheta, ndelta, mmu, mw, lda, maxmargin, own, margin, sobs,
                         logitdelta=lgd, guess=guess, jac=jac, fullret=fullret,
                         retry=retry, maxiter=maxiter, ftol=ftol)


def contraction(mu, mw, lda, verbose=False, blpguess=True):
    """Find parameters solving inner loop system of equations for each market."""
    theta, delta, ulo, uhi = [np.empty(0)] * 4
    duration = np.empty(nmarket)
    solvestatus = np.empty(nmarket)
    funval = np.empty(nmarket)
    retries = np.empty(nmarket)
    i = 0
    for m in marketlist:
        if verbose:
            print("Contracting market {:d}".format(m))
        ret = marketsolve(m, mu, mw, lda, fullret=True, blpdelta=blpguess)
        rtheta = ret[0][0]
        rdelta = ret[0][1]
        rulo = ret[0][2]
        ruhi = ret[0][3]
        if np.isnan(ret[1].x).any() or not np.isfinite(ret[1].x).any():
            raise ContractionError('Invalid values in contraction')
        theta = hstack((theta, rtheta))
        delta = hstack((delta, rdelta))
        ulo = hstack((ulo, rulo))
        uhi = hstack((uhi, ruhi))
        solvestatus[i] = ret[1].success
        funval[i] = np.max([np.abs(ret[1].fun).max(), ret[4]])
        duration[i] = ret[2]
        retries[i] = ret[3]
        if verbose:
            print("Duration: {:.2f}".format(duration[i]))
            print("Success? ", ret[1].success)
        i += 1
    theta = pd.DataFrame({'theta': theta, 'ulo': ulo, 'uhi': uhi}, index=thetaix)
    delta = pd.Series(delta, index = marketdata.index, name = 'meanutil')
    diagnostic = pd.DataFrame({'status': solvestatus, 'duration': duration,
                               'funval': funval, 'retry': retries}, index=marketlist)
    return theta, delta, diagnostic


def objective(parameter, blpguess=True):
    """Function for maximizing log likelihood."""
    start = dtim()
    print("Current parameter value: ", parameter)
    ugroupmean, ugroupdist, lda, eta = createindparams(parameter)
    mu, mw = updateutildev(eta, ugroupmean, ugroupdist)
    try:
        theta, delta, diagnostic = contraction(mu, mw, lda, blpguess=blpguess)
    except ContractionError as error:
        print(error)
        return 1e12, - np.sign(parameter) * 2500
    contractduration = diagnostic.duration.sum()
    print(f"Time in contraction: \t {contractduration:.1f} seconds")
    nonsuccess = (1 - diagnostic.status).sum()
    print(f"Number of unsuccessful contractions: \t {nonsuccess}")
    maxfunval = diagnostic.funval.max()
    print(f"Maximum funval in contraction: \t {maxfunval}")
    thetanuc = (~np.isclose(theta.theta, 1)).sum()
    print(f"Number of theta below one (0.99): \t {thetanuc}")
    
    ll, llgrad = mloglik(delta, ugroupdist, ugroupmean, lda, theta.theta, eta)
    
    print("Seconds in likelihood loop: \t %.1f" % (dtim() - start))
    print(f"Current log likelihood: \t {ll:.8f}")
    print(" ")
    return -ll, -llgrad

def hess_num(parameter, twosided=True):
    """Calculates the Hessian by finite differences of the gradient.
    Returns the negative of the Hessian (objective is negative of log likelihood)."""
    eps = 1e-6
    hess = zeros((nmlparam, nmlparam))
    if not twosided:
        _, grad = objective(parameter)
    for i in range(nmlparam - 22):
        par_hi = parameter.copy()
        par_hi[i] += eps
        _, grad_hi = objective(par_hi)
        if twosided:
            par_lo = parameter.copy()
            par_lo[i] -= eps
            _, grad_lo = objective(par_lo)
            hess[i] = (grad_hi - grad_lo) / (2 * eps)
        else:
            hess[i] = (grad_hi - grad) / eps
    return hess

def calccovmat(parameter, sandwich=True):
    """Calculate covariance matrix of estimator. Can use "sandwich" formula."""
    nhess = hess_num(parameter)
    invnhess = np.linalg.inv(nhess)
    if sandwich:
        ugroupmean, ugroupdist, lda, eta = createindparams(parameter)
        mu, mw = updateutildev(eta, ugroupmean, ugroupdist)
        theta, delta, _ = contraction(mu, mw, lda)
        scores = mscores(delta, ugroupdist, ugroupmean, lda, theta.theta, eta)
        ops = scores.T @ scores
        covmat = invnhess @ ops @ invnhess
    else:
        covmat = invnhess
    return covmat

def ugroupprobest(ugroupparam, covmat):
    """
    Create latent class probabilities and standard errors (delta method).
    """
    # Calculate probabilities
    ugroupprob = hstack((1, exp(ugroupparam))) / hstack((1, exp(ugroupparam))).sum()
    # Relevant jacobian of probabilities wrt to parameter
    Jugroupprob = diag(ugroupprob[1:]) - ugroupprob[1:, None].dot(ugroupprob[None,1:])
    # Covariance matrix for probabilities
    probcovmat = Jugroupprob.dot(covmat).dot(Jugroupprob.T)
    # Standard errors of probabilities
    probse = np.sqrt(probcovmat.diagonal())
    return ugroupprob, probse


def maketable(parameter, ll, caption=None, key=None, numdec=2, sandwich=True):
    """
    Create table of estimated model parameters.

    parameter is the vector of estimated parameters
    ll is the log likelihood of the objective function
    caption specifies table caption (optional)
    key specifies key for table reference (optional)
    numdec specifies the number of decimals to use
    sandwich specifies whether covmat is calculated with sandwich formula
    """
    #Index for parameters in LaTeX table
    paramindex = hstack(( [r'$\tau_g(\boldsymbol{\eta})$', r'$\eta_g$'],
                  r'$\delta_p^g$',
                  [rf'$\delta_{i}^g$' for i in range(2, nchain + 1)],
                  r'$\sigma_p^g$',
                  [rf'$\sigma_{i}^g$' for i in range(2, nchain + 1)]))
    #Column header for latent classes in LaTeX table
    groupindex = [rf'$g = {i}$' for i in range(nugroup)]
    
    #Covariance matrix of parameter estimates
    covmat = calccovmat(parameter, sandwich=sandwich)
    #Standard errors
    se = np.sqrt(covmat.diagonal())
    #Latent class probabilities and standard errors
    ugroupprob, ugroupprobse = ugroupprobest(parameter[-(nugroup - 1):], covmat[-(nugroup - 1):, -(nugroup - 1):])

    #Make array of parameters and std.err. in format appropriate for table generation
    parmat = row_stack([ugroupprob, hstack((0, parameter[-(nugroup - 1):])),
                        hstack((zeros(nugroupparams),
                                parameter[(nugroup * prefdistdim):((2*nugroup - 1) * prefdistdim)])).reshape((nugroup, prefdistdim)).T,
                        parameter[:(nugroup * prefdistdim)].reshape((nugroup, prefdistdim)).T])

    semat = row_stack([hstack((0, ugroupprobse)), hstack((0, se[-(nugroup - 1):])),
                       hstack((zeros(nugroupparams),
                               se[(nugroup * prefdistdim):((2*nugroup - 1) * prefdistdim)])).reshape((nugroup, prefdistdim)).T,
                       se[:(nugroup * prefdistdim)].reshape((nugroup, prefdistdim)).T])

    #Format table to text with number formatting
    semat = array([['$({:.2f})$'.format(x) for x in z] for z in semat])
    #Remove some of the standard errors for non-estimated parameters
    semat[0, 0] = '--'
    semat[parmat == 0] = '--'

    parmat = array([['${:.2f}$'.format(x) for x in z] for z in parmat])
    #Add on parameter labels to parameter matrix and empty column to std.err matrix
    parmat = column_stack([paramindex, parmat])
    semat = column_stack([[''] * len(paramindex), semat])

    #Join matrices to one
    tablemat = hstack([row_stack([parmat[i], semat[i]]).T for i in range(len(parmat))]).T
    tablerows = ['\t&\t'.join(x) + r' \\ ' + '\n' for x in tablemat]

    #Construct table
    table = r'\begin{table}[h]' + '\n'
    table += r'\centering' + '\n'

    #Make caption
    if caption is not None:
        captionstr = r'\caption{' + caption
        if key is not None:
            captionstr += r' \label{' + key + r'}'
        captionstr += r'}' + '\n'
        table += captionstr

    table += r'\begin{tabular}{l' + 'c' * nugroup + r'}' + '\n'
    table += r'\toprule' + '\n'
    table += '\t&\t' + '\t&\t'.join(groupindex) + r' \\' + ' \n'
    table += r'\midrule' + '\n'

    #Construct table body
    for i in range(len(tablerows)):
        table += tablerows[i]
        if i%2 == 1:
            table += r'\addlinespace' + '\n'

    #Statistics
    table += r'\midrule' + '\n'
    for i in range(nchain):
        table += r'$\lambda_{:d}$ &'.format(i + 1) + '\t'
        table += r'\multicolumn{%d}{c}{' % nugroup + '${:.2f}$'.format(parameter[-(nchain + nugroup - 1):-(nugroup - 1)][i]) + r'} \\ ' + '\n'
        table += '\t\t' + r'&'
        table += r'\multicolumn{%d}{c}{' % nugroup + '$({:.2f})$'.format(se[-(nchain + nugroup - 1):-(nugroup - 1)][i]) + r'} \\ ' + '\n'
        table += r'\addlinespace' + '\n'
    table += r'$\ln \mathcal{L}(\hat{\beta})$ &' + '\t'
    table += r'\multicolumn{%d}{c}{' % nugroup + '${:,.0f}$'.format(ll) + r'} \\ ' + '\n'
    table += r'$N$ &' + '\t'
    table += r'\multicolumn{%d}{c}{' % nugroup + '${:,.0f}$'.format(50000) + r'} \\ ' + '\n'
    #End of table
    table += r'\bottomrule' + '\n'
    table += r'\end{tabular}' + '\n'
    table += r'\end{table}'

    return table



#%% ESTIMATE MODEL AND STORE RESULTS
parguess = array([-3.25017361e-02,  3.52432130e+00,  3.24821916e+00,
                  4.17779558e-03,  7.07481951e+00,  3.77455283e+00,
                  6.17723402e-01,  8.56638406e+00,  3.21854851e+00,
                  3.84207774e-01,  2.27708323e+00,  2.14601547e+00,
                  1.43655433e-01,  4.67030915e+00, -1.46370919e+00,
                  3.66990252e-02,  1.51967772e+00,  7.12389661e+00,
                  5.25897917e-02, -4.90166393e+00, -4.36690930e+00,
                  2.78923114e-01,  6.78998660e-01,  6.25781574e-01,
                  1.43255248e+00,  1.60291909e+00,  1.77723490e+00])

print('Estimating model')
res = minimize(objective, parguess, jac=True)

print('Finished iteration')
print(f'Number of latent classes: \t {args.nugroup}')
print(f'Number of quadrature nodes: \t {args.quadnodes}')
print(f'Current parameter value: \t {res.x}')
print(f'Success: \t {res.success}')
print(res)

#Store results object
pickle.dump(res, open(f'{resultdir}demand_res{args.filepostfix}.pkl', 'wb'))

# Store inner loop parameters
parameter = res.x
ugroupmean, ugroupdist, lda, eta = createindparams(parameter)
mu, mw = updateutildev(eta, ugroupmean, ugroupdist)

theta, delta, diagnostic = contraction(mu, mw, lda)

delta.to_pickle(f'{resultdir}delta{args.filepostfix}.pkl')
theta.to_pickle(f'{resultdir}theta{args.filepostfix}.pkl')

# Store parameter guess
with open(f'{resultdir}parguess{args.filepostfix}.txt', 'w') as f:
    parguess.tofile(f, sep=';')



#%% TABLE 4.3
# Estimation table
tabcaption = 'Parameter estimates for consumer choice model with supply constraints'
tabkey = 'tbl:demand'

regtable = maketable(parameter, -res.fun, caption=tabcaption, key=tabkey)

print(regtable)


with open(f'{tabfigdir}demandtable{args.filepostfix}.tex', 'w') as f:
    f.write(regtable)

