#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Morten Sæthre

Paper: "On the Role of Parallel Trade on Manufacturers and Retailers Profits in the Pharmaceutical Sector"
Authors: Pierre Dubois (TSE) and Morten Saethre (NHH)

Create dataset of margins and sales for a larger set of drugs.
Used to perform extended reduced form analysis.
"""

#%% SETUP

import numpy as np
import pandas as pd
import mkl

mkl.set_num_threads(mkl.get_max_threads())

# Directories
rootdir = '/home/morten/'
projectroot = f'{rootdir}projects/parallel/'
datadir = f'{projectroot}data/'



#%% LOAD DATA

### Load information on each drug ID
druginfo = pd.read_pickle(f'{datadir}druginfo.pkl')
druginfo.drop(druginfo[druginfo.drugid.isnull()].index, inplace=True)
druginfo['drugid'] = druginfo.drugid.astype(int)

# Drop erroneous entries
dropmask = (druginfo.drugid == 460964) & (druginfo.varenavn.str.contains('COAPROVEL'))
druginfo.drop(druginfo[dropmask].index, inplace=True)

# Load parallel import license data
parlic = pd.read_pickle(f'{datadir}parallelmktlicense.pkl')
parlic.rename(columns={'varenummer': 'drugid'}, inplace=True)

# Load drug company name
drugcompany = pd.read_pickle(f'{datadir}drugcompany.pkl')
drugcompany.name = 'firmname'
firmdict = {'Farlic': 'Farmagon', 'Abacist Pharma': 'Abacist',
            'AstreZeneca': 'AstraZeneca', 'H. Lundbeck': 'Lundbeck',
            'Ares-Serono': 'Merck', 'Serono': 'Merck', 'NM Pharma': 'Mylan',
            'GEA Farmaceutisk Fabrik': 'Sandoz', 'Hexal': 'Sandoz',
            'Pliva Pharma': 'Pliva', 'ALK-Abello': 'ALK',
            'ACS Dobfar Generics': 'ACS Dobfar', 'Antula Healthcare': 'Antula',
            'B. Braun Melsungen': 'Braun', 'Besins International': 'Besins',
            'Grünenthal': 'Grunenthal', 'Gebro Pharma': 'Gebro',
            'G. Pohl-Boskamp': 'Pohl-Boskamp', 'Martindale Pharmaceuticals': 'Martindale',
            'Institut Produits Synthese': 'Ipsen',
            'Laboratoires Théa S.A.S': 'Novartis', 'Otsuka Pharmaceutical': 'Otsuka',
            'Sanofi Pharma Bristol-Myers Squibb': 'Sanofi-Aventis'}
drugcompany = drugcompany.replace(firmdict)

# Fix erroneous entries
drugcompany.loc[386706] = 'Paranova'
drugcompany.loc[[23548, 18807]] = 'Arrow'

# Load GIP (wholesale) data
gipdata = pd.read_csv(f'{datadir}gip_all.csv')

# Load additional drug ID information
drugdesc = pd.read_stata(f'{datadir}beskrivelse_varenummer.dta')
drugdesc.rename(columns={'varenummer': 'drugid', 'mt_comp': 'firmname'}, inplace=True)
firmdict = {'NYCOMED PHARMA': 'NYCOMED', 'GLAXOSMITHKLIN': 'GLAXOSMITHKLINE',
            'MERCK NM': 'MYLAN', 'SCHERING-PLOUGH': 'SCHERING',
            'B-MYERS SQUIBB': 'BRISTOL-MYERS SQUIBB', 'MERCK SERONO': 'MERCK',
            'ALPHARMA AS': 'ALPHARMA', 'FARMAGON AS': 'FARMAGON',
            'DN ETERFABRIKK': 'DEN NORSKE ETERFABRIKK',
            'BOEHRINGER ING': 'BOEHRINGER INGELHEIM', 'DUMEX-ALPHARMA': 'ALPHARMA',
            'ORION PHARMA': 'ORION', 'BAYER SCHERING': 'BAYER',
            'ALK, ALLERGOLOG': 'ALK', 'ACS DOBFAR GEN': 'ACS DOBFAR',
            'LAB. BESINS': 'BESINS', 'BIOGEN IDEC': 'BIOGEN',
            'CHAUVIN PHARMA': 'CHAUVIN', 'FARLIC': 'FARMAGON',
            'FERROSAN AS': 'FERROSAN', 'FRESENIUS KABI': 'FRESENIUS',
            'PHARMA NORD AS': 'PHARMA NORD', 'NORDIC DRUGS A': 'NORDIC DRUGS',
            'SEARLE': 'PFIZER'}
drugdesc['firmname'] = drugdesc.firmname.replace(firmdict)

# Load regulatory data
regdata = pd.read_pickle(f'{datadir}SLVregulation.pkl')


#%% Subsetting
# Restrict to ATC codes in GIP sample (200 most important ATCs)
gipatc = gipdata.atccode.unique()

druginfo.drop(druginfo[~druginfo.atccode.isin(gipatc)].index, inplace=True)
parlic.drop(parlic[~parlic.atckode.isin(gipatc)].index, inplace=True)

# ATC codes with parallel imports registered
paratc = parlic.atckode.unique()

druginfo = druginfo.join(drugcompany, on='drugid')
druginfo = druginfo.join(drugdesc.set_index('drugid')[['pi', 'generic_code', 'firmname']].
                         rename(columns={'firmname': 'firmname2'}), on='drugid')

# List of parallel importing companies
parfirm = ['Farmagon', 'Orifarm', 'Euromedica', 'Paranova', 'Sigma', 'Abacist',
           'Cross Pharma', 'Arrow']

# Dictionary with parallel firm substitutions in drug name
parfirmdict = {'Abacist': ['abacist', 'abac'],
               'Arrow': ['arrow', 'arro', 'arr'],
               'Euromedica': ['euro', 'eurom', 'euromed', 'euromedica'],
               'Cross Pharma': ['cross'],
               'Farmagon': ['farmagon', 'farmag', 'farma', 'farm', 'farlic', 'farl', 'far'],
               'Orifarm': ['orifarm', 'orif', 'ori'],
               'Paranova': ['paranova', 'para', 'par'],
               'Sigma': ['sigma']
               }

def searchfirm(firm, firmdict):
    if not firm.istitle():
        firm = firm.capitalize()
    return ' ' + ' | '.join(firmdict[firm]).upper() + ' '

# Parallel if parallel firm in package name
for firm in parfirmdict.keys():
    mask = druginfo.varenavn.str.contains(searchfirm(firm, parfirmdict))
    druginfo['firmname'] = druginfo.firmname.where(~mask, firm)
    
# Fix some parallel importer drug names
namedict = {'PAR': 'Paranova', 'PARA': 'Paranova', 'PARANOVA': 'Paranova',
            'FARLIC': 'Farmagon', 'FARMAGON': 'Farmagon', 'FARMAG': 'Farmagon',
            'FARMA': 'Farmagon', 'FARM': 'Farmagon', 'FARL': 'Farmagon',
            'FAR': 'Farmagon',
            'ORIFARM': 'Orifarm', 'ORIF': 'Orifarm', 'ORI': 'Orifarm',
            'EUROMED': 'Euromedica', 'EUROM': 'Euromedica',
            'EURO': 'Euromedica', 'EUROMEDICA': 'Euromedica',
            'SIGMA': 'Sigma',
            'ABAC': 'Abacist', 'ABACIST': 'Abacist',
            'CROSS': 'Cross Pharma'}

for k in namedict.keys():
    mask = druginfo.varenavn.str.contains(' {} '.format(k)) & druginfo.firmname.isnull()
    druginfo['firmname'] = druginfo.firmname.where(~mask, namedict[k])

# Fix some generic codes
genfirm = ['Actavis', 'Alpharma', 'Copyfarm', 'Sandoz', 'Ratiopharm', 'Teva', 'Mylan',
           'Hexal', 'Ranbaxy', 'NM Pharma', 'NMPharma']
genfirmre = '|'.join(genfirm).upper()

# Dictionary with generic firm substitutions in drug name and helper function
genfirmdict = {'Actavis': ['actavis', 'actav', 'acta', 'act', 'ac'],
               'Alpharma': ['alph', 'alpharma'],
               'Copyfarm': ['copyfarm', 'copyf', 'copy', 'cop'],
               'Sandoz': ['sa', 'san', 'sandoz', 'sand', 'biochemie', 'biochem', 'bioc'],
               'Mylan': ['nmparhma', 'nmpharm', 'nm', 'mercknm', 'mylan', 'myl'],
               'Hexal': ['hex', 'hexa', 'hexal'],
               'Pliva': ['pliva']
               }

# Generic if generic firm in package name
for firm in genfirmdict.keys():
    mask = druginfo.varenavn.str.contains(searchfirm(firm, genfirmdict))
    druginfo['firmname'] = druginfo.firmname.where(~mask, firm)

# Insert firmname2 if firmname is still missing
druginfo['firmname'] = druginfo.firmname.where(druginfo.firmname.notnull(),
                                               druginfo.firmname2.str.capitalize())



#%% Load wholesale data in other countries

imsdata = pd.read_feather(f'{datadir}imsdata_prepared.feather')



#%% Load sales data (from full transaction data)

salesdata = pd.read_feather(f'{datadir}sales_all.feather')

# Keep main chains
salesdata.drop(salesdata[~salesdata.chainid.isin(np.arange(1, 4))].index, inplace=True)

salesdrugs = salesdata.drugid.unique()



#%% Prepare regulatory data

coldict = {'Vare-nummer': 'drugid', 'Periode': 'period',
           'Styrke': 'strength', 'Legemiddelform': 'drugform',
           'Mengde per beholder': 'unitamount', 'Måle-enhet': 'measureunit',
           'Bytte-gruppe-kode': 'substitutioncode'}

regdata.rename(columns=coldict, inplace=True)
regdata['strength'] = regdata.strength.str.replace(',', '.').str.lower()
#Remove spaces in numbers
regdata['strength'] = regdata.strength.str.replace('(\d)\s(\d)', r'\1\2')
regdata['strengthval'] = pd.to_numeric(regdata.strength.str.
       extract('(\d*\.?\d*)')[0], 
       errors='coerce')

drugstrength = pd.Series(index=salesdrugs, name='strength')

# Milligram strength
mask = regdata.drugid.isin(salesdrugs) & regdata.strength.str.contains('mg')
mgdrugs = regdata[mask].drugid.unique()
nvals = regdata[mask].groupby('drugid').strengthval.nunique()
uniquedrugs = nvals[nvals == 1].index
mask = mask & regdata.drugid.isin(uniquedrugs)
tempstrength = regdata[mask].groupby('drugid').strengthval.mean().round(7)
drugstrength.update(tempstrength)

# Microgram strength
mask = (regdata.drugid.isin(salesdrugs) & ~regdata.drugid.isin(mgdrugs) & 
        regdata.strength.str.contains('mcg|mikrog')
        )
mcgdrugs = regdata[mask].drugid.unique()
nvals = regdata[mask].groupby('drugid').strengthval.nunique()
uniquedrugs = nvals[nvals == 1].index
mask = mask & regdata.drugid.isin(uniquedrugs)
tempstrength = regdata[mask].groupby('drugid').strengthval.mean().round(7) / 1000
drugstrength.update(tempstrength)

mask = (regdata.drugid.isin(salesdrugs) & ~regdata.drugid.isin(mgdrugs) & 
        ~regdata.drugid.isin(mcgdrugs) & 
        regdata.strength.str.contains('g$|g/')
        )
gdrugs = regdata[mask].drugid.unique()
nvals = regdata[mask].groupby('drugid').strengthval.nunique()
uniquedrugs = nvals[nvals == 1].index
mask = mask & regdata.drugid.isin(uniquedrugs)
tempstrength = regdata[mask].groupby('drugid').strengthval.mean().round(7) * 1000
drugstrength.update(tempstrength)

# Add strength onto drug info
druginfo = druginfo.join(drugstrength, on='drugid')

#%% Continue with preparation of transaction data

druginfo.drop(druginfo[~druginfo.drugid.isin(salesdrugs)].index, inplace=True)
drugdesc.drop(drugdesc[~drugdesc.drugid.isin(salesdrugs)].index, inplace=True)
gipdata.drop(gipdata[~gipdata.drugid.isin(salesdrugs)].index, inplace=True)

druginfo = druginfo.drop_duplicates('drugid')
druginfo.set_index('drugid', inplace=True)
drugdesc.set_index('drugid', inplace=True)

druginfo['firmname'] = druginfo.firmname.fillna(drugdesc.firmname.str.title())
druginfo.rename(columns={'varenavn': 'drugname'}, inplace=True)

salesdata = salesdata.join(druginfo[['firmname', 'pi', 'strength', 'drugname']], on='drugid')

typereplace = {'BGN': 'GEN', 'ORGWG': 'ORG', 'OTH': 'OTC'}
prodtype = druginfo.generic_code.replace(typereplace).rename('prodtype')

salesdata = salesdata.join(prodtype, on='drugid')

# Only keep relevant fields, and define index to prepare for reindexing
gipixcols = ['date', 'chainid', 'drugid']

# Join wholesale prices
salesdata = salesdata.join(gipdata.set_index(gipixcols).gip, on=gipixcols)

salesdata['strength'] = salesdata.strength.fillna(0)
salesdata['prodtype'] = salesdata.prodtype.fillna('UNKNOWN')
salesdata['pi'] = salesdata.pi == 1

salesdata['giprevenue'] = salesdata.gip * salesdata.packages
salesdata['giprevenueddd'] = salesdata.giprevenue.where(salesdata.ddd.notnull())
salesdata['gipddd'] = salesdata.ddd * salesdata.gip.notnull()
# Adjust strength for numerical errors
salesdata['strength'] = salesdata.strength.round(8)

# Aggregation
aggcols = ['date', 'chainid', 'atccode', 'strength', 'prodtype', 'pi']
agguse = ['revenue', 'maxrevenue', 'maxrevenueddd', 'packages', 'ddd',
          'maxddd', 'giprevenue', 'giprevenueddd', 'gipddd']

aggdata = salesdata.groupby(aggcols)[agguse].sum()

aggdata.reset_index(inplace=True)
aggdata['price'] = aggdata.revenue / aggdata.ddd
aggdata['maxprice'] = aggdata.maxrevenueddd / aggdata.maxddd.where(aggdata.maxddd != 0)
aggdata['gip'] = aggdata.giprevenueddd / aggdata.gipddd.where(aggdata.gipddd != 0)
# Aggregate IMS data
imsdata['ddd'] = imsdata.quantity * imsdata.strength / imsdata.dddvalue
imsdataaggcols = ['date', 'country', 'atccode', 'strength', 'prodtype']
imsagguse = ['sales', 'sales_pub', 'sales_trd', 'quantity', 'ddd']

imsagg = imsdata.groupby(imsdataaggcols)[imsagguse].sum()
imsagg['price'] = imsagg.sales / imsagg.ddd.where(imsagg.ddd != 0)

imsagg.reset_index(inplace=True)

imsprices = (imsagg[imsagg.prodtype == 'ORG'].
             set_index(['atccode', 'strength', 'country']).
             apply(lambda x: x.set_index('date').price.resample('MS').bfill()).
             unstack()
             )

imsprices.columns = [f'price_{c.lower()}' for c in imsprices.columns]

parpres = aggdata.merge(imsprices, how='left', right_index=True, left_on=['atccode', 'strength', 'date'])

dropcols = ['revenue', 'maxrevenue', 'maxrevenueddd', 'maxddd', 'giprevenue',
            'giprevenueddd', 'gipddd']

parpres.drop(dropcols, axis=1, inplace=True)
parpres['maxprice'] = parpres.maxprice.where(parpres.maxprice != 0)
parpres.to_csv(f'{datadir}parallel_presence.csv', index=False)

salesdata = salesdata.groupby(['atccode', 'strength']).filter(lambda x: x.pi.sum() > 0)

salesdata = salesdata.join(gipdata, on=gipixcols)
salesdata['giprevenue'] = salesdata.gip * salesdata.packages

salesdata['price'] = salesdata.revenue / salesdata.ddd
salesdata['maxprice'] = salesdata.maxrevenue / salesdata.maxddd

# Add on export country
sourcetemp = pd.get_dummies(parlic.exportcountry, prefix='src').groupby(parlic.drugid).max()
sourcetemp = sourcetemp.loc[salesdata.drugid.unique()]
sourcetemp = sourcetemp.loc[:, sourcetemp.sum() > 0]

salesdata = salesdata.join(sourcetemp, on='drugid')

#%% Aggregate data

agggroup = ['atccode', 'strength', 'date', 'chainid', 'firmname']
aggdata = salesdata.groupby(agggroup)[['revenue', 'maxrevenue', 'giprevenue', 'ddd', 'maxddd']].sum()

aggdata['price'] = aggdata.revenue / aggdata.ddd
aggdata['maxprice'] = aggdata.maxrevenue / aggdata.maxddd
aggdata['gip'] = aggdata.giprevenue / aggdata.ddd
aggdata['margin'] = aggdata.price - aggdata.gip

#Add source dummies
aggdata = aggdata.join(salesdata.groupby(agggroup)[sourcetemp.columns].max())
aggdata.reset_index(inplace=True)
aggdata['market'] = aggdata.groupby(['atccode', 'strength', 'date']).ngroup()
# Parallel trade information
aggdata['parallel'] = aggdata.firmname.isin(parfirm)
aggdata['parchainmarket'] = aggdata.groupby(['market', 'chainid']).parallel.transform('max')
aggdata['parmarket'] = aggdata.groupby('market').parallel.transform('max')
fullparmarket = (aggdata.groupby(['market', 'chainid']).parallel.max()
                 .groupby('market').sum() == 3
                )
fullparmarket.name = 'fullparmarket'
aggdata = aggdata.join(fullparmarket, on='market')

aggdata['ismaxoption'] = aggdata.margin == aggdata.groupby(['market', 'chainid']).margin.transform('max')

aggdata['marketddd'] = aggdata.groupby('market').ddd.transform('sum')
aggdata['share'] = aggdata.ddd / aggdata.marketddd

#Load exchange rates
exchdata = pd.read_table(f'{datadir}exchrates.csv')
exchdata['date'] = pd.to_datetime(exchdata.date, format='%Y-%m-%d')

# Add to market data
aggdata = aggdata.join(exchdata.set_index('date'), on='date')

# Add source country prices
exportdata = (imsdata[imsdata.atccode.isin(aggdata.atccode.unique())]
                       .groupby(['date', 'atccode', 'strength', 'country'], as_index=False)
                       [['quantity', 'sales']].sum()
             )
exportdata['price'] = exportdata.sales / exportdata.quantity
exportdata = (exportdata.groupby(['atccode', 'strength', 'country'])
              .apply(lambda x: x.set_index('date').price.resample('MS').bfill())
              .unstack('country')
             )

aggdata = aggdata.join(exportdata, on=['atccode', 'strength', 'date'])

# Convert to NOK/DDD
countrylist = ['Czech', 'France', 'Germany', 'Greece', 'Italy', 'Poland', 'Spain', 'UK']
exchlist = ['nokczk', 'nokeur', 'nokeur', 'nokeur', 'nokeur', 'nokpln', 'nokeur', 'nokgbp']

convertdata = exchdata.set_index('date')[exchlist]
convertdata.columns = countrylist

nokexport = exportdata.mul(convertdata, level='date')
nokexport.columns  = [f'{col}_nokddd' for col in nokexport.columns]

aggdata = aggdata.join(nokexport, on=['atccode', 'strength', 'date'])

dropcols = ['revenue', 'maxrevenue', 'giprevenue', 'maxddd']
aggdata.drop(dropcols, axis=1, inplace=True)
aggdata.rename(columns={'src_Czech Republic': 'src_Czech'}, inplace=True)
#Store aggregate data
varlabels = {'atccode': 'ATC code',
             'strength': 'milligram active ingredient per unit',
             'date': 'Date (monthly)',
             'chainid': 'Pharmacy chain ID',
             'firmname': 'Producer/importer name',
             'ddd': 'Product quantity in DDD',
             'price': 'NOK/DDD', 'maxprice': 'Max NOK/DDD',
             'gip': 'Input price NOK/DDD', 'margin': 'Margin NOK/DDD',
             'market': 'Market index (atccode x strength x date)',
             'parallel': 'Parallel trade products',
             'parchainmarket': 'Chain carries parallel trade in market',
             'parmarket': 'Parallel trade in market',
             'fullparmarket': 'All chains carry parallel trade in market',
             'ismaxoption': 'Product with highest margin in market for chain',
             'marketddd': 'Total DDD for market', 'share': 'market share for products',
             'nokchf': 'Exch. rate NOK/CHF', 'nokczk': 'Exch. rate NOK/CZK',
             'nokeur': 'Exch. rate NOK/EUR', 'nokgbp': 'Exch. rate NOK/GBP',
             'nokpln': 'Exch. rate NOK/PLN', 'nokusd': 'Exch. rate NOK/USD'}

sourcecountries = ['Belgium', 'Czech', 'EU', 'France', 'Greece',
                   'Italy', 'Poland', 'Spain', 'UK']
for country in sourcecountries:
    varlabels[f'src_{country}'] = f'License for import from {country}'

for country in countrylist:
    varlabels[country] = f'Manufacturer price in {country} in local currency per DDD'
    varlabels[f'{country}_nokddd'] = f'Manufacturer price in {country} in NOK/DDD'

aggdata.to_stata(f'{datadir}educed_form_data.dta',
                 write_index=False, convert_dates={'date': 'tm'},
                 variable_labels=varlabels)